var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var RogueRpg;
(function (RogueRpg) {
    var LosingLifeChunk = /** @class */ (function () {
        function LosingLifeChunk(widthAsPercent, offsetAsPercent, timeRemaining) {
            if (timeRemaining === void 0) { timeRemaining = LosingLifeChunk.DURATION; }
            this.widthAsPercent = widthAsPercent;
            this.offsetAsPercent = offsetAsPercent;
            this.timeRemaining = timeRemaining;
            this.offsetAsPercent = Math.max(this.offsetAsPercent, 0);
            this.widthAsPercent = Math.min(this.widthAsPercent, 1);
        }
        // Returns true if the chunk should be retained
        LosingLifeChunk.Update = function (chunk) {
            chunk.timeRemaining -= RogueRpg.Game.deltaTime;
            return chunk.timeRemaining > 0;
        };
        LosingLifeChunk.DURATION = 500;
        return LosingLifeChunk;
    }());
    RogueRpg.LosingLifeChunk = LosingLifeChunk;
    var FloatingText = /** @class */ (function () {
        // TODO: Floating Text rendering is very slow
        function FloatingText(text, color, drawX, drawY, combat, duration) {
            if (combat === void 0) { combat = false; }
            this.text = text;
            this.color = color;
            this.drawX = drawX;
            this.drawY = drawY;
            this.combat = combat;
            this.timeRemaining = duration || FloatingText.DEFAULT_DURATION;
            this.speed = 1.8;
        }
        FloatingText.GetSimpleColor = function (text) {
            text.color.a = 1;
            return Rgba.stringify(text.color);
        };
        FloatingText.GetColor = function (text) {
            text.color.a = FloatingText.GetOpacity(text);
            return Rgba.stringify(text.color);
        };
        FloatingText.GetOpacity = function (text) {
            return 0.4 + (text.timeRemaining / FloatingText.DEFAULT_DURATION);
        };
        FloatingText.DEFAULT_DURATION = 1500;
        return FloatingText;
    }());
    RogueRpg.FloatingText = FloatingText;
    // Represents a buff or other timed effect
    var ActiveEffect = /** @class */ (function () {
        function ActiveEffect(name, timeLeft, state, nameIsAnimationKey) {
            if (nameIsAnimationKey === void 0) { nameIsAnimationKey = false; }
            this.name = name;
            this.timeLeft = timeLeft;
            this.state = state;
            this.nameIsAnimationKey = nameIsAnimationKey;
        }
        // Because ActiveEffect is used as a prototype and an instance, you need to clone the prototype
        // before using the instance, or update() will mess up the prototype
        ActiveEffect.clone = function (effect) {
            return JSON.parse(JSON.stringify(effect));
        };
        return ActiveEffect;
    }());
    RogueRpg.ActiveEffect = ActiveEffect;
    var HeroState = /** @class */ (function () {
        function HeroState(useDefaults) {
            this.classChosen = false;
            this.healingPotionCount = 0;
            this.pets = [];
            this.itemStates = [];
            // Similar to other games, the items available are specific to the hero and not the map - in multiplayer the map can be shared
            this.itemsOnGround = [];
            if (useDefaults) {
                this.life = 40;
                this.baseMaxLife = 36;
                this.level = 1;
                this.baseStrength = 1;
                this.baseDexterity = 1;
                this.baseDamage = 0;
                this.baseDefense = 0;
                this.baseReduction = 0;
                this.luck = 1;
                this.baseLifesteal = 0;
                this.currentExperience = 0;
                this.neededExperience = 10;
                this.itemStates = [];
                this.potions = [];
                this.activeEffects = [];
                this.pets = [];
                this.x = 700;
                this.y = 700;
                this.floor = 1;
                this.levelUpPoints = 0;
                this.lastLevelBought = 1;
                this.classIndex = 0;
                this.multiplier = 1;
            }
        }
        return HeroState;
    }());
    RogueRpg.HeroState = HeroState;
    var DudeSpriteUtils = /** @class */ (function () {
        function DudeSpriteUtils() {
        }
        DudeSpriteUtils.GetImageNameForDude = function (dude) {
            // if monster is attacking
            if (dude.attackingTimer > 1) {
                if (dude.attackImages) {
                    var index = Math.floor((251 - dude.attackingTimer) / 250 * dude.attackImages.length);
                    return dude.attackImages[index];
                }
            }
            // if monster is hit
            if (dude.wasHitTimer > 0) {
                if (dude.hitImages) {
                    return dude.hitImages[0]; // TODO: Add more tracking
                }
            }
            // if monster is moving
            if (dude.movedThisLoop || dude.movingWhileStill) {
                if (dude.movingImages) {
                    var index = Math.floor(dude.movingTimer / (dude.movingAnimationDuration || 100)) % dude.movingImages.length;
                    return dude.movingImages[index];
                }
            }
            return dude.type;
        };
        return DudeSpriteUtils;
    }());
    RogueRpg.DudeSpriteUtils = DudeSpriteUtils;
    var Mood = /** @class */ (function () {
        function Mood() {
        }
        Mood.sad = function (frames) {
            Mood.Happiness -= frames;
        };
        Mood.happy = function (frames) {
            Mood.Happiness += frames;
        };
        Mood.GetHeroFace = function () {
            if (Mood.Happiness > 0) {
                Mood.Happiness--;
            }
            else if (Mood.Happiness < 0) {
                Mood.Happiness++;
            }
            if (Mood.Looking > 0) {
                Mood.Looking--;
            }
            else if (Mood.Looking < 0) {
                Mood.Looking++;
            }
            else {
                if (Math.random() < 0.0003) {
                    Mood.Looking = 37;
                }
                else if (Math.random() < 0.0003) {
                    Mood.Looking = -37;
                }
            }
            if (RogueRpg.Helpers.hero.attackingTime > 0) {
                Mood.Looking = 0;
                Mood.Happiness = 0;
                return Images.get("heroFaceAttacking");
            }
            else if (Mood.Happiness > 0) {
                return Images.get("heroHappy");
            }
            else if (Mood.Happiness < 0) {
                return Images.get("heroSad");
            }
            else if (Mood.Looking > 0) {
                return Images.get("heroLookingLeft");
            }
            else if (Mood.Looking < 0) {
                return Images.get("heroLookingRight");
            }
            else {
                return Images.get("heroFace");
            }
        };
        Mood.GetHeroImage = function () {
            var movementRatio = 50;
            var speed = RogueRpg.Helpers.hero.speed;
            if (RogueRpg.Level[RogueRpg.L.SpeedLimit] && speed > 35) {
                speed = 35;
            }
            Mood.MovingTimer = (Mood.MovingTimer + RogueRpg.Game.deltaTime) % (Mood.MoveSpeed * movementRatio / (speed));
            Mood.StoppingTimer -= RogueRpg.Game.deltaTime;
            if (RogueRpg.Helpers.hero.attackingTime > 0) {
                Mood.Looking = 0;
                Mood.Happiness = 0;
                var attackRatio = 1 - (RogueRpg.Helpers.hero.attackingTime / (RogueRpg.Helpers.hero.weaponDelay / 2));
                // return new HeroRenderingInfo("heroAttacking", new Location(6 - 10 * Math.cos(attackRatio), 37), 40 - 120 * (attackRatio), new Location(33, 47), 0);
                if (RogueRpg.Helpers.hero.attackingTime > RogueRpg.Helpers.hero.weaponDelay / 2 * 0.75) {
                    return HeroRender.Attacking[0];
                }
                else if (RogueRpg.Helpers.hero.attackingTime > RogueRpg.Helpers.hero.weaponDelay / 2 * 0.3) {
                    return HeroRender.Attacking[1];
                }
                return HeroRender.Attacking[2];
            }
            else if (RogueRpg.Helpers.game.movedThisCycle) {
                Mood.wasMoving = true;
                var movingSpeed = Mood.MoveSpeed * movementRatio / speed;
                var index = Math.floor(Mood.MovingTimer / movingSpeed * HeroRender.Moving.length);
                return HeroRender.Moving[index];
            } /*else if (Mood.wasMoving && Mood.StoppingTimer <= 0) {
                Mood.wasMoving = false;
                Mood.StoppingTimer = 400;
                return Images.get("hero stopping");
            } else if (Mood.StoppingTimer > 0) {
                return Images.get("hero stopping");
            } */
            else {
                return HeroRender.Idle;
            }
        };
        Mood.MovingTimer = 0;
        Mood.Happiness = 0;
        Mood.Looking = 0;
        Mood.wasMoving = false;
        Mood.StoppingTimer = 100;
        Mood.MoveSpeed = 400;
        return Mood;
    }());
    RogueRpg.Mood = Mood;
    var HeroRenderingInfo = /** @class */ (function () {
        function HeroRenderingInfo(image, weaponAttachmentPoint, weaponRotation, shieldAttachmentPoint, shieldRotation) {
            this.image = image;
            this.weaponAttachmentPoint = weaponAttachmentPoint;
            this.weaponRotation = weaponRotation;
            this.shieldAttachmentPoint = shieldAttachmentPoint;
            this.shieldRotation = shieldRotation;
        }
        return HeroRenderingInfo;
    }());
    RogueRpg.HeroRenderingInfo = HeroRenderingInfo;
    var HeroRender = /** @class */ (function () {
        function HeroRender() {
        }
        HeroRender.defaultWeapon = new RogueRpg.Location(7, 46);
        HeroRender.defaultShield = new RogueRpg.Location(33, 47);
        HeroRender.Idle = new HeroRenderingInfo("hero", new RogueRpg.Location(0, 0), 10, new RogueRpg.Location(0, 0), 0);
        HeroRender.Moving = [
            new HeroRenderingInfo("hero move0", new RogueRpg.Location(0, 0), 10, new RogueRpg.Location(0, 0), 0),
            new HeroRenderingInfo("hero move1", new RogueRpg.Location(0, 0), 10, new RogueRpg.Location(0, 0), 0),
            new HeroRenderingInfo("hero move2", new RogueRpg.Location(0, 0), 10, new RogueRpg.Location(0, 0), 0),
            new HeroRenderingInfo("hero move3", new RogueRpg.Location(0, 0), 10, new RogueRpg.Location(0, 0), 0),
            //new HeroRenderingInfo("hero move4", new Location(0, 0), 10, new Location(0, 0), 0),
            new HeroRenderingInfo("hero move5", new RogueRpg.Location(0, 0), 10, new RogueRpg.Location(0, 0), 0),
            new HeroRenderingInfo("hero move6", new RogueRpg.Location(0, 0), 10, new RogueRpg.Location(0, 0), 0),
            new HeroRenderingInfo("hero move7", new RogueRpg.Location(0, 0), 10, new RogueRpg.Location(0, 0), 0),
            new HeroRenderingInfo("hero move8", new RogueRpg.Location(0, 0), 10, new RogueRpg.Location(0, 0), 0),
        ];
        HeroRender.Attacking = [
            new HeroRenderingInfo("heroAttacking", new RogueRpg.Location(5, -5), 30, new RogueRpg.Location(-3, 3), 10),
            new HeroRenderingInfo("heroAttacking2", new RogueRpg.Location(5, 5), -50, new RogueRpg.Location(2, -6), -20),
            new HeroRenderingInfo("heroAttacking3", new RogueRpg.Location(8, 10), -75, new RogueRpg.Location(3, -9), -30),
        ];
        return HeroRender;
    }());
    RogueRpg.HeroRender = HeroRender;
    var Hero = /** @class */ (function () {
        function Hero(state) {
            this.baseSpeed = 29;
            this.remainingDelay = 0;
            this.specialDelay = 0;
            this.specialDelayStart = 5000; // todo: hard coded for Sweep
            this.drawWidth = 40;
            this.drawHeight = 80;
            this.attackingTime = 0;
            this.orbAngle = 0;
            this.orbPositions = [];
            this.cachedHeroState = false;
            if (state) {
                this.state = state;
                this.items = [];
                // Account for going from one version to the next
                if (!this.state.potions) {
                    this.state.potions = [];
                }
                for (var i = 0; i < this.state.itemStates.length; i++) {
                    var item = this.state.itemStates[i];
                    if (item.baseDamage) {
                        this.items.push(new RogueRpg.Weapon(0, 0, RogueRpg.Helpers.game.startingWeapon || "Scumbasher", item));
                    }
                    else if (item.baseReduction) {
                        this.items.push(new RogueRpg.Shield(0, 0, "", item));
                    }
                    else if (item.baseDefense) {
                        this.items.push(new RogueRpg.Armor(0, 0, "", item));
                    }
                    else {
                        this.items.push(new RogueRpg.Accessory(0, 0, "", item));
                    }
                }
                if (this.state.equippedWeaponState) {
                    this.equippedWeapon = new RogueRpg.Weapon(0, 0, "", this.state.equippedWeaponState);
                }
                if (this.state.equippedShieldState) {
                    this.equippedShield = new RogueRpg.Shield(0, 0, "", this.state.equippedShieldState);
                }
                if (this.state.equippedArmorState) {
                    this.equippedArmor = new RogueRpg.Armor(0, 0, "", this.state.equippedArmorState);
                }
                if (this.state.equippedAccessoryState) {
                    this.equippedAccessory = new RogueRpg.Accessory(0, 0, "", this.state.equippedAccessoryState);
                }
            }
            else {
                this.state = new HeroState(/*useDefaults=*/ true);
                this.equippedWeapon = new RogueRpg.Weapon(0, 1, RogueRpg.Helpers.game.startingWeapon || "Scumbasher"); // Start off with a Club
                this.life = this.maxLife;
                this.items = [];
                this.itemsOnGround = [];
            }
        }
        Object.defineProperty(Hero.prototype, "drawX", {
            // drawX is the top-left corner of the hero
            get: function () { return this.x - this.drawWidth / 2; },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Hero.prototype, "drawY", {
            get: function () { return this.y - this.drawHeight; },
            enumerable: true,
            configurable: true
        });
        Hero.prototype.getHeroState = function () {
            if (!this.cachedHeroState) {
                this.cachedHeroState = true;
                this.state.equippedWeaponState = this.equippedWeapon && this.equippedWeapon.getItemState();
                this.state.equippedShieldState = this.equippedShield && this.equippedShield.getItemState();
                this.state.equippedArmorState = this.equippedArmor && this.equippedArmor.getItemState();
                this.state.equippedAccessoryState = this.equippedAccessory && this.equippedAccessory.getItemState();
                this.state.itemStates = [];
                if (this.items) {
                    for (var i = 0; i < this.items.length; i++) {
                        this.state.itemStates.push(this.items[i].getItemState());
                    }
                    /* TODO: ISM: Clean this up but make sure the logic is extended into the map variant
                    this.state.itemsOnGround = [];
                    for (var i = 0; i < this.itemsOnGround.length; i++) {
                        this.state.itemsOnGround.push(this.itemsOnGround[i].getItemState());
                    }
                    */
                }
            }
            return this.state;
        };
        Object.defineProperty(Hero.prototype, "baseDelay", {
            get: function () {
                return this.equippedWeapon ? this.equippedWeapon.delay : 1000;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Hero.prototype, "baseLuck", {
            get: function () {
                return this.state.luck + RogueRpg.Level.Luck * 2 + RogueRpg.Level[RogueRpg.L.Luck2] * 3;
            },
            enumerable: true,
            configurable: true
        });
        Hero.prototype.clearStatCaches = function () {
            this.cachedMaxLife = null;
            this.cachedStr = null;
            this.cachedDex = null;
        };
        Hero.prototype.getExperienceMultiplier = function () {
            var mult = 1;
            mult *= 1 + (RogueRpg.Level[RogueRpg.L.DoubleExp]);
            mult *= 1 + (RogueRpg.Level[RogueRpg.L.BonusExp] * 0.12);
            if (this.life < this.maxLife * 0.6) {
                mult *= 1 + 0.25 * RogueRpg.Level[RogueRpg.L.BonusExpLowLife];
            }
            mult *= 1 + RogueRpg.Level[RogueRpg.L.ExpButWeak] * 0.35;
            mult *= 1 + this.sumFromItems("percent_experience") * 0.01;
            return mult;
        };
        Object.defineProperty(Hero.prototype, "critChance", {
            get: function () {
                var critChance = 0.1 + 0.1 * RogueRpg.Level[RogueRpg.L.IncreasedCritChance];
                if (RogueRpg.Level[RogueRpg.L.LuckOnCombatEffects]) {
                    critChance = this.getLuckBasedChance(critChance);
                }
                return critChance;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Hero.prototype, "maxLife", {
            // NOTE - maxLife cannot ever depend on other stats because they depend on it
            get: function () {
                if (!this.cachedMaxLife) {
                    var life = this.getMaxLifeFlatPart();
                    var mult = this.getMaxLifeMultiplierPart();
                    life *= mult;
                    this.cachedMaxLife = Math.round(life);
                }
                return this.cachedMaxLife;
            },
            enumerable: true,
            configurable: true
        });
        Hero.prototype.getMaxLifeFlatPart = function () {
            var life = this.baseMaxLife;
            life += RogueRpg.Level.Life * 6 + RogueRpg.Level[RogueRpg.L.Life2] * 9;
            life += RogueRpg.Level[RogueRpg.L.FlatLifeLuck] * 16;
            life += (RogueRpg.Level[RogueRpg.L.MoreMaxLife] || 0) * this.level * 5;
            life += this.sumFromItems("flat_life");
            return life;
        };
        Hero.prototype.getMaxLifeMultiplierPart = function () {
            var multiplier = 1;
            multiplier *= this.getStrengthLifeMult();
            ;
            multiplier *= 1 + this.sumFromItems("percent_life") / 100;
            multiplier *= 1 + ((RogueRpg.Level[RogueRpg.L.Life3] || 0) * 0.10);
            return multiplier;
        };
        Object.defineProperty(Hero.prototype, "strength", {
            get: function () {
                if (!this.cachedStr) {
                    var strength = this.baseStrength;
                    strength += RogueRpg.Level.Str * 2 + RogueRpg.Level[RogueRpg.L.Str2] * 3;
                    strength += RogueRpg.Level[RogueRpg.L.FlatDmgStr] * 4 + RogueRpg.Level[RogueRpg.L.FlatDefStr] * 4 + RogueRpg.Level[RogueRpg.L.FlatStrLuck] * 4;
                    strength += this.sumFromItems("flat_strength");
                    strength *= (1 + this.sumFromItems("percent_strength") / 100);
                    strength *= 1 + ((RogueRpg.Level[RogueRpg.L.Str3] || 0) * 0.10);
                    if (RogueRpg.Game.superHero) {
                        strength = (strength + 50) * 30;
                    }
                    this.cachedStr = Math.round(strength);
                }
                return this.cachedStr;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Hero.prototype, "dexterity", {
            get: function () {
                if (!this.cachedDex) {
                    var dexterity = this.baseDexterity;
                    dexterity += RogueRpg.Level.Dex * 2 + RogueRpg.Level[RogueRpg.L.Dex2] * 3;
                    dexterity += RogueRpg.Level[RogueRpg.L.FlatDefDex] * 4;
                    dexterity += this.sumFromItems("flat_dexterity");
                    dexterity *= (1 + this.sumFromItems("percent_dexterity") / 100);
                    dexterity *= 1 + ((RogueRpg.Level[RogueRpg.L.Dex3] || 0) * 0.10);
                    if (RogueRpg.Game.superHero) {
                        dexterity = (dexterity + 40) * 3;
                    }
                    this.cachedDex = Math.round(dexterity);
                }
                return this.cachedDex;
            },
            enumerable: true,
            configurable: true
        });
        Hero.prototype.getCooldownReduction = function () {
            var percentReduction = this.sumFromItems("percent_cooldowns");
            return percentReduction / 100;
        };
        Hero.prototype.getDexterityDelayMultiplierPercent = function () {
            var percent;
            var adjustedDex = this.dexterity * (1 + 0.2 * RogueRpg.Level[RogueRpg.L.DexBonusDelayRatio]);
            if (adjustedDex > 100) {
                adjustedDex = 100 + (adjustedDex - 100) * 0.95;
            }
            if (adjustedDex > 250) {
                adjustedDex = 250 + (adjustedDex - 250) * 0.9;
            }
            if (adjustedDex > 400) {
                adjustedDex = 400 + (adjustedDex - 400) * 0.85;
            }
            // SImulate Ratio boost by multiplying dex
            percent = 100 * (adjustedDex) / (adjustedDex + 100);
            return Math.floor(percent);
        };
        Hero.prototype.getDexterityDefenseMultiplierPercent = function () {
            var percent = 100 * (this.dexterity / 100);
            percent *= 1 + (RogueRpg.Level[RogueRpg.L.DexBonusDefRatio] || 0) * 0.2;
            return Math.round(percent);
        };
        Object.defineProperty(Hero.prototype, "speed", {
            get: function () {
                var speed = this.baseSpeed;
                speed += RogueRpg.Level.Speed + RogueRpg.Level[RogueRpg.L.Speed2];
                speed += RogueRpg.Level[RogueRpg.L.Sprinter] * 5;
                speed += this.sumFromItems("flat_speed");
                speed += 1 * (RogueRpg.Level[RogueRpg.L.FlatDefSpeed] + RogueRpg.Level[RogueRpg.L.FlatLuckSpeed]);
                if ((RogueRpg.Level[RogueRpg.L.SpeedWithFast] && this.equippedWeapon.isFast()) || (RogueRpg.Level[RogueRpg.L.SpeedWithSlow] && !this.equippedWeapon.isFast())) {
                    speed += 3;
                }
                speed *= (1 + this.sumFromItems("percent_speed") / 100);
                speed *= 1 + ((RogueRpg.Level[RogueRpg.L.Speed3] || 0) * 0.10);
                speed -= RogueRpg.Level[RogueRpg.L.ExpButWeak] * 0.10 * speed;
                if (RogueRpg.Game.superHero) {
                    speed = (speed + 12) * 1.2;
                }
                if (this.life >= this.maxLife * 0.95) {
                    speed += speed * RogueRpg.Level[RogueRpg.L.SpeedOnFullLife] * 0.40;
                }
                return Math.round(speed);
            },
            enumerable: true,
            configurable: true
        });
        // Default is 1% damage per strength
        Hero.prototype.getStrengthMultiplierPercent = function () {
            var percent = 100 * this.strength * (0.01 + ((RogueRpg.Level[RogueRpg.L.StrBonusDmgRatio] || 0) * 0.002));
            return Math.round(percent);
        };
        Hero.prototype.getStrengthLifeMult = function () {
            return 1 + this.strength * (0.01 + RogueRpg.Level[RogueRpg.L.StrBonusHpRatio] * 0.002);
        };
        Hero.prototype.getLifestealPercent = function () {
            var lifeSteal = this.baseLifesteal;
            lifeSteal += RogueRpg.Level[RogueRpg.L.FlatLifesteal] * 4;
            lifeSteal += this.sumFromItems("flat_lifesteal");
            return lifeSteal;
        };
        Object.defineProperty(Hero.prototype, "weaponDamage", {
            get: function () {
                var damage = this.getDamageFlatPart();
                var mult = this.getDamageMultPart();
                damage *= mult;
                damage += RogueRpg.Level[RogueRpg.L.LifeToDmg] * 0.01 * this.maxLife;
                if (RogueRpg.Game.superHero) {
                    damage = (damage + 125) * 25;
                }
                return Math.floor(damage);
            },
            enumerable: true,
            configurable: true
        });
        Hero.prototype.getDamageFlatPart = function () {
            var damage = this.state.baseDamage;
            damage += this.equippedWeapon ? this.equippedWeapon.damage : 1; // There should ALWAYS be a weapon, but whatever
            damage += this.sumFromItems("flat_damage", true);
            damage += RogueRpg.Level.Dmg + RogueRpg.Level[RogueRpg.L.Dmg2] * 2;
            damage += RogueRpg.Level[RogueRpg.L.FlatDmgStr] * 4;
            return damage;
        };
        Hero.prototype.getDamageMultPart = function () {
            var mult = 1;
            if ((RogueRpg.Level[RogueRpg.L.SpeedWithSlow] && !this.equippedWeapon.isFast())) {
                mult *= 1.15;
            }
            var statMultiplier = this.getStrengthMultiplierPercent() / 100;
            if (RogueRpg.Level[RogueRpg.L.UseDexInDmg]) {
                statMultiplier = statMultiplier / 2 + (this.dexterity / 100);
            }
            mult *= 1 + statMultiplier;
            mult *= 1 + ((RogueRpg.Level[RogueRpg.L.Dmg3] || 0) * 0.10);
            if (RogueRpg.Level[RogueRpg.L.SpeedLimit] && this.speed > 35) {
                mult *= 1 + (this.speed - 35) * 0.02;
            }
            mult *= (1 + this.sumFromItems("percent_damage", true) / 100);
            mult *= 1 + RogueRpg.Level[RogueRpg.L.Berserking] * 0.6 * (1 - this.life / this.maxLife); // Increases damage by up to 50% based on your missing life
            mult *= 1 + RogueRpg.Level[RogueRpg.L.SlowButStrong] * 0.60;
            mult *= 1 - RogueRpg.Level[RogueRpg.L.ExpButWeak] * 0.10;
            return mult;
        };
        Object.defineProperty(Hero.prototype, "reduction", {
            get: function () {
                var red = this.state.baseReduction;
                red += this.sumFromItems("flat_reduction", true);
                var prismaReductionRanks = RogueRpg.Helpers.game.prismaState.levels["Reduction"] || 0;
                red += prismaReductionRanks * PermanentUpgrade.Upgrades.Reduction.costPerRank; // Not a percent
                return red + (this.equippedShield ? this.equippedShield.reduction : 0);
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Hero.prototype, "defense", {
            get: function () {
                var def = this.state.baseDefense;
                def += this.equippedArmor ? this.equippedArmor.defense : 0;
                def += (RogueRpg.Level[RogueRpg.L.ToughSkin] || 0) * this.level;
                def += RogueRpg.Level.Def + RogueRpg.Level[RogueRpg.L.Def2];
                def += RogueRpg.Level[RogueRpg.L.FlatDefDex] * 8 + RogueRpg.Level[RogueRpg.L.FlatDefStr] * 8 + RogueRpg.Level[RogueRpg.L.FlatDefLuck] * 8 + RogueRpg.Level[RogueRpg.L.FlatDefSpeed] * 8;
                def += this.sumFromItems("flat_defense", true);
                def *= 1 + ((RogueRpg.Level[RogueRpg.L.Def3] || 0) * 0.10);
                var dexModifier = this.getDexterityDefenseMultiplierPercent() / 100;
                def *= 1 + dexModifier;
                def *= (1 + this.sumFromItems("percent_defense", true) / 100);
                if (RogueRpg.Game.superHero) {
                    def = (def + 900) * 40;
                }
                def += RogueRpg.Level[RogueRpg.L.DefOnLowLife] * (1 - this.life / this.maxLife) * 0.45 * def; // Adds up to 45% defense as your life approaches 0
                return Math.floor(def);
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Hero.prototype, "luck", {
            get: function () {
                var luck = this.getLuckFlatPart();
                luck *= this.getLuckMultPart();
                return Math.round(luck);
            },
            enumerable: true,
            configurable: true
        });
        Hero.prototype.getLuckFlatPart = function () {
            var luck = this.baseLuck;
            luck += this.sumFromItems("flat_luck");
            luck += 4 * (RogueRpg.Level[RogueRpg.L.FlatStrLuck] + RogueRpg.Level[RogueRpg.L.FlatDefLuck] + RogueRpg.Level[RogueRpg.L.FlatLifeLuck] + RogueRpg.Level[RogueRpg.L.FlatLuckSpeed]);
            luck += 12 * RogueRpg.Level[RogueRpg.L.BigFlatLuck];
            return luck;
        };
        Hero.prototype.getLuckMultPart = function () {
            var mult = 1;
            mult *= 1 + RogueRpg.Level[RogueRpg.L.Luck3] * 0.10;
            mult *= 1 + RogueRpg.Level[RogueRpg.L.Greedless] * 0.75;
            return mult;
        };
        // If you have 10 defense against a level 1 enemy, you have 50% reduction
        // If you have 20 defense against a level 1 enemy, you have 66% reduction
        // If you have 20 defense against a level 3 enemy, you have 50% reduction
        // You need 5 defense per enemy level 
        Hero.prototype.getDefenseReductionPercent = function (enemyLevel) {
            return 100 * (this.defense) / (this.defense + 5 + enemyLevel * 5);
        };
        Object.defineProperty(Hero.prototype, "weaponDelay", {
            get: function () {
                var baseDelay = this.getDelayFlatPart();
                if ((RogueRpg.Level[RogueRpg.L.SpeedWithFast] && this.equippedWeapon.isFast())) {
                    baseDelay *= 0.9;
                }
                baseDelay = baseDelay * (1 - (RogueRpg.Level[RogueRpg.L.ReducedDelay] || 0) * 0.1);
                var dexModifier = this.getDexterityDelayMultiplierPercent() / 100;
                var delay = baseDelay * (1 - dexModifier);
                delay *= 1 - RogueRpg.Level[RogueRpg.L.Berserking] * 0.3 * (1 - this.life / this.maxLife); // Increases attack speed by up to 25% based on your missing life
                delay *= this.accessoryAttackSpeedPercent(delay) / 100;
                delay *= 1 + 0.5 * RogueRpg.Level[RogueRpg.L.SlowButStrong];
                if (delay < 25) {
                    delay = (delay + 25) / 2;
                }
                if (isNaN(delay) || delay < 17) {
                    delay = 17;
                }
                var delayReductionFromUpgrades = RogueRpg.Helpers.game.prismaState.levels[PermanentUpgrade.Upgrades.AttackSpeed.id] || 0;
                delay *= 1 - (delayReductionFromUpgrades * PermanentUpgrade.Upgrades.AttackSpeed.gainPerRank * 0.02);
                return Math.floor(delay);
            },
            enumerable: true,
            configurable: true
        });
        Hero.prototype.getDelayFlatPart = function () {
            var delay = this.baseDelay;
            delay -= RogueRpg.Level[RogueRpg.L.FlatDelayReduction] * 70;
            return delay;
        };
        // Accessory reports in attacks per second, this converts to a multiplier
        Hero.prototype.accessoryAttackSpeedPercent = function (inputDelay) {
            var attacksPerSecond = 1000 / inputDelay;
            attacksPerSecond += this.equippedAccessory ? this.equippedAccessory.attacksPerSecond : 0;
            var newDelay = 1000 / attacksPerSecond;
            return (newDelay / inputDelay) * 100;
        };
        Object.defineProperty(Hero.prototype, "attackRange", {
            get: function () {
                if (HeroClass.Classes[this.getHeroState().classIndex].isRanged) {
                    return 1000;
                }
                return 75;
            },
            enumerable: true,
            configurable: true
        });
        // In theory there may be different heros, so this encapsulates stat gain and needed exp
        Hero.prototype.levelUp = function () {
            Mood.happy(60);
            this.level++;
            this.state.levelUpPoints++;
            if (this.level == 1) {
                this.neededExperience = 7;
            }
            else if (this.level <= 30) {
                this.neededExperience = 17 * this.level * Math.pow(1.061, this.level);
            }
            else if (this.level <= 60) {
                this.neededExperience = 22 * this.level * Math.pow(1.063, this.level);
            }
            else if (this.level <= 100) {
                this.neededExperience = 27 * this.level * Math.pow(1.066, this.level);
            }
            else if (this.level <= 200) {
                this.neededExperience = 27 * this.level * Math.pow(1.071, this.level);
            }
            else if (this.level <= 300) {
                this.neededExperience = 30 * this.level * Math.pow(1.075, this.level);
            }
            else if (this.level <= 400) {
                this.neededExperience = 30 * this.level * Math.pow(1.08, this.level);
            }
            else if (this.level <= 500) {
                this.neededExperience = 30 * this.level * Math.pow(1.09, this.level);
            }
            else if (this.level <= 600) {
                this.neededExperience = 30 * this.level * Math.pow(1.1, this.level);
            }
            else {
                this.neededExperience = 30 * this.level * Math.pow(1.1, this.level);
                this.neededExperience *= Math.pow(2, this.level - 600);
            }
            var lifegain = 5 + this.level * Math.pow(this.level / 2, 0.21);
            this.life += lifegain;
            this.baseMaxLife += lifegain;
            this.baseStrength += 2;
            this.baseDexterity += 2;
            this.clearStatCaches();
        };
        Hero.prototype.sumFromItems = function (property, onlyNonItems) {
            if (onlyNonItems === void 0) { onlyNonItems = false; }
            var val = 0;
            if (!onlyNonItems) {
                val += this.equippedWeapon ? this.equippedWeapon.getPropertyValue(property) : 0;
                val += this.equippedShield ? this.equippedShield.getPropertyValue(property) : 0;
                val += this.equippedArmor ? this.equippedArmor.getPropertyValue(property) : 0;
                val += this.equippedAccessory ? this.equippedAccessory.getPropertyValue(property) : 0;
            }
            if (property == HeroClass.Classes[this.getHeroState().classIndex].bonus) {
                val += HeroClass.Classes[this.getHeroState().classIndex].amount;
            }
            for (var i = 0; i < Relic.Relics.length; i++) {
                if (Relic.RelicLevels[i] > 0) {
                    var stat = Relic.Relics[i].statsPerLevel[property];
                    if (stat && stat > 0) {
                        val += stat * Relic.RelicLevels[i];
                    }
                }
            }
            return val;
        };
        // Given some number, like 0.3, and a luck like 200, computes the equivalent chance
        // TODO: Only works on positive numbers
        Hero.prototype.getLuckBasedChance = function (originalChance) {
            var luckCounter = this.luck;
            var antiChance = 1 - originalChance;
            while (luckCounter >= 100) {
                luckCounter -= 100;
                antiChance *= (1 - originalChance);
            }
            // luckCounter is now somewhere between 0 and 99 inclusive
            antiChance *= 1 - (luckCounter / 100) * originalChance;
            return 1 - antiChance;
        };
        Object.defineProperty(Hero.prototype, "x", {
            // All this is used to make persisting the hero state easy
            get: function () { return this.state.x; },
            set: function (val) { this.state.x = val; },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Hero.prototype, "y", {
            get: function () { return this.state.y; },
            set: function (val) { this.state.y = val; },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Hero.prototype, "level", {
            get: function () { return this.state.level; },
            set: function (val) { this.state.level = val; },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Hero.prototype, "baseStrength", {
            get: function () { return this.state.baseStrength; },
            set: function (val) { this.state.baseStrength = val; },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Hero.prototype, "baseDexterity", {
            get: function () { return this.state.baseDexterity; },
            set: function (val) { this.state.baseDexterity = val; },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Hero.prototype, "baseLifesteal", {
            get: function () { return this.state.baseLifesteal; },
            set: function (val) { this.state.baseLifesteal = val; },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Hero.prototype, "life", {
            get: function () { return this.state.life; },
            set: function (val) { this.state.life = val; },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Hero.prototype, "baseMaxLife", {
            get: function () { return this.state.baseMaxLife; },
            set: function (val) { this.state.baseMaxLife = val; },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Hero.prototype, "currentExperience", {
            get: function () { return this.state.currentExperience; },
            set: function (val) { this.state.currentExperience = val; },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Hero.prototype, "neededExperience", {
            get: function () { return this.state.neededExperience; },
            set: function (val) { this.state.neededExperience = val; },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Hero.prototype, "floor", {
            get: function () { return this.state.floor; },
            set: function (val) { this.state.floor = val; },
            enumerable: true,
            configurable: true
        });
        return Hero;
    }());
    RogueRpg.Hero = Hero;
    var Tile = /** @class */ (function () {
        function Tile() {
        }
        return Tile;
    }());
    RogueRpg.Tile = Tile;
    // Note that map is a pure-data item
    var Map = /** @class */ (function () {
        function Map() {
            this.tileSize = 80;
            this.mapWidth = 10;
            this.mapHeight = 10;
            this.leftMostRoom = 1000;
            this.bottomMostRoom = 0;
            this.rightMostRoom = 0;
            this.topMostRoom = 1000;
            this.tiles = [];
            this.monsters = [];
            this.healingOnGround = [];
            this.shardsOnGround = [];
            this.gemstonesOnGround = [];
            this.specialWorldObjects = [];
            this.deadMonsters = [];
            this.potionsOnGround = [];
            this.itemsOnGround = [];
            this.goldOnGround = [];
        }
        // The DroppedItem contains a 'Item' that is not persisted as a static object, but as a separate state-only state object,
        // and it needs to be reconstructed. This should be improved at some point, but I'm not digging in too deep while I migrate
        // itemsOnGround from the hero+heroState to the Map so it works well with the sublevel stacks
        Map.HydrateDroppedItems = function (map) {
            var cache = map.itemsOnGround;
            map.itemsOnGround = [];
            map.goldOnGround = map.goldOnGround || [];
            if (cache) {
                for (var i = 0; i < cache.length; i++) {
                    var di = cache[i];
                    if (di.state) {
                        map.itemsOnGround.push(new RogueRpg.DroppedItem(di.state));
                    }
                }
            }
        };
        return Map;
    }());
    RogueRpg.Map = Map;
    var WorldObject = /** @class */ (function () {
        function WorldObject(type, data) {
            this.drawWidth = 40;
            this.drawHeight = 80;
            this.isUnit = false; // If true, this is also a unit, presumably attackable? Maybe I should make this a monster instead
            if (type) {
                this.type = type;
            }
            if (data) {
                this.data = data;
            }
        }
        return WorldObject;
    }());
    RogueRpg.WorldObject = WorldObject;
    var TestMonsterWorldObject = /** @class */ (function (_super) {
        __extends(TestMonsterWorldObject, _super);
        function TestMonsterWorldObject(monsterType) {
            var _this = _super.call(this, "TestMonster", monsterType) || this;
            _this.timer = 0;
            var monster = new Monster(monsterType, null, 0);
            _this.monster = monster;
            return _this;
        }
        TestMonsterWorldObject.Update = function (obj) {
            obj.timer += RogueRpg.Game.deltaTime;
            obj.monster.attackingTimer -= RogueRpg.Game.deltaTime;
            if (obj.timer <= 2000) {
                obj.monster.movedThisLoop = true;
                obj.monster.movingTimer += RogueRpg.Game.deltaTime;
            }
            else if (obj.timer < 4000) {
                obj.monster.movedThisLoop = false;
                if (obj.monster.attackingTimer <= -obj.monster.delay / 2) {
                    obj.monster.attackingTimer = 250;
                }
            }
            else if (obj.timer < 6000) {
                obj.monster.movedThisLoop = true;
                obj.monster.facingRight = true;
                obj.monster.movingTimer += RogueRpg.Game.deltaTime;
            }
            else if (obj.timer < 8000) {
                obj.monster.movedThisLoop = false;
                if (obj.monster.attackingTimer <= -obj.monster.delay) {
                    obj.monster.attackingTimer = 250;
                }
            }
            else if (obj.timer >= 8000) {
                obj.timer = 0;
                obj.monster.facingRight = false;
            }
        };
        TestMonsterWorldObject.GetImage = function (mon) {
            var imageName = DudeSpriteUtils.GetImageNameForDude(mon.monster);
            return Images.get(imageName);
        };
        // Performs each of these actions for 2 seconds then repeats
        TestMonsterWorldObject.animationStates = ["MoveLeft", "Attack", "MoveRight", "Attack"];
        return TestMonsterWorldObject;
    }(WorldObject));
    RogueRpg.TestMonsterWorldObject = TestMonsterWorldObject;
    var DeadMonster = /** @class */ (function () {
        function DeadMonster(monster) {
            this.monster = monster;
            this.timer = 0;
        }
        DeadMonster.Update = function (dm) {
            if (dm.timer < 1000) {
                dm.timer += RogueRpg.Game.deltaTime;
                var type = dm.monster.movingImages ? dm.monster.movingImages[0] : dm.monster.type;
                RogueRpg.Helpers.game.renderer.context.drawImage(Images.get(type), dm.monster.drawX - RogueRpg.Helpers.OriginX, dm.monster.drawY + (dm.monster.drawHeight * dm.timer / 1000) - RogueRpg.Helpers.OriginY, dm.monster.drawWidth, dm.monster.drawHeight * (1000 - dm.timer) / 1000);
                return true;
            }
            else {
                return false;
            }
        };
        return DeadMonster;
    }());
    RogueRpg.DeadMonster = DeadMonster;
    var Dude = /** @class */ (function () {
        function Dude() {
            this.tookDamageTimer = 0;
            this.movingTimer = 0;
            this.movedThisLoop = false;
            this.attackingTimer = 0;
            this.movingWhileStill = false;
            this.wasHitTimer = 0;
            this.relatiateTimer = RogueRpg.Game.deltaTime * 4;
            // Amount to reduce all incoming damage by
            this.toughness = 1;
        }
        Dude.containsX = function (x, dude, margin) {
            if (margin === void 0) { margin = 0; }
            return x + margin > dude.x - (dude.drawWidth / 2) && x - margin < dude.x + (dude.drawWidth / 2);
        };
        Dude.containsY = function (y, dude, margin) {
            if (margin === void 0) { margin = 0; }
            return y + margin > dude.y - (dude.drawHeight / 2) && y - margin < dude.y + (dude.drawHeight / 2);
        };
        Dude.addActiveEffect = function (dude, effect) {
            for (var i = 0; i < dude.activeEffects.length; i++) {
                var e = dude.activeEffects[i];
                if (e.name === effect.name) {
                    dude.activeEffects.splice(i, 1);
                    i--;
                }
            }
            dude.activeEffects.push(ActiveEffect.clone(effect));
        };
        return Dude;
    }());
    RogueRpg.Dude = Dude;
    var Pet = /** @class */ (function (_super) {
        __extends(Pet, _super);
        function Pet(type, x, y) {
            var _this = _super.call(this) || this;
            _this.agroRange = 500;
            _this.activeEffects = [];
            _this.x = x;
            _this.y = y;
            _this.isPet = true;
            _this.type = type;
            _this.speed = 25;
            _this.drawWidth = 50;
            _this.drawHeight = 50;
            _this.agroRange = 500;
            _this.attackRange = 5;
            _this.movingAnimationDuration = 100;
            _this.drawX = _this.x - (_this.drawWidth / 2);
            _this.drawY = _this.y - _this.drawHeight;
            if (type === "Sparks1") {
                _this.maxLife = 50000;
                _this.delay = 1000;
                _this.damage = 2;
                _this.level = 1;
                _this.drawWidth = 50;
                _this.movingWhileStill = true;
                _this.attackImages = ["Sparks3"];
                _this.movingImages = ["Sparks1", "Sparks2"];
                _this.movingAnimationDuration = 250;
            }
            else if (type === "Puppy") {
                _this.maxLife = 50;
                _this.delay = 1000;
                _this.damage = 2;
                _this.level = 1;
                _this.drawWidth = 50;
                _this.movingWhileStill = true;
                _this.attackImages = [];
                _this.movingImages = ["Puppy Walk 1", "Puppy Walk 2", "Puppy Walk 1", "Puppy Walk 3"];
                _this.stateImages = {
                    "Sit": ["Puppy Sit 1", "Puppy Sit 2", "Puppy Sit 1", "Puppy Sit 3"]
                };
                _this.movingAnimationDuration = 200;
            }
            return _this;
        }
        Pet.PetFromMonster = function (mon) {
            mon.path = null;
            var pet = JSON.parse(JSON.stringify(mon));
            // TODO: Add any pet-specific stuff here
            pet.isMonster = false;
            pet.isPet = true;
            pet.life = pet.maxLife;
            return pet;
        };
        return Pet;
    }(Dude));
    RogueRpg.Pet = Pet;
    var Monster = /** @class */ (function (_super) {
        __extends(Monster, _super);
        function Monster(type, map, multiplierTier, locationOverride, flatMultiplier) {
            var _this = _super.call(this) || this;
            _this.isAgroed = false;
            _this.skipLoot = false;
            _this.isMonster = true;
            _this.type = type;
            _this.speed = 17;
            _this.drawWidth = 40;
            _this.drawHeight = 80;
            _this.attackRange = 50;
            _this.movingAnimationDuration = 100;
            _this.activeEffects = [];
            if (type === "Goblin") {
                _this.maxLife = 5;
                _this.delay = 1000;
                _this.damage = 2;
                _this.experienceGrantedOnKill = 1;
                _this.level = 1;
                _this.drawWidth = 60;
                // The centerOffset is the x difference between the center of the image and the center of the unit in the image
                _this.movingWhileStill = false;
                _this.attackImages = ["GoblinAttack1"];
                _this.movingImages = ["GoblinMove1", "GoblinMove2"];
                _this.movingAnimationDuration = 100;
            }
            else if (type === "Bat") {
                _this.maxLife = 12;
                _this.delay = 1200;
                _this.damage = 4;
                _this.experienceGrantedOnKill = 3;
                _this.level = 2;
                _this.drawHeight = 80;
                _this.drawWidth = 80;
                _this.attackImages = ["BatAttack"];
                _this.hitImages = null;
                _this.movingWhileStill = true;
                _this.movingImages = ["BatMove1", "BatMove2", "BatMove3", "BatMove4"];
            }
            else if (type == "Boarman") {
                _this.maxLife = 40;
                _this.delay = 2000;
                _this.damage = 15;
                _this.experienceGrantedOnKill = 14; // This gets *2 for rare
                _this.level = 1; // Will be bumped up by "isRare" to '3' for loot
                _this.isRare = true;
                _this.attackRange = 90;
                _this.drawWidth = 90;
                _this.drawHeight = 90;
            }
            else if (type === "Mud Clown") {
                _this.maxLife = 23;
                _this.delay = 850;
                _this.damage = 9;
                _this.speed = 12;
                _this.experienceGrantedOnKill = 5;
                _this.level = 4;
                _this.drawWidth = 60;
                _this.attackImages = ["Mud Clown Attack"];
                _this.movingWhileStill = false;
                _this.movingImages = ["Mud Clown Move 1", "Mud Clown Move 2"];
                // TODO: Mud Blood is disabled because image doesn't fit
                /*if (Helpers.game.playerRandom(true) > 0.985 && false) {
                    this.type = "Mud Blood";
                    this.isRare = true;
                }*/
            }
            else if (type === "Goo") {
                _this.maxLife = 36;
                _this.delay = 1000;
                _this.damage = 15;
                _this.experienceGrantedOnKill = 8;
                _this.level = 6;
                _this.attackImages = ["Goo Attack 1", "Goo Attack 2", "Goo Attack 3", "Goo Attack 4", "Goo Attack 5"];
                _this.movingWhileStill = false;
                _this.movingImages = ["Goo Move 1", "Goo Move 2", "Goo Move 3", "Goo Move 4"];
                _this.drawWidth = 80;
                _this.drawHeight = 80;
                /*if (false && Helpers.game.playerRandom(true) > 0.99) {
                    this.type = "Kingu";
                    this.isRare = true;
                }*/
            }
            else if (type === "Thorny Cactus") {
                _this.speed = 0;
                _this.maxLife = 75;
                _this.delay = 750;
                _this.damage = 20;
                _this.experienceGrantedOnKill = 12;
                _this.level = 8;
                _this.attackRange = 90;
                _this.drawWidth = 80;
                _this.attackImages = ["Thorny Cactus Attack 1", "Thorny Cactus Attack 2"];
                // if (Math.random() < 0.02) {
                //     this.type = "Sombrero Cactus";
                //     this.isRare = true;
                // }
            }
            else if (type === "Scorpiar") {
                _this.maxLife = 130;
                _this.delay = 1000;
                _this.damage = 31;
                _this.experienceGrantedOnKill = 16;
                _this.level = 10;
                _this.drawHeight = 80;
                _this.drawWidth = 80;
                _this.attackImages = ["ScorpiarAttack"];
                _this.movingWhileStill = false;
                _this.movingImages = ["ScorpiarMove1", "ScorpiarMove2"];
                /*if (false && Math.random() < 0.02) {  // TODO: NO images for this
                    this.type = "Triscorpiar";
                    this.isRare = true;
                }*/
            }
            else if (type === "Flying Skull") {
                _this.speed = 29;
                _this.maxLife = 155;
                _this.delay = 800;
                _this.damage = 38;
                _this.experienceGrantedOnKill = 21;
                _this.level = 13;
                _this.drawWidth = 100;
                _this.drawHeight = 100;
                _this.attackImages = ["Flying Skull Attack"];
                _this.movingWhileStill = false;
                _this.movingImages = ["Flying Skull Move1", "Flying Skull Move2"];
                /*if (false && Math.random() < 0.05) { // TODO: NO animated images for this
                    this.type = "Diseased Skull";
                    this.isRare = true;
                }*/
            }
            else if (type === "Flytrap") {
                _this.speed = 0;
                _this.maxLife = 270;
                _this.delay = 900;
                _this.damage = 78;
                _this.experienceGrantedOnKill = 27;
                _this.level = 15;
                _this.attackRange = 90;
                _this.drawWidth = 70;
                _this.attackImages = ["Flytrap Attack"];
                // TODO: Disabled due to attack animation missing
                /*if (Helpers.game.playerRandom(true) > 0.99 && false) {
                    this.type = "Flytrapped";
                    this.isRare = true;
                }*/
            }
            else if (type === "Wee Demon") {
                _this.maxLife = 380;
                _this.delay = 1000;
                _this.damage = 90;
                _this.experienceGrantedOnKill = 35;
                _this.level = 16;
                if (RogueRpg.Helpers.game.playerRandom(true) > 0.992) {
                    _this.type = "Lava Demon";
                    _this.isRare = true;
                }
            }
            else if (type === "Mad Cube") {
                _this.maxLife = 490;
                _this.delay = 575;
                _this.damage = 100;
                _this.experienceGrantedOnKill = 43;
                _this.level = 17;
                if (RogueRpg.Helpers.game.playerRandom(true) > 0.992) {
                    _this.type = "Sad Cube";
                    _this.isRare = true;
                }
                else {
                    _this.attackImages = ["MadCubeAttack"];
                    _this.movingWhileStill = false;
                    _this.movingImages = ["MadCubeMove1", "MadCubeMove2"];
                }
            }
            else if (type === "Bad Hero") {
                _this.maxLife = 595;
                _this.delay = 800;
                _this.damage = 115;
                _this.experienceGrantedOnKill = 52;
                _this.level = 19;
                if (RogueRpg.Helpers.game.playerRandom(true) > 0.995) {
                    _this.type = "Super Bad Hero";
                    _this.isRare = true;
                }
            }
            else if (type === "Ghost") {
                _this.maxLife = 660;
                _this.delay = 800;
                _this.damage = 125;
                _this.speed = 29;
                _this.experienceGrantedOnKill = 62;
                _this.level = 21;
                if (RogueRpg.Helpers.game.playerRandom(true) > 0.99) {
                    _this.type = "Roast Ghost";
                    _this.isRare = true;
                }
                else {
                    _this.attackImages = ["GhostAttack"];
                    _this.movingWhileStill = true;
                    _this.movingImages = ["GhostMove1", "GhostMove2"];
                }
            }
            else if (type === "Snowman") {
                _this.maxLife = 850;
                _this.delay = 800;
                _this.damage = 170;
                _this.experienceGrantedOnKill = 72;
                _this.level = 22;
                _this.drawWidth = 80;
                _this.drawHeight = 80;
                _this.attackImages = ["Snowman Attack"];
                _this.movingWhileStill = false;
                _this.movingImages = ["Snowman Move 1", "Snowman Move 2"];
                /*if (false && Helpers.game.playerRandom(true) > 0.99) { // TODO: No images for this
                    this.type = "Bowman";
                    this.isRare = true;
                }*/
            }
            else if (type === "Eye") {
                _this.maxLife = 1040;
                _this.delay = 800;
                _this.damage = 280;
                _this.experienceGrantedOnKill = 85;
                _this.level = 24;
                if (RogueRpg.Helpers.game.playerRandom(true) > 0.98) {
                    _this.type = "Maw";
                    _this.isRare = true;
                }
                else {
                    _this.attackImages = ["EyeAttack"];
                    _this.movingWhileStill = true;
                    _this.movingImages = ["EyeMove1", "EyeMove2"];
                }
            }
            else if (type === "Snake") {
                _this.maxLife = 1250;
                _this.delay = 800;
                _this.damage = 390;
                _this.experienceGrantedOnKill = 90;
                _this.level = 26;
                if (RogueRpg.Helpers.game.playerRandom(true) > 0.98) {
                    _this.type = "Lava Snake";
                    _this.isRare = true;
                }
            }
            else if (type === "Minotaur") {
                _this.maxLife = 1300;
                _this.delay = 1300;
                _this.damage = 600;
                _this.experienceGrantedOnKill = 100;
                _this.level = 28;
                _this.drawWidth = 100;
                _this.drawHeight = 150;
                _this.attackRange = 120;
                _this.attackImages = ["MinotaurAttack"];
            }
            else if (type === "Fireling") {
                _this.level = RogueRpg.Helpers.game.hero.level;
                _this.maxLife = Math.ceil(4200 * RogueRpg.Helpers.game.hero.weaponDamage / RogueRpg.Helpers.game.hero.weaponDelay);
                var reduc = 1 - (RogueRpg.Helpers.game.hero.getDefenseReductionPercent(_this.level) / 100);
                _this.damage = Math.ceil(RogueRpg.Helpers.game.hero.maxLife * 0.032 / reduc + RogueRpg.Helpers.game.hero.reduction);
                _this.delay = 600;
                _this.experienceGrantedOnKill = Math.ceil(RogueRpg.Helpers.game.hero.neededExperience * 0.1);
                _this.drawWidth = 50;
                _this.drawHeight = 75;
                _this.speed = 30;
                if (Math.random() < 0.15) {
                    _this.type = "Torchling";
                    _this.isRare = true;
                }
                else {
                    _this.attackImages = ["FirelingAttack"];
                    _this.movingWhileStill = false;
                    _this.movingImages = ["FirelingMove1", "FirelingMove2"];
                }
            }
            if (_this.speed > 5) {
                _this.speed += randIntMap(0, 5) - 2;
            }
            var heroMultiplier = RogueRpg.Helpers.hero.state.multiplier;
            if (type !== "Fireling" && type !== "Torchling") {
                _this.maxLife *= heroMultiplier;
                _this.damage *= heroMultiplier;
            }
            // Additional Multiplier changes
            _this.toughness = 1;
            if (heroMultiplier >= 4) {
                var scalingFactor = {
                    '4': 0.02,
                    '8': 0.025,
                    '16': 0.03,
                    '50': 0.04,
                    '100': 0.05,
                    '200': 0.05
                };
                var defFactor = {
                    '4': 0.01,
                    '8': 0.02,
                    '16': 0.03,
                    '50': 0.04,
                    '100': 0.05,
                    '200': 0.06
                };
                var floorBasedScaling = scalingFactor['' + heroMultiplier];
                if (floorBasedScaling) {
                    // Increase damage by a small amount per floor and life by half of that
                    var floorBasedMultiplier = 1 + (RogueRpg.Helpers.hero.state.floor * floorBasedScaling / 2);
                    _this.maxLife *= floorBasedMultiplier;
                    _this.damage *= floorBasedMultiplier;
                }
                var floorBasedToughness = defFactor['' + heroMultiplier];
                if (floorBasedToughness && RogueRpg.Helpers.hero.state.floor > 2) {
                    var floor = RogueRpg.Helpers.hero.state.floor - 2;
                    var toughness = Math.pow((1 - floorBasedToughness), floor);
                    _this.toughness = toughness;
                }
            }
            if (flatMultiplier) {
                _this.maxLife *= flatMultiplier;
                _this.experienceGrantedOnKill *= flatMultiplier;
                _this.damage *= flatMultiplier;
                _this.dropRateModifier = flatMultiplier;
            }
            if (_this.isRare) {
                _this.maxLife = _this.maxLife * 3;
                _this.level = _this.level + 2;
                _this.delay = Math.ceil(_this.delay * 0.8);
                _this.experienceGrantedOnKill = _this.experienceGrantedOnKill * 2;
            }
            var mult = 1.0;
            while (multiplierTier > 0) {
                _this.maxLife *= 55 * mult;
                _this.maxLife += 1700;
                _this.damage *= 60 * mult;
                _this.damage += 390;
                _this.delay *= 0.7;
                _this.experienceGrantedOnKill *= 5;
                _this.experienceGrantedOnKill += 90;
                _this.level += 26;
                if (_this.speed > 0) {
                    _this.speed = _this.speed * 1.06;
                    _this.speed = _this.speed + 2;
                }
                multiplierTier -= 1;
                mult += 0.15;
            }
            _this.life = _this.maxLife;
            if (map) {
                var position = locationOverride || RogueRpg.Helpers.game.getRandomWalkableTilePosition(map);
                _this.x = position.x;
                _this.y = position.y;
                _this.drawX = _this.x - (_this.drawWidth / 2);
                _this.drawY = _this.y - _this.drawHeight;
            }
            return _this;
        }
        Monster.ResetButtonMonster = {
            type: "click on the Restart Button"
        };
        return Monster;
    }(Dude));
    RogueRpg.Monster = Monster;
    var Queue = /** @class */ (function () {
        function Queue() {
            this.inner = [];
        }
        Queue.prototype.enqueue = function (t) {
            this.inner.push(t);
        };
        Queue.prototype.dequeue = function () {
            return this.inner.splice(0, 1)[0];
        };
        Queue.prototype.empty = function () {
            return this.inner.length === 0;
        };
        return Queue;
    }());
    RogueRpg.Queue = Queue;
    var BridgedImage = /** @class */ (function () {
        function BridgedImage(image, x1, y1, x2, y2, timeLeft) {
            this.image = image;
            this.x1 = x1;
            this.y1 = y1;
            this.x2 = x2;
            this.y2 = y2;
            this.timeLeft = timeLeft;
        }
        return BridgedImage;
    }());
    RogueRpg.BridgedImage = BridgedImage;
    var Messages = /** @class */ (function () {
        function Messages() {
        }
        Messages.life = function (hero) {
            var text = "你目前有 " + green(Math.round(hero.life)) + "</span> 生命的最大限度 " + green(hero.maxLife) + ".";
            text += Messages.newlines + "你有 " + green(Math.round(hero.getMaxLifeFlatPart())) + " 基本的生命。这进一步增加了。 " + green(percentIncr(hero.getMaxLifeMultiplierPart())) + "% 从装备，力量和技能。.";
            text += Messages.newlines + "你会痊愈的" + green(Math.round(hero.maxLife * 0.25)) + " 当你拾起一颗心。";
            if (RogueRpg.Level[RogueRpg.L.NoLifesteal]) {
                text += Messages.newlines + "你治好了 " + green(Math.round(hero.maxLife * 0.006)) + " 每次攻击。";
            }
            else {
                text += Messages.newlines + "你治好了 " + green(Math.round(hero.getLifestealPercent())) + "% 你所造成的伤害。";
            }
            return text;
        };
        Messages.exp = function (hero) {
            var text = "你目前的水平 " + green(hero.level) + ".";
            text += Messages.newlines + "你需要 " + green(Math.ceil(hero.neededExperience - hero.currentExperience)) + " 更多的经验达到水平。" + (hero.level + 1) + ".";
            var expMult = hero.getExperienceMultiplier() * (1 + 0.12 * RogueRpg.Level[RogueRpg.L.ChanceDoubleExp]);
            text += Messages.newlines + "你获得的所有经验都是增加的。" + green(percentIncr(expMult)) + "%";
            return text;
        };
        Messages.damage = function (hero) {
            var text = "";
            var m = RogueRpg.Level[RogueRpg.L.LuckyStrike] ? 1 : 0.5;
            text += "你会交易 " + green(Math.round(hero.weaponDamage * (1 - m))) + "-" + green(Math.round(hero.weaponDamage * (1 + m))) +
                " 每一次攻击造成的伤害 " + green(hero.weaponDelay / 1000) + " 秒。";
            var averageDamageMultiplier = (1 - m + 2 * m * RogueRpg.Helpers.game.expectedValueFromCurrentLuck());
            text += Messages.newlines + "运气好的话，你的平均命中率将会是。 " +
                yellow("" + Math.round((100 * hero.weaponDamage * averageDamageMultiplier)) / 100.0);
            text += Messages.newlines + "你的基础伤害 " + green(hero.getDamageFlatPart()) + ". 这是增加了 " +
                green(percentIncr(hero.getDamageMultPart())) + "%";
            text += Messages.newlines + "你的基本CD是 " + green(hero.getDelayFlatPart()) + ",减少到 " +
                green(hero.weaponDelay) + " 受到各种影响。";
            var critDamage = 200 + RogueRpg.Level[RogueRpg.L.IncreasedCritDamage] * 100;
            text += Messages.newlines + "你有一个" + yellow(Math.floor(hero.critChance * 100)) + "% 致命一击的机会, " + yellow(critDamage) + "% 伤害";
            var aMon = RogueRpg.Helpers.game.getAMonsterForFloor(hero.floor);
            if (aMon && aMon.toughness < 1) {
                var reducedPercent = Math.round(1000 * (1 - aMon.toughness)) / 10;
                text += Messages.newlines + "这里的怪物很厉害！他们拿走了 " + red('' + reducedPercent) + "% 更少的伤害";
            }
            return text;
        };
        Messages.defense = function (hero) {
            var text = "你有 " + green(hero.defense) + " 防御和" + green(hero.reduction) + " 降低.";
            var defenseReductionAmountPretty = Math.round(hero.getDefenseReductionPercent(RogueRpg.Game.MonsterLevelForFloor(hero.floor)));
            text += Messages.newlines + "所造成的损失将会减少 " + green(hero.reduction) +
                " 减少的伤害，然后进一步减少。 " + green(defenseReductionAmountPretty) +
                "% 从防御对这个地板上的怪物。";
            // text += Messages.newlines + "Your armor has " + green(hero.equippedArmor ? hero.equippedArmor.defense : 0) + " defense and this is increased by " + green(hero.getDexterityDefenseMultiplierPercent()) + "% from dexterity.";
            return text;
        };
        Messages.strength = function (hero) {
            var text = "你有 " + green(hero.strength) + " strength.";
            text += Messages.newlines + "你的基本力量是 " + green(hero.baseStrength) + ". 增加了+" + green(hero.sumFromItems("flat_strength")) + " 然后进一步增加 " + green(hero.sumFromItems("percent_strength")) + "%.";
            text += Messages.newlines + "力量增加你的生命 " + green(percentIncr(hero.getStrengthLifeMult())) + "%";
            text += Messages.newlines + "力量增加你的伤害 " + green(hero.getStrengthMultiplierPercent()) + "%.";
            return text;
        };
        Messages.dexterity = function (hero) {
            var text = "你有 " + green(hero.dexterity) + " dexterity.";
            text += Messages.newlines + "你的基本灵敏是 " + green(hero.baseDexterity) + ". 增加了 +" + green(hero.sumFromItems("flat_dexterity")) + " a然后进一步增加。 " + green(hero.sumFromItems("percent_dexterity")) + "%.";
            text += Messages.newlines + "敏捷增加你的防御能力。" + green(hero.getDexterityDefenseMultiplierPercent()) + "%.";
            text += Messages.newlines + "敏捷可以减少你的攻击延迟。" + green(hero.getDexterityDelayMultiplierPercent()) + "%.";
            return text;
        };
        Messages.speed = function (hero) {
            var text = "你有 " + green(hero.speed) + " 的移动速度。";
            text += Messages.newlines + "你有 " + green(hero.baseSpeed) + " 基地的速度。这是增加了+" + green(hero.sumFromItems("flat_speed")) + " 然后进一步增加。 " + green(hero.sumFromItems("percent_speed")) + "%.";
            return text;
        };
        Messages.luck = function (hero) {
            var text = "你有 " + yellow(hero.luck) + " 运气.";
            text += Messages.newlines + "你有 " + yellow(hero.getLuckFlatPart()) + "基础运气，增加了。 " + yellow(percentIncr(hero.getLuckMultPart())) + "% 从技能.";
            text += Messages.newlines + "运气会增加物品掉落的质量，并影响你在你的范围内所造成的伤害。";
            if (RogueRpg.Level[RogueRpg.L.Dodge]) {
                text += Messages.newlines + "你会躲避 " + Math.round((RogueRpg.Helpers.hero.getLuckBasedChance(RogueRpg.Level[RogueRpg.L.Dodge] * 0.1) * 100)) + "% 的攻击.";
            }
            text += Messages.newlines + "例如，如果运气好，所有的卷都要做两次，最好的值被选中。." +
                "无论你的运气好坏，这个天平都很平稳。";
            return text;
        };
        Messages.death = function (hero, monster) {
            var text = "恭喜！你死了！ " + Messages.newlines;
            text += "你的 " + light_blue(RogueRpg.Helpers.game.shards) + " 碎片现在可以用来使你的文物更强壮，所以你可以在下一场比赛中进一步！" + Messages.newlines;
            text += "你到达了地板 " + green(hero.floor) + " 英雄级别是 " + green(hero.level) + " 在灭亡之前 " + green(monster.type) + ".";
            return text;
        };
        Messages.approxRatios = function (effect) {
            var text = "伤害:  ";
            if (effect.dpsRatio > 1) {
                text += green(Math.round(effect.dpsRatio * 100)) + "%";
            }
            else if (effect.dpsRatio < 1) {
                text += red(Math.round(effect.dpsRatio * 100)) + "%";
            }
            else {
                text += Math.round(effect.dpsRatio * 100) + "%";
            }
            text += "<br />防御: ";
            if (effect.defRatio > 1) {
                text += green(Math.round(effect.defRatio * 100)) + "%";
            }
            else if (effect.defRatio < 1) {
                text += red(Math.round(effect.defRatio * 100)) + "%";
            }
            else {
                text += Math.round(effect.defRatio * 100) + "%";
            }
            return text;
        };
        Messages.newlines = "<br/><br/>";
        return Messages;
    }());
    RogueRpg.Messages = Messages;
    var PermanentUpgrade = /** @class */ (function () {
        function PermanentUpgrade(id, name, description, gainPerRank, ranks, initialCost, costPerRank, resource, inAppPurchaseId) {
            if (inAppPurchaseId === void 0) { inAppPurchaseId = ''; }
            this.id = id;
            this.name = name;
            this.description = description;
            this.gainPerRank = gainPerRank;
            this.ranks = ranks;
            this.initialCost = initialCost;
            this.costPerRank = costPerRank;
            this.resource = resource;
            this.inAppPurchaseId = inAppPurchaseId;
        }
        PermanentUpgrade.Upgrades = {
            ShardBonus: new PermanentUpgrade("ShardBonus", "碎片的奖金", "获得 +{0}% 每次运行碎片 (+{1}% 每级)", 2, 50, 500, 1000, "gold"),
            AutoOpenChests: new PermanentUpgrade("AutoOpenChests", "自动打开箱子", "自动打开箱子范围", 1, 1, 30000, 1, "gold"),
            AutoItemSell: new PermanentUpgrade("AutoItemSell", "自动装备销售", "自动出售比您的装备差得多的物品", 1, 1, 200000, 1, "gold"),
            AutoItemPickup: new PermanentUpgrade("AutoItemPickup", "自动装备拾取", "自动拾取丢弃的装备", 1, 1, 500000, 1, "gold"),
            LessDamage: new PermanentUpgrade("LessDamage", "减少伤害", "减少受到的伤害 {0}%\n (+{1}% 每级)", 4, 15, 2, 2, "prisma"),
            StrongerWeapons: new PermanentUpgrade("StrongerWeapons", "更强的武器", "武器随着+{0}% 防御\n (+{1}% 每级)", 5, 20, 2, 1, "prisma"),
            StrongerArmors: new PermanentUpgrade("StrongerArmors", "坚固的盔甲", "	盔甲掉落 +{0}% 防御\n (+{1}% 每级)", 5, 20, 3, 1, "prisma"),
            Reduction: new PermanentUpgrade("Reduction", "防御", "增加 +{0} 防御    (+{1}每级)", 1, 20, 1, 1, "prisma"),
            AttackSpeed: new PermanentUpgrade("AttackSpeed", "攻击速度", "减少攻击延迟 {0}%  (-{1}% 每级)", 4, 10, 8, 4, "prisma"),
            SkillChoice: new PermanentUpgrade("SkillChoice", "技能选择", "增加 +{0} 选择技能时的选项(+{1} 每级)", 1, 1, 20, 80, "prisma"),
            BuyPrisma: new PermanentUpgrade("BuyPrisma", "买Prisma", "B买1 Prisma", 1, 1000, 250000, 5000, "gold"),
            UnlockAuto: new PermanentUpgrade("UnlockAuto", "解锁自动项目", "解锁箱子物品卖出和拾取", 1, 1, 500, 1, "iap", "unlock_auto_items"),
        };
        return PermanentUpgrade;
    }());
    RogueRpg.PermanentUpgrade = PermanentUpgrade;
    var Relic = /** @class */ (function () {
        // If price is present, the relic may be bought
        function Relic(name, statsPerLevel, maxRanks, price) {
            this.name = name;
            this.statsPerLevel = statsPerLevel;
            this.maxRanks = maxRanks;
            this.price = price;
            this.image = new Image();
        }
        Relic.getExpNeededAtLevel = function (level, relic) {
            if (level >= relic.maxRanks) {
                return 0;
            }
            if (relic.name === "Glittering Chalice") {
                if (level < 20) {
                    return 200 + level * 50;
                }
                else if (level < 35) {
                    return 1200 + 100 * (level - 20);
                }
                else {
                    return 2700 + 150 * (level - 35);
                }
            }
            if (relic.name === "Bloody Fang") {
                return 1500 + level * 500;
            }
            else if (relic.name === "Iron Spike") {
                return 2000 + level * 1000;
            }
            else if (relic.name === "Boots of Speed") {
                return 500 + level * 100;
            }
            else if (relic.name === "Defense Gem") {
                return 150 + level * 50;
            }
            else if (relic.name === "Keen Spectacles") {
                if (level < 50) {
                    return 50 + level * 25;
                }
                else if (level < 100) {
                    return 1050 + (level - 50) * 50;
                }
                else if (level < 150) {
                    return 3550 + (level - 100) * 100;
                }
                else {
                    return 8550 + (level - 150) * 150;
                }
            }
            else if (relic.name === "Brilliant Bauble") {
                if (level < 30) {
                    return 50 + level * 20;
                }
                else if (level < 45) {
                    return 450 + 50 * (level - 30);
                }
                else if (level < 60) {
                    return 1200 + 100 * (level - 45);
                }
                else if (level < 70) {
                    return 2700 + 400 * (level - 60);
                }
            }
            else if (relic.name === "Wicked Tome" && level < 40) {
                if (level < 7) {
                    return level * 10;
                }
                else if (level < 20) {
                    return 75 + (level - 7) * 5;
                }
                else if (level < 40) {
                    return 135 + (level - 20) * 10;
                }
            }
            else {
                if (level < 10) {
                    return level * 10;
                }
                else if (level < 20) {
                    return 50 + level * 10;
                }
                else if (level < 30) {
                    return 150 + level * 10;
                }
                else if (level < 40) {
                    return 250 + level * 10;
                }
                else if (level < 50) {
                    return 500 + level * 10;
                }
                else if (level < 70) {
                    return 1000 + (level - 50) * 25;
                }
                else {
                    return 1000 + 20 * 25 + (level - 70) * 50;
                }
            }
            // Default
            return;
        };
        Relic.FindRelicIndex = function (type) {
            for (var i = 0; i < Relic.Relics.length; i++) {
                if (Relic.Relics[i].name == type) {
                    return i;
                }
            }
            return -1;
        };
        Relic.prototype.load = function () {
            this.image.src = "images/" + this.name + ".png";
        };
        Relic.StatToName = function (input) {
            switch (input) {
                case "percent_life":
                    return "生命力";
                case "percent_dexterity":
                    return "敏捷";
                case "percent_strength":
                    return "力量";
                case "percent_defense":
                    return "防御";
                case "percent_damage":
                    return "伤害";
                case "flat_damage":
                    return "伤害";
                case "percent_cooldowns":
                    return "技能延迟";
                case "flat_luck":
                    return "运气";
                case "percent_speed":
                    return "速度";
                case "flat_speed":
                    return "速度";
                case "percent_experience":
                    return "经验";
                case "flat_defense":
                    return "防御";
                case "flat_lifesteal":
                    return "生命偷取";
            }
        };
        Relic.Relics = [
            new Relic("Wicked Tome", { "percent_life": 1 }, 200),
            new Relic("Shattered Blade", { "percent_damage": 1 }, 200),
            new Relic("Shielding Amulet", { "percent_defense": 1 }, 200),
            new Relic("Heroic Bracers", { "percent_strength": 1 }, 200),
            new Relic("Ceremonial Knife", { "percent_dexterity": 1 }, 200),
            new Relic("Glittering Chalice", { "flat_luck": 1 }, 50),
            new Relic("Brilliant Bauble", { "percent_cooldowns": 1 }, 70, 500),
            new Relic("Boots of Speed", { "flat_speed": 1 }, 10, 1000),
            new Relic("Defense Gem", { "flat_defense": 1 }, 20),
            new Relic("Keen Spectacles", { "percent_experience": 1 }, 200),
            new Relic("Bloody Fang", { "flat_lifesteal": 1 }, 5),
            new Relic("Iron Spike", { "flat_damage": 1 }, 9),
        ];
         Relic.Unlockables = [
            {
                difficulty: 0,
                relic: "Wicked Tome",
                relicIndex: 0,
                minFloor: 0,
                chance: 0.2
            },
            {
                difficulty: 1,
                relic: "Shattered Blade",
                relicIndex: 1,
                minFloor: 9,
                chance: 0.15
            },
            {
                difficulty: 1,
                relic: "Shielding Amulet",
                relicIndex: 2,
                minFloor: 18,
                chance: 0.1
            },
            {
                difficulty: 1,
                relic: "Heroic Bracers",
                relicIndex: 3,
                minFloor: 27,
                chance: 0.2
            },
            {
                difficulty: 1,
                relic: "Ceremonial Knife",
                relicIndex: 4,
                minFloor: 36,
                chance: 0.2
            },
            {
                difficulty: 1,
                relic: "Glittering Chalice",
                relicIndex: 5,
                minFloor: 46,
                chance: 0.2
            },
            {
                difficulty: 2,
                relic: "Defense Gem",
                relicIndex: 8,
                minFloor: 5,
                chance: 1
            },
            {
                difficulty: 2,
                relic: "Keen Spectacles",
                relicIndex: 9,
                minFloor: 11,
                chance: 1
            },
            {
                difficulty: 4,
                relic: "Bloody Fang",
                relicIndex: 10,
                minFloor: 5,
                chance: 1
            },
            {
                difficulty: 8,
                relic: "Iron Spike",
                relicIndex: 11,
                minFloor: 11,
                chance: 1
            }
        ];
        Relic.RelicLevels = [1, 0, 0, 0, 0, 0, 0, 0, 0, 0];
        Relic.RelicExperience = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0];
        return Relic;
    }());
    RogueRpg.Relic = Relic;
    var ItemEffect = /** @class */ (function () {
        function ItemEffect() {
        }
        return ItemEffect;
    }());
    RogueRpg.ItemEffect = ItemEffect;
    var AnimatedSprite = /** @class */ (function () {
        function AnimatedSprite(imagePaths, timePerFrame, loop, size) {
            if (loop === void 0) { loop = false; }
            this.imagePaths = imagePaths;
            this.timePerFrame = timePerFrame;
            this.loop = loop;
            this.size = size;
            this.images = [];
            for (var i = 0; i < imagePaths.length; i++) {
                var img = new Image();
                img.src = imagePaths[i];
                this.images.push(img);
            }
        }
        AnimatedSprite.prototype.getInstance = function (center, size) {
            return new AnimatedSpriteInstance(this, center);
        };
        return AnimatedSprite;
    }());
    RogueRpg.AnimatedSprite = AnimatedSprite;
    var AnimatedSpriteInstance = /** @class */ (function () {
        function AnimatedSpriteInstance(parent, drawCenter) {
            this.parent = parent;
            this.drawCenter = drawCenter;
            this.frameIndex = 0;
            this.countdownToSwitch = 0;
            this.countdownToSwitch = parent.timePerFrame;
        }
        AnimatedSpriteInstance.prototype.draw = function (context) {
            var image = this.parent.images[this.frameIndex];
            var width = this.parent.size || image.width;
            var height = this.parent.size || image.height;
            var offsetX = this.drawCenter.x - width / 2 - RogueRpg.Helpers.OriginX;
            var offsetY = this.drawCenter.y - height / 2 - RogueRpg.Helpers.OriginY;
            if (this.parent.size) {
                context.drawImage(image, offsetX, offsetY, width, height);
            }
            else {
                context.drawImage(image, offsetX, offsetY);
            }
        };
        // Returns true as long as this can be rendered
        AnimatedSpriteInstance.prototype.update = function (loopTime) {
            this.countdownToSwitch -= loopTime;
            if (this.countdownToSwitch <= 0) {
                this.frameIndex++;
                this.countdownToSwitch += this.parent.timePerFrame;
            }
            if (this.frameIndex >= this.parent.images.length) {
                return false;
            }
            return true;
        };
        return AnimatedSpriteInstance;
    }());
    RogueRpg.AnimatedSpriteInstance = AnimatedSpriteInstance;
    // Technically this is NOT an In App Purchase. This is an item that can be purchased with gemstones
    // or shards at any point throughout the game
    var InGameStoreItem = /** @class */ (function () {
        function InGameStoreItem(imageName, name, quantity, shardsPrice, gemstonePrice) {
            this.imageName = imageName;
            this.name = name;
            this.quantity = quantity;
            this.shardsPrice = shardsPrice;
            this.gemstonePrice = gemstonePrice;
        }
        return InGameStoreItem;
    }());
    RogueRpg.InGameStoreItem = InGameStoreItem;
    var MonsterImageNames = ["Goblin", "GoblinMove1", "GoblinMove2", "GoblinAttack1",
        "Torchling", "Fireling", "FirelingMove1", "FirelingMove2", "FirelingAttack",
        "Goo", "Kingu", "Goo Attack 1", "Goo Attack 2", "Goo Attack 3", "Goo Attack 4", "Goo Attack 5", "Goo Move 1",
        "Goo Move 2", "Goo Move 3", "Goo Move 4",
        "Flytrap", "Flytrapped", "Flytrap Attack",
        "Wee Demon", "Lava Demon",
        "Mud Clown", "Mud Clown Move 1", "Mud Clown Move 2", "Mud Blood", "Mud Clown Attack",
        "Mad Cube", "Sad Cube",
        "MadCubeMove1", "MadCubeMove2", "MadCubeAttack",
        "Bad Hero", "Super Bad Hero",
        "Ghost", "Roast Ghost",
        "GhostMove1", "GhostMove2", "GhostAttack",
        "Snowman", "Bowman", "Snowman Attack", "Snowman Move 1", "Snowman Move 2",
        "Scorpiar", "Triscorpiar", "ScorpiarMove1", "ScorpiarMove2", "ScorpiarAttack",
        "Thorny Cactus", "Sombrero Cactus", "Thorny Cactus Attack 1", "Thorny Cactus Attack 2",
        "Eye", "Maw", "EyeMove1", "EyeMove2", "EyeAttack",
        "Snake", "Lava Snake",
        "Flying Skull", "Diseased Skull", "Boarman", "Flying Skull Attack", "Flying Skull Move1", "Flying Skull Move2",
        "BatAttack", "BatMove1", "BatMove2", "BatMove3", "BatMove4",];
    var ImageNames = [
        // Hero images first
        "hero", "hero move0", "hero move1", "hero move2", "hero move3", "hero move4", "hero move5",
        "hero move6", "hero move7", "hero move8", "hero move9", "hero move10",
        "heroAttacking", "heroAttacking2", "heroAttacking3", "heroFaceAttacking", "heroFace",
        "hero stopping", "heroHappy", "heroSad", "heroLookingLeft", "heroLookingRight",
        "Headband", "Bowman Cap", "Wizard Hat", "inventoryButton",
        // Item images second
        "Sword", "Mace", "Spear", "Shield", "Protector", "Bulwark", "Leather Vest", "Dark Armor",
        "Breastplate", "Chainmail", "Longsword", "Aegis", "Tunic",
        "Scumbasher", "Scumbasher +1", "Scumbasher +2",
        "Wraps", "Blade of the Dreadking", "Greataxe",
        "Hatchet", "Halberd", "Morning Star", "Scimitar", "Decapitator", "Kukri", "Heavy Shield",
        "Ice Sword", "Fire Sword", "Greatsword", "Kingsword", "Crystal Shard", "Round Shield", "Ward", "Aspis", "Holy Shield",
        "Splint Mail", "Red Robes", "Battle Robe",
        "Tiger Claws", "Katar", "Gladius", "Kama", "Knife", "Dagger", "Katana", "Lance",
        "Arrow", "Simple Bow", "Bow", "Simple Bow", "Recurve Bow", "Golden Bow", "Sylvan Bow",
        // Floors, walls
        "Mine Floor", "Mine Floor 2", "Mine Wall", "Mine Wall 2", "Mine Wall Pickaxe", "Mine Wall Torch",
        "MineAlt Floor", "MineAlt Wall", "MineAlt Wall 2",
        "Orc Cave Floor", "Orc Cave Wall",
        "DesertSand",
        "Dungeon Floor", "Dungeon Wall", "Small Floor", "Small Wall", "Yellow Brick Floor", "Yellow Brick Wall",
        "Magma Floor", "Magma Wall", "Earth Floor", "Earth Wall",
        "Cave Floor", "Cave Wall", "Small Mud Puddle",
        "Ice Tunnel Floor", "Ice Tunnel Wall", "Cosmic Floor", "Cosmic Wall",
        "heart", "gemstones", "shards", "Prisma", "Gold", "30bonus", "shardEnergy",
        "rarity6glow", "Smash", "Portal", "Portal Empty", "Potion Slot",
        // World Objects
        "Item Upgrade Shrine", "Used Item Upgrade Shrine", "Healing Shrine", "Used Healing Shrine", "Treasure Chest",
        "Used Treasure Chest", "Level Exit", "Dark Treasure Chest", "Used Dark Treasure Chest",
        "Bowman Cap Item", "Wizard Hat Item",
        "Empty Mine Cart", "Filled Mine Cart",
        "Prisma Found",
        "Wicked Tome World", "Shattered Blade World", "Glittering Chalice World", "Shielding Amulet World", "Ceremonial Knife World",
        "Heroic Bracers World", "Pedestal", "Keen Spectacles World", "Bloody Fang World", "Defense Gem World",
        "Iron Spike World",
        "Sitting Wizard", "Frozen Monk", "Unfrozen Monk", "Flying Kick", "Flying Kick Reverse", "Ghost Girl", "Alchemy Table",
        "LightningWhite", "LightningPurple",
        "sparkle", "PurpleCloud", "TinySparkle",
        "Stab", "Sparks1", "Sparks2", "Sparks3",
        "Puppy Sit 1", "Puppy Sit 2", "Puppy Sit 3",
        // Effects, etc
        "Starburst", "Orb", "hero fly1",
        // Inventory images last
        "PotionBlue", "PotionRed", "PotionGreen", "PotionYellow", "PotionPurple", "PotionOrange", "PotionGrey", "PotionWhite", "Pointer",
        "PotionDamage", "PotionDefense", "PotionReduction", "PotionRainbow",
        "GoldRing", "GoldRing2", "GoldRing3", "GoldRing4", "GoldRing5",
        "SilverRing", "SilverRing2", "SilverRing3", "SilverRing4", "SilverRing5",
    ].concat(MonsterImageNames);
    var ImageSprites = {
        Smoke: new AnimatedSprite(["images/smoke0.png",
            "images/smoke1.png",
            "images/smoke2.png"], 100, false),
        Sweep: new AnimatedSprite(["images/sweep0.png",
            "images/sweep1.png",
            "images/sweep2.png",
            "images/sweep3.png",
            "images/sweep4.png",
            "images/sweep5.png",], 50, false),
        Snap: new AnimatedSprite(["images/snap0.png",
            "images/snap1.png",
            "images/snap2.png",
            "images/snap3.png",
            "images/snap4.png",
            "images/snap5.png",], 50, false),
        Slash: new AnimatedSprite(["images/slash0.png",
            "images/slash1.png",
            "images/slash2.png",
            "images/slash3.png",], 90, false),
        Advance: new AnimatedSprite(["images/advance0.png",
            "images/advance1.png",
            "images/advance2.png",
            "images/advance3.png",
            "images/advance4.png",
            "images/advance5.png",
            "images/advance6.png",
            "images/advance7.png"], 40, false),
        Explosion: new AnimatedSprite(["images/explosion0.png",
            "images/explosion1.png",
            "images/explosion2.png"], 200, false, 400),
        Slow: new AnimatedSprite(["images/Slow0.png",
            "images/Slow1.png",
            "images/Slow2.png",], 150, false, 30),
        Smash: new AnimatedSprite(["images/Smash2.png",
            "images/Smash3.png",
            "images/Smash4.png",], 80, false),
    };
    var Images = /** @class */ (function () {
        function Images() {
        }
        Images.has = function (name) {
            var index = Images._names.indexOf(name);
            if (index >= 0) {
                if (Images._images[index].complete) {
                    return true;
                }
            }
            return false;
        };
        Images.setImage = function (name, img) {
            Images._names.push(name);
            Images._images.push(img);
        };
        Images.get = function (name) {
            var index = Images._names.indexOf(name);
            if (index > -1) {
                if (Images._images[index].complete) {
                    return Images._images[index];
                }
            }
            index = Images._svgs.indexOf(name);
            if (index > -1) {
                if (Images._svgImages[index].complete) {
                    return Images._svgImages[index];
                }
            }
            return Images.placeholder; // Let the game start before all assets are loaded
        };
        Images.load = function () {
            Images.placeholder = new Image();
            Images.placeholder.src = "data:image/gif;base64,R0lGODlhAQABAIAAAAUEBAAAACwAAAAAAQABAAACAkQBADs=";
            for (var i = 0; i < Images._names.length; i++) {
                var image = new Image();
                image.src = "images/" + Images._names[i] + ".png";
                this._images.push(image);
            }
            for (var i = 0; i < Images._svgs.length; i++) {
                var image = new Image();
                image.src = "images/" + Images._svgs[i] + ".svg";
                this._svgImages.push(image);
            }
            for (var i = 0; i < Relic.Relics.length; i++) {
                Relic.Relics[i].load();
            }
            // Preload mirrored and damage enemies to avoid flicker
            for (var i = 0; i < MonsterImageNames.length; i++) {
                // TODO: Load Morped, Mirrored and 2M images
            }
        };
        Images._svgs = [];
        Images._images = [];
        Images._svgImages = [];
        Images.placeholder = new Image();
        Images._names = ImageNames;
        Images.Sprites = ImageSprites;
        return Images;
    }());
    RogueRpg.Images = Images;
    var Projectile = /** @class */ (function () {
        function Projectile(source, imageName, x, y, xVelocity, yVelocity, damage, zOffset) {
            if (zOffset === void 0) { zOffset = 0; }
            this.source = source;
            this.imageName = imageName;
            this.x = x;
            this.y = y;
            this.xVelocity = xVelocity;
            this.yVelocity = yVelocity;
            this.damage = damage;
            this.zOffset = zOffset;
            this.direction = 0;
            // Marked after the projectile is done, for a final rendering frame
            this.dead = false;
            this.direction = Math.atan2(this.yVelocity, this.xVelocity);
        }
        return Projectile;
    }());
    RogueRpg.Projectile = Projectile;
    var MovingShard = /** @class */ (function () {
        function MovingShard(location, target, value) {
            this.location = location;
            this.target = target;
            this.value = value;
            // Give a random direction
            this.dir = Math.random() * Math.PI * 2;
            this.velocity = 4.2;
            this.attractionVelocity = 7;
        }
        // Move the shard on its path, apply the target's pull
        MovingShard.Update = function (shard) {
            shard.location.x += Math.cos(shard.dir) * shard.velocity;
            shard.location.y += Math.sin(shard.dir) * shard.velocity;
            shard.velocity = shard.velocity * 0.99;
            shard.attractionVelocity += 0.10; // Accelerate towards target
            var dirToLog = Math.atan2(shard.target.y - shard.location.y, shard.target.x - shard.location.x);
            shard.location.x += Math.cos(dirToLog) * shard.attractionVelocity;
            shard.location.y += Math.sin(dirToLog) * shard.attractionVelocity;
        };
        return MovingShard;
    }());
    RogueRpg.MovingShard = MovingShard;
    var InAppPurchase = /** @class */ (function () {
        // The real price will be queried
        function InAppPurchase(name, baseGems, mockPriceUsd, image, id) {
            this.name = name;
            this.baseGems = baseGems;
            this.mockPriceUsd = mockPriceUsd;
            this.image = image;
            this.id = id;
        }
        InAppPurchase.Purchases = [
            // new InAppPurchase("Get 10", 10, 0, "images/gemstones.png"), // Base price
            new InAppPurchase("Get 200", 200, 1.99, "images/gemstones1.png", "currency_gemstones_200"),
            new InAppPurchase("Get 600", 600, 4.99, "images/gemstones2.png", "currency_gemstones_500"),
            new InAppPurchase("Get 1,250", 1250, 9.99, "images/gemstones3.png", "currency_gemstones_1000"),
            new InAppPurchase("Get 3,500", 3500, 24.99, "images/gemstones4.png", "currency_gemstones_2500"),
        ];
        return InAppPurchase;
    }());
    RogueRpg.InAppPurchase = InAppPurchase;
    var MultiplierModes = /** @class */ (function () {
        function MultiplierModes() {
        }
        MultiplierModes.Modes = [
            1,
            2,
            4,
            8,
            16,
            50,
            100,
            200
        ];
        return MultiplierModes;
    }());
    RogueRpg.MultiplierModes = MultiplierModes;
    var HeroClass = /** @class */ (function () {
        function HeroClass(name, skill, bonus, amount, description, hat, isRanged) {
            if (hat === void 0) { hat = ""; }
            if (isRanged === void 0) { isRanged = false; }
            this.name = name;
            this.skill = skill;
            this.bonus = bonus;
            this.amount = amount;
            this.description = description;
            this.hat = hat;
            this.isRanged = isRanged;
        }
        HeroClass.Classes = [
            new HeroClass("Warrior", "sweep", "flat_reduction", 1, "技能：对附近所有敌人造成1点伤害，延迟：5秒。 \r\n\r\n 加成: +1 免伤"),
            // new HeroClass("Rogue", "Smash", "flat_damage", 1, "其特殊攻击变得非常强大。 \r\n\r\n Bonus: +1 Damage", "Bowman Cap"),
            new HeroClass("Rogue", "Stab", "flat_damage", 1, "技能：对2个附近敌人以2倍速度附加攻击1秒，延迟：8秒。 \r\n\r\n 加成: +1 伤害", "Bowman Cap"),
            new HeroClass("Mage", "snap", "flat_speed", 4, "技能:传送到一个新的地点，在开始和结束计算，延迟:6s。 \r\n\r\n 加成: +4 速度", "Wizard Hat"),
            new HeroClass("Monk", "Flying Kick", "percent_luck", 10, "技能：向远方的敌人飞行，造成1点伤害，延迟：0.5s。 \r\n\r\n 加成: + 10%的运气", "Headband"),
        ];
        return HeroClass;
    }());
    RogueRpg.HeroClass = HeroClass;
    var Rgba = /** @class */ (function () {
        function Rgba(r, g, b, a) {
            if (a === void 0) { a = 1; }
            this.r = r;
            this.g = g;
            this.b = b;
            this.a = a;
            this.r = Math.floor(r);
            this.g = Math.floor(g);
            this.b = Math.floor(b);
        }
        Rgba.clone = function (rgba, newOpacity) {
            if (typeof (newOpacity) == "number") {
                return new Rgba(rgba.r, rgba.g, rgba.b, newOpacity);
            }
            else {
                return new Rgba(rgba.r, rgba.g, rgba.b, rgba.a);
            }
        };
        Rgba.stringify = function (rgba) {
            if (rgba.a < 1) {
                return "rgba(" + rgba.r + ", " + rgba.g + ", " + rgba.b + ", " + rgba.a + ")";
            }
            else {
                return "rgb(" + rgba.r + ", " + rgba.g + ", " + rgba.b + ")";
            }
        };
        return Rgba;
    }());
    RogueRpg.Rgba = Rgba;
})(RogueRpg || (RogueRpg = {}));
//# sourceMappingURL=GameExtras.js.map