var RogueRpg;
(function (RogueRpg) {
    var Helpers = /** @class */ (function () {
        function Helpers() {
        }
        Object.defineProperty(Helpers, "TileSize", {
            get: function () {
                return (Helpers.game.map && Helpers.game.map.tileSize) || 80;
            },
            enumerable: true,
            configurable: true
        });
        Helpers.formatNumber = function (num, doNotMinimize) {
            if (num < 1000000 || doNotMinimize) {
                return num.toLocaleString();
            }
            else if (num < 1000000000) {
                return Math.floor(num / 1000) / 1000 + "M";
            }
            return num.toLocaleString();
        };
        // Accepts an image and returns a canvas with the morphed image drawn in it (draw that canvas to your canvas)
        Helpers.MorphImage = function (imageName, w, h) {
            if (!RogueRpg.Images.has(imageName)) {
                return RogueRpg.Images.placeholder;
            }
            if (RogueRpg.Images.has(imageName + Helpers.MORPH_SUFFIX)) {
                return RogueRpg.Images.get(imageName + Helpers.MORPH_SUFFIX);
            }
            var image = RogueRpg.Images.get(imageName);
            var canvas = document.createElement("canvas");
            canvas.width = w;
            canvas.height = h;
            var placeholderContext = canvas.getContext("2d");
            placeholderContext.clearRect(0, 0, w, h);
            placeholderContext.drawImage(image, 0, 0, w, h);
            var imgData = placeholderContext.getImageData(0, 0, w, h);
            Helpers.pinkImageFilter(imgData);
            placeholderContext.putImageData(imgData, 0, 0);
            var img = document.createElement('img');
            img.src = canvas.toDataURL();
            if (img.width > 1) {
                RogueRpg.Images.setImage(imageName + Helpers.MORPH_SUFFIX, img);
            }
            return canvas;
        };
        Helpers.MirrorImage = function (imageName, center, drawWidth, drawHeight) {
            if (!RogueRpg.Images.has(imageName)) {
                return RogueRpg.Images.placeholder;
            }
            if (RogueRpg.Images.has(imageName + Helpers.MIRROR_SUFFIX)) {
                return RogueRpg.Images.get(imageName + Helpers.MIRROR_SUFFIX);
            }
            // Otherwise, get the image, draw it mirrored and save it
            var image = RogueRpg.Images.get(imageName);
            var w = drawWidth; // image.width; // Need extra space to rotate around
            var h = drawHeight; // image.height;
            var canvas = document.createElement("canvas");
            canvas.width = w * 2 - center;
            canvas.height = h;
            var context = canvas.getContext("2d");
            context.translate(center, 0);
            context.scale(-1, 1);
            context.translate(-(center), 0);
            context.clearRect(0, 0, w * 2 - center, h);
            context.drawImage(image, 0, 0, w, h);
            var img = document.createElement('img');
            img.src = canvas.toDataURL();
            if (img.width > 1) {
                RogueRpg.Images.setImage(imageName + Helpers.MIRROR_SUFFIX, img);
            }
            return canvas;
        };
        Helpers.pinkImageFilter = function (pixels) {
            var d = pixels.data;
            for (var i = 0; i < d.length; i += 4) {
                if (d[i + 3] !== 0) {
                    d[i] = 255;
                    d[i + 1] = Math.floor((d[i + 1] + 200) / 2);
                    d[i + 2] = Math.floor((d[i + 2] + 200) / 2);
                }
            }
        };
        Helpers.whiteImageFilter = function (pixels) {
            var d = pixels.data;
            for (var i = 0; i < d.length; i += 4) {
                if (d[i + 3] !== 0) {
                    d[i] = 255;
                    d[i + 1] = 255;
                    d[i + 2] = 255;
                }
            }
        };
        Helpers.SelectOne = function (array) {
            var val = Math.floor(Math.random() * array.length);
            return array[val];
        };
        Object.defineProperty(Helpers, "HeroDrawX", {
            // Returns the X-offset to draw the hero at (he's centered)
            get: function () {
                return Helpers.CenterX - Helpers.HeroWidth / 2;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Helpers, "HeroDrawY", {
            // Returns the Y-Offset to draw the hero character at
            get: function () {
                return Helpers.CenterY - Helpers.HeroHeight;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Helpers, "CenterX", {
            get: function () {
                return Helpers.renderer.ww / 2;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Helpers, "CenterY", {
            get: function () {
                return Helpers.renderer.wh / 2;
            },
            enumerable: true,
            configurable: true
        });
        Helpers.distance = function (x1, x2, y1, y2) {
            return Math.sqrt(Math.pow(x1 - x2, 2) + Math.pow(y1 - y2, 2));
        };
        Helpers.angle = function (rotateThis, towardsThis) {
            if (!rotateThis.y) {
                rotateThis.y = rotateThis.drawY + Helpers.OriginY + rotateThis.drawHeight / 2;
            }
            if (!rotateThis.x) {
                rotateThis.x = rotateThis.drawX + Helpers.OriginX + rotateThis.drawWidth / 2;
            }
            if (!towardsThis.y) {
                towardsThis.y = towardsThis.drawY + Helpers.OriginY + towardsThis.drawHeight / 2;
            }
            if (!towardsThis.x) {
                towardsThis.x = towardsThis.drawX + Helpers.OriginX + towardsThis.drawWidth / 2;
            }
            return Math.atan2(towardsThis.y - rotateThis.y, towardsThis.x - rotateThis.x);
        };
        Helpers.dist = function (a, b) {
            return Math.sqrt(Math.pow(a.x - b.x, 2) + Math.pow(a.y - b.y, 2));
        };
        Helpers.IsWalkable = function (location, map, avoidNonUpWallsBy) {
            if (avoidNonUpWallsBy === void 0) { avoidNonUpWallsBy = 0; }
            var tileX = Math.floor(location.x / map.tileSize);
            var tileY = Math.floor(location.y / map.tileSize);
            var tile = map.tiles[tileY] && map.tiles[tileY][tileX];
            if (tile) {
                var remainderX = location.x % map.tileSize;
                var remainderY = location.y % map.tileSize;
                if (remainderX < avoidNonUpWallsBy || remainderX > map.tileSize - avoidNonUpWallsBy || remainderY > map.tileSize - avoidNonUpWallsBy) {
                    return false;
                }
                return tile.walkable;
            }
            return false;
        };
        Object.defineProperty(Helpers, "OriginX", {
            // Returns the game-coordinates of the (0, 0) canvas point
            get: function () {
                return Helpers.hero.x - Helpers.CenterX;
            },
            enumerable: true,
            configurable: true
        });
        Object.defineProperty(Helpers, "OriginY", {
            get: function () {
                return Helpers.hero.y - Helpers.CenterY;
            },
            enumerable: true,
            configurable: true
        });
        Helpers.ToTileLocation = function (x, y) {
            return new RogueRpg.Location(Math.floor(x / Helpers.TileSize), Math.floor(y / Helpers.TileSize));
        };
        Helpers.HeroHeight = 80;
        Helpers.HeroWidth = 40;
        Helpers.TileOverlap = 50;
        Helpers.LoopTime = 24; // 25ms is 40fps
        Helpers.MaxItemCount = 25;
        Helpers.RarityColors = ["rgb(0, 0, 0)", "rgb(255, 255, 255)", "rgb(90, 110, 255)", "rgb(240, 240, 0)", "rgb(220, 150, 50)", "rgb(50, 255, 20)", "rgb(110, 110, 110)", "rgb(255, 0, 255)"];
        Helpers.DefaultAgroRadius = 280;
        Helpers.DefaultAssistRadius = 170;
        Helpers.DefaultFollowRadius = 700;
        Helpers.MORPH_SUFFIX = "_MORPHED_";
        Helpers.MIRROR_SUFFIX = "(MIRRORED)";
        return Helpers;
    }());
    RogueRpg.Helpers = Helpers;
})(RogueRpg || (RogueRpg = {}));
//# sourceMappingURL=Helpers.js.map