/// <reference path="game.ts" />
// For an introduction to the Blank template, see the following documentation:
// http://go.microsoft.com/fwlink/?LinkID=397705
// To debug code on page load in Ripple or on Android devices/emulators: launch your app, set breakpoints, 
// and then run "window.location.reload()" in the JavaScript Console.
var RogueRpg;
(function (RogueRpg) {
    "use strict";
    // TODO: It seemed like later levels needed more monsters
    var MapGenerator = /** @class */ (function () {
        function MapGenerator() {
            this.exitPlaced = false;
            this.enemyCount = 0;
            this.currentLevelNumber = 1;
        }
        MapGenerator.prototype.getFloorTile = function () {
            if (this.tileType == "Mine" && Math.random() > 0.5) {
                return this.tileType + " Floor 2";
            }
            return this.tileType + " Floor";
        };
        MapGenerator.prototype.getWallTile = function () {
            if (this.tileType == "Mine" && Math.random() > 0.5) {
                return this.tileType + " Wall 2";
            }
            return this.tileType + " Wall";
        };
        MapGenerator.prototype.makeFloor = function (y, x) {
            this._mapInProgress.tiles[y][x].type = this.getFloorTile();
            this._mapInProgress.tiles[y][x].walkable = true;
        };
        MapGenerator.prototype.createLevel = function (levelNumber) {
            this.currentLevelNumber = levelNumber;
            MapGenerator.addedRelicThisStage = false;
            MapGenerator.addedClassThisStage = false;
            MapGenerator.itemsForTiles = {};
            if (Math.random() > 10) {
                this.createPotionFixMap();
            }
            else if (Math.random() > 10) {
                this.createUpRoomMap();
            }
            else if (ENABLE_ART_REVIEW) {
                this.createArtReviewMap(); // test testmonster 
            }
            else if (levelNumber % RogueRpg.Game.NumLevels == 1) {
                this.createMineHallwaysMap("Mine");
            }
            else if (levelNumber % RogueRpg.Game.NumLevels == 2) {
                this.createLinkedRoomsMapOrcCave("Orc Cave");
            }
            else if (levelNumber % RogueRpg.Game.NumLevels == 3) {
                this.createLabyrinthMap("Cave");
                // this.addFloorDecor([{ name: "Small Mud Puddle", size: 130 }], 15);
            }
            else if (levelNumber % RogueRpg.Game.NumLevels == 4) {
                this.createRiverTunnelsMap("Ice Tunnel");
            }
            else if (levelNumber % RogueRpg.Game.NumLevels == 5) {
                this.createOutdoorDesertMap();
            }
            else if (levelNumber % RogueRpg.Game.NumLevels == 6) {
                this.createUndergroundDesertLevel();
            }
            else if (levelNumber % RogueRpg.Game.NumLevels == 7) {
                this.createUndergroundDesertLevelAlternate();
            }
            else {
                //if (Math.random() < 0.3) {
                //    this.createUpRoomMap();
                //} else
                if (Math.random() < 0.3) {
                    this.createHallwaysMap();
                }
                else if (Math.random() < 0.7) {
                    if (Math.random() > 0.5) {
                        this.createUndergroundDesertLevel(MapGenerator.getRandomTilesetName());
                    }
                    else {
                        this.createUndergroundDesertLevelAlternate(MapGenerator.getRandomTilesetName());
                    }
                }
                else {
                    this.createLabyrinthMap();
                }
            }
            this.addSomeGold();
            // Helpers.game.spawnAllStorylineNotes(this._mapInProgress);
            this.computeLevelBounds();
            // There appears to be a tiny chance that the exit is not properly placed, but I don't know the cause
            var map = this._mapInProgress;
            if (!this.exitPlaced) {
                this.placeExit(map.heroStartX / map.tileSize, map.heroStartY / map.tileSize);
            }
            else {
                // TODO: Figure out if the exit is not walkable
                var exit = this.findExit();
                if (!RogueRpg.Helpers.IsWalkable(new RogueRpg.Location(exit.drawX + 0.5 * exit.drawWidth, exit.drawY + 0.5 * exit.drawWidth), map)) {
                    var index = map.specialWorldObjects.indexOf(exit);
                    map.specialWorldObjects.splice(index, 1);
                    this.placeExit(map.heroStartX / map.tileSize, map.heroStartY / map.tileSize);
                }
            }
            return this._mapInProgress;
        };
        MapGenerator.prototype.addSomeGold = function () {
            var count = randInt(0, 3);
            for (var i = 0; i < count; i++) {
                var location = RogueRpg.Helpers.game.getRandomWalkableTilePosition(this._mapInProgress);
                var amount = Math.floor(randInt(1, 4) * (2 * this.currentLevelNumber) + randInt(1, 10)) * RogueRpg.Helpers.game.hero.state.multiplier;
                this._mapInProgress.goldOnGround.push({ x: location.x, y: location.y, amount: amount, v: 0 });
            }
        };
        MapGenerator.prototype.createPotionFixMap = function () {
            this.setupMap(40, ["Ice Tunnel"], "Ice Tunnel");
            var x = 18;
            var y = 20;
            this._mapInProgress.heroStartX = this._mapInProgress.tileSize * x;
            this._mapInProgress.heroStartY = this._mapInProgress.tileSize * y;
            var room = {
                left: x - 1,
                top: y - 2,
                right: x + 1,
                bottom: y + 2
            };
            this.drawRoomToMap(room);
            var room = {
                left: x - 2,
                top: y - 1,
                right: x + 2,
                bottom: y + 1
            };
            this.drawRoomToMap(room);
            this.placeExitInRoom(room);
            var dp = new RogueRpg.DroppedPotion(RogueRpg.Game.randomPotion(), new RogueRpg.Location(RogueRpg.Helpers.TileSize * (room.right + room.left) / 2, RogueRpg.Helpers.TileSize * (room.top + room.bottom) / 2));
            this._mapInProgress.potionsOnGround.push(dp);
        };
        MapGenerator.prototype.findExit = function () {
            for (var i = 0; i < this._mapInProgress.specialWorldObjects.length; i++) {
                var obj = this._mapInProgress.specialWorldObjects[i];
                if (obj.type === "Level Exit") {
                    return obj;
                }
            }
        };
        MapGenerator.prototype.createArtReviewMap = function () {
            this.setupMap(40, ["Cave"], "Cave");
            var room = {
                left: 2,
                top: 2,
                right: 30,
                bottom: 30
            };
            this.drawRoomToMap(room);
            this.placeExitInRoom(room);
            for (var i = 0; i < RogueRpg.Game.MonstersByFloor.length; i++) {
                this.addWorldObjectToTile(new RogueRpg.Location(i + 2, 4), "TestMonster", RogueRpg.Game.MonstersByFloor[i]);
            }
            this._mapInProgress.heroStartX = 750;
            this._mapInProgress.heroStartY = 750;
        };
        MapGenerator.prototype.createSpecialLevel = function (tileType, floor) {
            MapGenerator.itemsForTiles = {};
            this.createPlaneMap(tileType);
            this.computeLevelBounds();
            return this._mapInProgress;
        };
        MapGenerator.prototype.computeLevelBounds = function () {
            var map = this._mapInProgress;
            for (var i = 0; i < map.mapHeight; i++) {
                for (var j = 0; j < map.mapWidth; j++) {
                    if (map.tiles[i][j].walkable) {
                        if (j < map.leftMostRoom) {
                            map.leftMostRoom = j;
                        }
                        if (j > map.rightMostRoom) {
                            map.rightMostRoom = j;
                        }
                        if (i < map.topMostRoom) {
                            map.topMostRoom = i;
                        }
                        if (i > map.bottomMostRoom) {
                            map.bottomMostRoom = i;
                        }
                    }
                }
            }
        };
        MapGenerator.getRandomTilesetName = function () {
            return RogueRpg.Helpers.SelectOne(MapGenerator.AvailableTilesets);
        };
        MapGenerator.prototype.createLabyrinthMap = function (tileset) {
            var tilesets = ["Ice Tunnel", "Small", "Yellow Brick", "Cave"];
            this.setupMap(50, tilesets, tileset);
            var origin = new RogueRpg.Location(25, 25);
            var numBranches = this.branchOutFrom(origin, "none", 1);
            // Note: The branchOutFrom function should place an exit, but I think there's a chance it doesn't
            this._mapInProgress.heroStartX = 25.5 * this._mapInProgress.tileSize;
            this._mapInProgress.heroStartY = 25.5 * this._mapInProgress.tileSize;
            for (var i = 0; i < this.enemyCount + numBranches * 0.5; i++) {
                RogueRpg.Helpers.game.spawnMonster(this._mapInProgress);
            }
        };
        MapGenerator.prototype.createOutdoorDesertMap = function () {
            this.enemyCount = 35;
            this._mapInProgress = new RogueRpg.Map();
            this._mapInProgress.tileSize = 200;
            this._mapInProgress.mapHeight = 8;
            this._mapInProgress.mapWidth = 12;
            this._mapInProgress.heroStartX = 150;
            this._mapInProgress.heroStartY = 750;
            for (var i = 0; i < this._mapInProgress.mapHeight; i++) {
                var tileRow = [];
                for (var j = 0; j < this._mapInProgress.mapWidth; j++) {
                    var tile = new RogueRpg.Tile();
                    tile.type = "DesertSand";
                    tile.walkable = true;
                    tileRow.push(tile);
                }
                this._mapInProgress.tiles.push(tileRow);
            }
            this.placeExit(randIntMap(10, 11), randIntMap(2, 7));
            if (Math.random() > 0.5) {
                this.spawnRandomPotionAt(randIntMap(2, 9), randIntMap(2, 7));
            }
            for (var i = 0; i < this.enemyCount; i++) {
                RogueRpg.Helpers.game.spawnMonster(this._mapInProgress);
            }
            if (Math.random() > 0.5) {
                this.addWorldObjectToTile(new RogueRpg.Location(randIntMap(2, this._mapInProgress.mapWidth - 1), randIntMap(2, this._mapInProgress.mapHeight - 1)), "Item Upgrade Shrine");
            }
            else {
                this.addWorldObjectToTile(new RogueRpg.Location(randIntMap(2, this._mapInProgress.mapWidth - 1), randIntMap(2, this._mapInProgress.mapHeight - 1)), "Healing Shrine");
            }
        };
        // Creates a map that goes mostly upwards, consisting of closely linked rooms
        // TODO: Note that rooms are not quite the right construct here - too much loot density
        MapGenerator.prototype.createUpRoomMap = function (tileset) {
            // TODO: Figure out a better way of managing tilesets
            var tilesets = ["Dungeon", "Small", "Yellow Brick", "Cave"];
            this.setupMap(60, tilesets, tileset);
            var start = new RogueRpg.Location(30, 55);
            var startingRoom = this.roomAt(start.y - 1, start.x - 1, start.y + 2, start.x + 2);
            // Unitary operation
            // Place a room up/right and up/left
            var leftRoom = this.addConnectedRoomAbove(start.x - 2, start.y - 4, startingRoom, MapGenerator.UpRoomOptions);
            var rightRoom = this.addConnectedRoomAbove(start.x + 2, start.y - 4, startingRoom, MapGenerator.UpRoomOptions);
            var rooms = [];
            this.createUpRoomRecursive(leftRoom, rooms);
            this.createUpRoomRecursive(rightRoom, rooms);
            // Algorithm: As a binary tree, add rooms. If a room ends up on the wrong side of half, skip it.
            // If a room is too close to the edges, skip it
            // Push a list of all rooms into an array
            // Place the exit in one of the lowest-y-value room,
            // If depth > N or the room is too close to 0, skip it
            // TODO: Place the exit in a room from rooms
            var highestRoom = startingRoom;
            var highestRoomY = 55;
            for (var i = 0; i < rooms.length; i++) {
                var y = FindRoomLocation(rooms[i]).y;
                if (y < highestRoomY) {
                    highestRoomY = y;
                    highestRoom = rooms[i];
                }
            }
            var exitRoomPos = FindRoomLocation(highestRoom);
            this.placeExit(exitRoomPos.x, exitRoomPos.y - 1);
            this._mapInProgress.heroStartX = (start.x + 0.5) * this._mapInProgress.tileSize;
            this._mapInProgress.heroStartY = (start.y + 0.5) * this._mapInProgress.tileSize;
        };
        MapGenerator.prototype.createUpRoomRecursive = function (from, list) {
            var roll = randIntMap(1, 3);
            var start = FindRoomLocation(from);
            if (start.y < 30) {
                return;
            }
            if (start.x <= 10 || start.x >= this._mapInProgress.mapWidth - 10) {
                return;
            }
            if (start.y < 40) {
                var narrowing = true;
            }
            if (roll == 1 || roll == 3) {
                if (!narrowing || start.x > this._mapInProgress.mapWidth / 2 - 3) {
                    var leftRoom = this.addConnectedRoomAbove(start.x - randIntMap(1, 2), start.y - randIntMap(4, 5), from, narrowing ? MapGenerator.UpRoomOptionsNarrowing : MapGenerator.UpRoomOptions);
                    list.push(leftRoom);
                    this.createUpRoomRecursive(leftRoom, list);
                }
            }
            if (roll == 1 || roll == 2) {
                if (!narrowing || start.x < this._mapInProgress.mapWidth / 2 + 3) {
                    var rightRoom = this.addConnectedRoomAbove(start.x + randIntMap(1, 2), start.y - randIntMap(4, 5), from, narrowing ? MapGenerator.UpRoomOptionsNarrowing : MapGenerator.UpRoomOptions);
                    list.push(rightRoom);
                    this.createUpRoomRecursive(rightRoom, list);
                }
            }
        };
        // Add a room centered(ish) around the numbers given, with hallways as needed to reach it
        MapGenerator.prototype.addConnectedRoomAbove = function (x, y, from, roomOptions) {
            var width = randIntMap(roomOptions.minWidth, roomOptions.maxWidth);
            var height = randIntMap(roomOptions.minHeight, roomOptions.maxHeight);
            var newRoom = this.roomAt(y - Math.floor(height / 2), x - Math.floor(width / 2), y + Math.ceil(height / 2), x + Math.ceil(width / 2));
            if (newRoom.right < from.right) {
                var hallX = from.left;
            }
            else {
                var hallX = from.right - 1; // Right is apparently non-inclusive
            }
            var hall = this.verticalHallway(hallX, newRoom.bottom - 1, 2 + (from.top - newRoom.bottom));
            return newRoom;
        };
        MapGenerator.prototype.createMineHallwaysMap = function (tileset) {
            var tilesets = ["Dungeon", "Small", "Yellow Brick", "Cave", "Cosmic"];
            this.setupMap(50, tilesets, tileset);
            var middleHorizontal = this.horizontalHallway(4, 20, 23);
            var topHorizontal = this.horizontalHallway(20, 12, 8);
            var bottomHorizontal = this.horizontalHallway(15, 28, 14);
            var crossBoth = this.verticalHallway(25, topHorizontal.startY, bottomHorizontal.startY - topHorizontal.startY);
            var topCross = this.verticalHallway(20, topHorizontal.startY, middleHorizontal.startY - topHorizontal.startY);
            var bottomCross = this.verticalHallway(15, middleHorizontal.startY, bottomHorizontal.startY - middleHorizontal.startY);
            if (roll(60)) {
                this.roomAtStart(crossBoth, 3, false);
            }
            if (roll(60)) {
                this.roomAtStart(topCross, 3, false);
            }
            if (roll(80)) {
                this.roomAtEnd(middleHorizontal, 3, true);
            }
            if (roll(60)) {
                this.roomAtEnd(topHorizontal, 3, true);
            }
            if (roll(60)) {
                this.roomAtEnd(bottomHorizontal, 3, true);
            }
            if (roll(60)) {
                this.roomAtEnd(crossBoth, 3, false);
            }
            if (roll(60)) {
                this.roomAtEnd(bottomCross, 3, false);
            }
            if (roll(70)) {
                this.roomAt(topHorizontal.startY + 2, topCross.startX - 5, middleHorizontal.startY - 1, topCross.startX - 1);
                this.horizontalHallway(topCross.startX - 1, topHorizontal.startY + 4, 1);
            }
            if (roll(70)) {
                this.roomAt(middleHorizontal.startY + 2, bottomCross.startX - 5, bottomHorizontal.startY - 1, bottomCross.startX - 1);
                this.horizontalHallway(bottomCross.startX - 1, middleHorizontal.startY + 4, 1);
            }
            if (roll(70)) {
                this.roomAt(middleHorizontal.startY + 2, crossBoth.startX - 5, bottomHorizontal.startY - 1, crossBoth.startX - 1);
                this.horizontalHallway(crossBoth.startX - 1, middleHorizontal.startY + 4, 1);
            }
            var exitHallway = Math.random() > 0.5 ? topHorizontal : bottomHorizontal;
            this.placeExit(exitHallway.startX + exitHallway.length - 1, exitHallway.startY);
            this._mapInProgress.heroStartX = (middleHorizontal.startX + 0.5) * this._mapInProgress.tileSize;
            this._mapInProgress.heroStartY = (middleHorizontal.startY + 0.5) * this._mapInProgress.tileSize;
            for (var i = 0; i < this.enemyCount; i++) {
                RogueRpg.Helpers.game.spawnMonster(this._mapInProgress);
            }
        };
        MapGenerator.prototype.createHallwaysMap = function (tileset) {
            var tilesets = ["Dungeon", "Small", "Yellow Brick", "Cave", "Cosmic"];
            this.setupMap(35, tilesets, tileset);
            var mainHallway = this.mainHorizontalHallway();
            var crossHallway = this.crossHallway(mainHallway);
            var topRoom = this.createTopRoom(crossHallway);
            var bottomRoom = this.createBottomRoom(crossHallway);
            var parallelHallway = this.parallelHallway(mainHallway);
            var crossHallway2 = this.crossHallway(mainHallway);
            var topRoom2 = this.createTopRoom(crossHallway2);
            var hallwayRoom = this.createHallwayRoom(mainHallway);
            var hallwayRoom = this.createHallwayRoom(parallelHallway);
            var sideRoom = this.createSideRoom(mainHallway);
            this._mapInProgress.heroStartX = 100;
            this._mapInProgress.heroStartY = mainHallway.startY * this._mapInProgress.tileSize + 60;
            for (var i = 0; i < this.enemyCount; i++) {
                RogueRpg.Helpers.game.spawnMonster(this._mapInProgress);
            }
        };
        MapGenerator.prototype.createRiverTunnelsMap = function (tileset) {
            var tilesets = ["Dungeon", "Small", "Yellow Brick", "Cave", "Cosmic"];
            this.setupMap(45, tilesets, tileset);
            var xOffset = 5;
            var increment = 0;
            var hallways = [];
            for (var i = 0; i < 8; i++) {
                hallways.push(this.verticalHallway(xOffset, randIntMap(4, 7), randIntMap(18, 23)));
                if (i < 7) {
                    increment = randIntMap(2, 4);
                    this.horizontalHallway(xOffset, randIntMap(7, 20), increment);
                    xOffset += increment;
                }
            }
            this._mapInProgress.heroStartX = (hallways[0].startX + 0.5) * this._mapInProgress.tileSize;
            this._mapInProgress.heroStartY = (hallways[0].startY + 0.5) * this._mapInProgress.tileSize;
            if (RogueRpg.Helpers.game.highestClassUnlocked < 3) {
                var room = { left: xOffset + 2, top: 7, right: xOffset + 7, bottom: 12 };
                this.drawRoomToMap(room);
                this.addWorldObjectToRoom(room, "Frozen Monk");
            }
            this.placeExit(hallways[hallways.length - 1].startX, hallways[hallways.length - 1].startY + hallways[hallways.length - 1].length - 1);
            for (var i = 0; i < this.enemyCount; i++) {
                RogueRpg.Helpers.game.spawnMonster(this._mapInProgress);
            }
        };
        MapGenerator.prototype.createPlaneMap = function (tileType) {
            this.setupMap(40, [tileType]);
            var room = {
                left: 11,
                top: 11,
                right: 19,
                bottom: 19
            };
            this.drawRoomToMap(room);
            this.placeExitInRoom(room);
            if (tileType === "Magma") {
                for (var i = 0; i < 8; i++) {
                    var mon = new RogueRpg.Monster("Fireling", this._mapInProgress, 0, this.randomLocationInRoom(room));
                    this._mapInProgress.monsters.push(mon);
                }
                var location = new RogueRpg.Location(room.right - 2, room.bottom - 2);
                this.addWorldObjectToTile(location, "Dark Treasure Chest");
            }
            else {
                for (var i = 0; i < 50; i++) {
                    var mon = RogueRpg.Helpers.game.spawnMonster(this._mapInProgress, undefined, undefined, undefined, undefined, 0.40);
                    mon.dropRateModifier = 0.15;
                    mon.experienceGrantedOnKill *= 0.6;
                }
                var location = new RogueRpg.Location(room.right - 2, room.bottom - 2);
                this.addWorldObjectToTile(location, "Treasure Chest");
            }
            this._mapInProgress.heroStartX = 15 * this._mapInProgress.tileSize;
            this._mapInProgress.heroStartY = 16 * this._mapInProgress.tileSize;
        };
        MapGenerator.prototype.createUndergroundDesertLevel = function (tileType) {
            if (tileType === void 0) { tileType = "Yellow Brick"; }
            this.setupMap(55, [tileType]);
            var entryRoom = this.roomAt(20, 11, 23, 14);
            var topRoom = this.hlpGenerateRandomAdjacentRoomFrom(entryRoom, "top", false, true);
            if (topRoom) {
                if (roll(70)) {
                    this.hlpGenerateRandomAdjacentRoomFrom(topRoom, "top", false, true);
                }
                if (roll(50)) {
                    this.hlpGenerateRandomAdjacentRoomFrom(topRoom, "right", false, true);
                }
                if (roll(70)) {
                    this.hlpGenerateRandomAdjacentRoomFrom(topRoom, "left", false, true);
                }
            }
            var bottomRoom = this.hlpGenerateRandomAdjacentRoomFrom(entryRoom, "bottom", false, true);
            if (bottomRoom) {
                if (roll(70)) {
                    this.hlpGenerateRandomAdjacentRoomFrom(bottomRoom, "bottom", false, true);
                }
                if (roll(50)) {
                    this.hlpGenerateRandomAdjacentRoomFrom(bottomRoom, "right", false, true);
                }
                if (roll(70)) {
                    this.hlpGenerateRandomAdjacentRoomFrom(bottomRoom, "left", false, true);
                }
            }
            var rightRoom = this.hlpGenerateRandomAdjacentRoomFrom(entryRoom, "right", true, true);
            var right2Room = this.hlpGenerateRandomAdjacentRoomFrom(rightRoom, "right", true, true);
            var right3Room = this.hlpGenerateRandomAdjacentRoomFrom(right2Room, "right", true, true);
            this.hlpGenerateRandomAdjacentRoomFrom(rightRoom, "bottom", false, true);
            this.hlpGenerateRandomAdjacentRoomFrom(rightRoom, "top", false, true);
            if (roll(60)) {
                this.hlpGenerateRandomAdjacentRoomFrom(right2Room, "bottom", false, true);
            }
            if (roll(60)) {
                this.hlpGenerateRandomAdjacentRoomFrom(right2Room, "top", false, true);
            }
            this.placeExitInRoom(right3Room);
            this._mapInProgress.heroStartX = (entryRoom.left + entryRoom.right + 0.5) / 2 * this._mapInProgress.tileSize;
            this._mapInProgress.heroStartY = (entryRoom.top + entryRoom.bottom + 0.5) / 2 * this._mapInProgress.tileSize;
        };
        MapGenerator.prototype.createUndergroundDesertLevelAlternate = function (tilesetName) {
            if (tilesetName === void 0) { tilesetName = "Yellow Brick"; }
            this.setupMap(40, [tilesetName]);
            var entryRoom = this.roomAt(7, 7, 11, 11);
            var rightRoom = this.hlpGenerateRandomAdjacentRoomFrom(entryRoom, "right", true, true);
            var rightDownRoom = this.hlpGenerateRandomAdjacentRoomFrom(rightRoom, "bottom", true, true);
            var rdrRoom = this.hlpGenerateRandomAdjacentRoomFrom(rightDownRoom, "right", true, true);
            var rdrdRoom = this.hlpGenerateRandomAdjacentRoomFrom(rdrRoom, "bottom", true, true);
            var rdrdlRoom = this.hlpGenerateRandomAdjacentRoomFrom(rdrdRoom, "left", true, true);
            if (roll(50)) {
                var opt1Room = this.hlpGenerateRandomAdjacentRoomFrom(rightRoom, "right", true, true);
            }
            else {
                var opt1Room = this.hlpGenerateRandomAdjacentRoomFrom(rdrRoom, "top", true, true);
            }
            var opt1bRoom = this.hlpGenerateRandomAdjacentRoomFrom(opt1Room, "right", true, false);
            var downRoom = this.hlpGenerateRandomAdjacentRoomFrom(entryRoom, "bottom", true, true);
            var down2Room = this.hlpGenerateRandomAdjacentRoomFrom(downRoom, "bottom", true, true);
            var ddrRoom = this.hlpGenerateRandomAdjacentRoomFrom(down2Room, "right", true, true);
            var ddrrRoom = this.hlpGenerateRandomAdjacentRoomFrom(ddrRoom, "right", true, true);
            if (roll(50)) {
                var opt2Room = this.hlpGenerateRandomAdjacentRoomFrom(down2Room, "bottom", true, true);
            }
            else {
                var opt2Room = this.hlpGenerateRandomAdjacentRoomFrom(ddrRoom, "bottom", true, true);
            }
            if (roll(50)) {
                var opt2bRoom = this.hlpGenerateRandomAdjacentRoomFrom(opt2Room, "right", true, true);
            }
            else {
                var opt2bRoom = this.hlpGenerateRandomAdjacentRoomFrom(opt2Room, "bottom", true, true);
            }
            if (roll(50)) {
                var opt3Room = this.hlpGenerateRandomAdjacentRoomFrom(rdrRoom, "right", true, true);
            }
            else {
                var opt3Room = this.hlpGenerateRandomAdjacentRoomFrom(rdrdRoom, "right", true, true);
            }
            if (roll(50)) {
                var opt3bRoom = this.hlpGenerateRandomAdjacentRoomFrom(opt3Room, "right", true, true);
            }
            else {
                var opt3bRoom = this.hlpGenerateRandomAdjacentRoomFrom(opt3Room, "bottom", true, true);
            }
            var opt3cRoom = this.hlpGenerateRandomAdjacentRoomFrom(opt3bRoom, "bottom", true, false);
            this.placeExitInRoom(opt3cRoom);
            this._mapInProgress.heroStartX = (entryRoom.left + entryRoom.right + 0.5) / 2 * this._mapInProgress.tileSize;
            this._mapInProgress.heroStartY = (entryRoom.top + entryRoom.bottom + 0.5) / 2 * this._mapInProgress.tileSize;
        };
        MapGenerator.prototype.createLinkedRoomsMapOrcCave = function (tilesetOverride) {
            var tilesets = ["Dungeon", "Small", "Yellow Brick", "Cave", "Cosmic"];
            this.setupMap(50, tilesets, tilesetOverride);
            var entryRoom = this.roomAt(20, 11, 23, 14);
            var topRoom = this.hlpGenerateRandomAdjacentRoomFrom(entryRoom, "top");
            if (topRoom) {
                if (roll(70)) {
                    this.hlpGenerateRandomAdjacentRoomFrom(topRoom, "top");
                }
                if (roll(50)) {
                    this.hlpGenerateRandomAdjacentRoomFrom(topRoom, "right");
                }
                if (roll(70)) {
                    this.hlpGenerateRandomAdjacentRoomFrom(topRoom, "left");
                }
            }
            var bottomRoom = this.hlpGenerateRandomAdjacentRoomFrom(entryRoom, "bottom");
            if (bottomRoom) {
                if (roll(70)) {
                    this.hlpGenerateRandomAdjacentRoomFrom(bottomRoom, "bottom");
                }
                if (roll(50)) {
                    this.hlpGenerateRandomAdjacentRoomFrom(bottomRoom, "right");
                }
                if (roll(70)) {
                    this.hlpGenerateRandomAdjacentRoomFrom(bottomRoom, "left");
                }
            }
            var rightRoom = this.hlpGenerateRandomAdjacentRoomFrom(entryRoom, "right", true);
            var right2Room = this.hlpGenerateRandomAdjacentRoomFrom(rightRoom, "right", true);
            var right3Room = this.hlpGenerateRandomAdjacentRoomFrom(right2Room, "right", true);
            var right4Room = this.hlpGenerateRandomAdjacentRoomFrom(right3Room, "right", true, false, true);
            this.hlpGenerateRandomAdjacentRoomFrom(rightRoom, "bottom");
            this.hlpGenerateRandomAdjacentRoomFrom(rightRoom, "top");
            if (roll(60)) {
                this.hlpGenerateRandomAdjacentRoomFrom(right2Room, "bottom");
            }
            if (roll(60)) {
                this.hlpGenerateRandomAdjacentRoomFrom(right2Room, "top");
            }
            var bossLocation = new RogueRpg.Location(this._mapInProgress.tileSize * (right3Room.right + right3Room.left) / 2, this._mapInProgress.tileSize * (right3Room.bottom + right3Room.top) / 2);
            RogueRpg.Helpers.game.spawnSpecificMonster(this._mapInProgress, bossLocation, "Boarman");
            this.placeExitInRoom(right4Room);
            this._mapInProgress.heroStartX = (entryRoom.left + entryRoom.right + 0.5) / 2 * this._mapInProgress.tileSize;
            this._mapInProgress.heroStartY = (entryRoom.top + entryRoom.bottom + 0.5) / 2 * this._mapInProgress.tileSize;
        };
        MapGenerator.prototype.hlpGenerateRandomAdjacentRoomFrom = function (room, direction, allowOverlap, orderlyHalls, skipPlacingContent) {
            var hallLength = randIntMap(2, 3);
            var roomWidth = orderlyHalls ? 3 : randIntMap(2, 4);
            var roomHeight = orderlyHalls ? 3 : randIntMap(2, 4);
            var roomOffset = 0;
            if (direction == "top" || direction == "bottom") {
                roomOffset = orderlyHalls ? -Math.floor(roomWidth / 2) : randIntMap(-roomWidth, 1);
            }
            else {
                roomOffset = orderlyHalls ? -Math.floor(roomHeight / 2) : randIntMap(-1, 1);
            }
            var newRoom;
            if (direction == "top") {
                var hallOffset = orderlyHalls ? Math.floor(roomWidth / 2) : randIntMap(-1, room.right - room.left);
                newRoom = {
                    top: room.top - (roomHeight + hallLength),
                    left: room.left + hallOffset + roomOffset,
                    bottom: room.top - hallLength,
                    right: room.left + hallOffset + roomOffset + roomWidth
                };
                if (this.isRoomSpaceUsable(allowOverlap, newRoom.top, newRoom.left, newRoom.bottom, newRoom.right)) {
                    this.drawRoomToMap(newRoom);
                    this.pathTo(new RogueRpg.Location(room.left + hallOffset, room.top + 1), new RogueRpg.Location(room.left + hallOffset, newRoom.bottom - 1));
                }
                else {
                    return null;
                }
            }
            else if (direction == "bottom") {
                var hallOffset = orderlyHalls ? Math.floor(roomWidth / 2) : randIntMap(-1, room.right - room.left);
                newRoom = {
                    top: room.bottom + hallLength,
                    left: room.left + hallOffset + roomOffset,
                    bottom: room.bottom + hallLength + roomHeight,
                    right: room.left + hallOffset + roomOffset + roomWidth
                };
                if (this.isRoomSpaceUsable(allowOverlap, newRoom.top, newRoom.left, newRoom.bottom, newRoom.right)) {
                    this.drawRoomToMap(newRoom);
                    this.pathTo(new RogueRpg.Location(room.left + hallOffset, room.bottom - 1), new RogueRpg.Location(room.left + hallOffset, newRoom.top + 1));
                }
                else {
                    return null;
                }
            }
            else if (direction == "right") {
                var hallOffset = orderlyHalls ? Math.floor(roomHeight / 2) : randIntMap(-1, room.bottom - room.top);
                newRoom = {
                    top: room.top + hallOffset + roomOffset,
                    left: room.right + hallLength,
                    bottom: room.top + hallOffset + roomOffset + roomWidth,
                    right: room.right + hallLength + roomWidth
                };
                if (this.isRoomSpaceUsable(allowOverlap, newRoom.top, newRoom.left, newRoom.bottom, newRoom.right)) {
                    this.drawRoomToMap(newRoom);
                    this.pathTo(new RogueRpg.Location(room.right - 1, room.top + hallOffset), new RogueRpg.Location(newRoom.left, room.top + hallOffset));
                }
                else {
                    return null;
                }
            }
            else if (direction == "left") {
                var hallOffset = orderlyHalls ? Math.floor(roomHeight / 2) : randIntMap(-1, room.bottom - room.top);
                newRoom = {
                    top: room.top + hallOffset + roomOffset,
                    left: room.left - hallLength - roomWidth,
                    bottom: room.top + hallOffset + roomOffset + roomWidth,
                    right: room.left - (hallLength)
                };
                if (this.isRoomSpaceUsable(allowOverlap, newRoom.top, newRoom.left, newRoom.bottom, newRoom.right)) {
                    this.drawRoomToMap(newRoom);
                    this.pathTo(new RogueRpg.Location(newRoom.right - 1, room.top + hallOffset), new RogueRpg.Location(room.left, room.top + hallOffset));
                }
                else {
                    return null;
                }
            }
            if (newRoom && !skipPlacingContent) {
                var roomReward = 1;
                var enemyPopulation = RogueRpg.Helpers.SelectOne(MapGenerator.RoomTypeEnemies);
                if (enemyPopulation == "Empty") {
                    roomReward--;
                }
                else if (enemyPopulation == "Normal") {
                    RogueRpg.Helpers.game.spawnMonster(this._mapInProgress, this.randomLocationInRoom(newRoom));
                    RogueRpg.Helpers.game.spawnMonster(this._mapInProgress, this.randomLocationInRoom(newRoom));
                    if (roll(15)) {
                        RogueRpg.Helpers.game.spawnMonster(this._mapInProgress, this.randomLocationInRoom(newRoom), 1);
                        roomReward++;
                    }
                }
                else if (enemyPopulation == "Dense") {
                    RogueRpg.Helpers.game.spawnMonster(this._mapInProgress, this.randomLocationInRoom(newRoom));
                    RogueRpg.Helpers.game.spawnMonster(this._mapInProgress, this.randomLocationInRoom(newRoom));
                    RogueRpg.Helpers.game.spawnMonster(this._mapInProgress, this.randomLocationInRoom(newRoom));
                    RogueRpg.Helpers.game.spawnMonster(this._mapInProgress, this.randomLocationInRoom(newRoom));
                    roomReward++;
                }
                else if (enemyPopulation == "StrongSingle") {
                    RogueRpg.Helpers.game.spawnMonster(this._mapInProgress, this.randomLocationInRoom(newRoom), 1);
                    roomReward++;
                    if (roll(15)) {
                        RogueRpg.Helpers.game.spawnMonster(this._mapInProgress, this.randomLocationInRoom(newRoom), 1);
                        roomReward++;
                    }
                }
                else if (enemyPopulation == "WeakHorde") {
                    RogueRpg.Helpers.game.spawnMonster(this._mapInProgress, this.randomLocationInRoom(newRoom), -1);
                    RogueRpg.Helpers.game.spawnMonster(this._mapInProgress, this.randomLocationInRoom(newRoom), -1);
                    RogueRpg.Helpers.game.spawnMonster(this._mapInProgress, this.randomLocationInRoom(newRoom), -1);
                    RogueRpg.Helpers.game.spawnMonster(this._mapInProgress, this.randomLocationInRoom(newRoom), -1);
                    RogueRpg.Helpers.game.spawnMonster(this._mapInProgress, this.randomLocationInRoom(newRoom), -1);
                    RogueRpg.Helpers.game.spawnMonster(this._mapInProgress, this.randomLocationInRoom(newRoom), -1);
                    roomReward++;
                    if (roll(15)) {
                        RogueRpg.Helpers.game.spawnMonster(this._mapInProgress, this.randomLocationInRoom(newRoom));
                        RogueRpg.Helpers.game.spawnMonster(this._mapInProgress, this.randomLocationInRoom(newRoom));
                        roomReward++;
                    }
                }
                roomReward += randInt(-1, 1);
                var special = this.tryAddSpecialReward(newRoom);
                if (special) {
                    return special;
                }
                if (RogueRpg.Helpers.game.hero.state.multiplier >= 50 && Math.random() < 0.012) {
                    this.addWorldObjectToRoom(newRoom, "Prisma Found");
                }
                if (roomReward <= 0) {
                    RogueRpg.Helpers.game.spawnMonster(this._mapInProgress, this.randomLocationInRoom(newRoom), -2);
                    RogueRpg.Helpers.game.spawnMonster(this._mapInProgress, this.randomLocationInRoom(newRoom), -2);
                    RogueRpg.Helpers.game.spawnMonster(this._mapInProgress, this.randomLocationInRoom(newRoom), -2);
                }
                else if (roomReward == 1 && Math.random() < 0.18) {
                    this.addWorldObjectToRoom(newRoom, "Healing Shrine");
                }
                else if (roomReward == 2 || roomReward == 3) {
                    var bonus = RogueRpg.Helpers.hero.state.multiplier > 1 ? 0.05 : 0;
                    if (Math.random() < 0.16 + bonus) {
                        if (Math.random() < 0.12) {
                            this.addWorldObjectToRoom(newRoom, "Sitting Wizard");
                        }
                        else if (Math.random() > 0.4) {
                            var potion = RogueRpg.Game.randomPotion();
                            var dp = new RogueRpg.DroppedPotion(potion, new RogueRpg.Location(this._mapInProgress.tileSize * (newRoom.right + newRoom.left) / 2, this._mapInProgress.tileSize * (newRoom.top + newRoom.bottom) / 2));
                            this._mapInProgress.potionsOnGround.push(dp);
                        }
                    }
                    else {
                        var type = "Treasure Chest";
                        if (this.tileType === "Dungeon" && Math.random() > 0.97) {
                            type = "Dark Treasure Chest";
                        }
                        this.addWorldObjectToRoom(newRoom, type);
                    }
                }
                else if (roomReward >= 4 && Math.random() > 0.5) {
                    if (Math.random() > 0.5) {
                        this.addWorldObjectToRoom(newRoom, "Item Upgrade Shrine");
                    }
                    else {
                        this.addWorldObjectToRoom(newRoom, "Portal");
                    }
                }
            }
            return newRoom;
        };
        // Special rewards are all the unlockables!
        MapGenerator.prototype.tryAddSpecialReward = function (newRoom) {
            var floor = RogueRpg.Helpers.game.floor;
            var mult = RogueRpg.Helpers.hero.state.multiplier;
            var highestClass = RogueRpg.Helpers.game.highestClassUnlocked;
            var maxPotions = RogueRpg.Helpers.game.maxPotions;
            if (!MapGenerator.addedClassThisStage) {
                if (RogueRpg.Helpers.game.floor > 10 && RogueRpg.Helpers.game.highestClassUnlocked == 0) {
                    if (Math.random() > 0.95) {
                        this.addWorldObjectToRoom(newRoom, "Bowman Cap Item");
                        MapGenerator.addedClassThisStage = true;
                        return newRoom; // Skip the other rewards
                    }
                }
                else if (RogueRpg.Helpers.game.floor > 15 && RogueRpg.Helpers.game.highestClassUnlocked == 1) {
                    if (Math.random() > 0.95) {
                        this.addWorldObjectToRoom(newRoom, "Wizard Hat Item");
                        MapGenerator.addedClassThisStage = true;
                        return newRoom; // Skip the other rewards
                    }
                }
            }
            // New: Life, Dmg, Def, Str, Dex, luck, Cdr, spd, 
            if (!MapGenerator.addedRelicThisStage) {
                if (RogueRpg.Helpers.hero.state.multiplier >= 2 && RogueRpg.Helpers.game.floor > 5 && !RogueRpg.Helpers.game.startingWeapon) {
                    this.addWorldObjectToRoom(newRoom, "Scumbasher +1");
                    MapGenerator.addedRelicThisStage = true;
                    return newRoom;
                }
                else if (RogueRpg.Helpers.hero.state.multiplier >= 4 && RogueRpg.Helpers.game.floor > 5 && RogueRpg.Helpers.game.startingWeapon === "Scumbasher +1") {
                    this.addWorldObjectToRoom(newRoom, "Scumbasher +2");
                    MapGenerator.addedRelicThisStage = true;
                    return newRoom;
                }
                if (maxPotions <= 5 && mult === 2 && floor === 17) {
                    this.addWorldObjectToRoom(newRoom, "Potion Slot");
                    MapGenerator.addedRelicThisStage = true;
                    return newRoom;
                }
                else if (maxPotions <= 6 && mult === 4 && floor === 17) {
                    this.addWorldObjectToRoom(newRoom, "Potion Slot");
                    MapGenerator.addedRelicThisStage = true;
                    return newRoom;
                }
                else if (maxPotions <= 7 && mult === 8 && floor === 17) {
                    this.addWorldObjectToRoom(newRoom, "Potion Slot");
                    MapGenerator.addedRelicThisStage = true;
                    return newRoom;
                }
                for (var i = 0; i < RogueRpg.Relic.Unlockables.length; i++) {
                    var relic = RogueRpg.Relic.Unlockables[i];
                    if (!RogueRpg.Relic.RelicLevels[relic.relicIndex] && RogueRpg.Helpers.game.floor >= relic.minFloor &&
                        Math.random() > (1 - relic.chance) && RogueRpg.Helpers.hero.state.multiplier == relic.difficulty) {
                        this.addWorldObjectToRoom(newRoom, relic.relic + " World");
                        MapGenerator.addedRelicThisStage = true;
                        return newRoom;
                    }
                }
            }
        };
        MapGenerator.prototype.randomLocationInRoom = function (room) {
            var x = (room.left * this._mapInProgress.tileSize) + 30 + Math.random() * ((room.right - room.left) * this._mapInProgress.tileSize - 60);
            var y = (room.top * this._mapInProgress.tileSize) + Math.random() * ((room.bottom - room.top) * this._mapInProgress.tileSize - 50);
            return new RogueRpg.Location(x, y);
        };
        /** Generic: Returns true if the point is a valid floor space inside the current map bounds */
        MapGenerator.prototype.isPointInvalid = function (y, x) {
            return (x < 0 || y < 1 || x >= this._mapInProgress.mapWidth || y >= this._mapInProgress.mapHeight);
        };
        MapGenerator.prototype.isRoomSpaceUsable = function (allowOverlap, top, left, bottom, right) {
            if (this.isPointInvalid(top, left) || this.isPointInvalid(bottom, right)) {
                return false;
            }
            if (!allowOverlap) {
                for (var i = top - 1; i <= bottom; i++) {
                    for (var j = left; j <= right; j++) {
                        if (this._mapInProgress.tiles[i][j].type) {
                            return false;
                        }
                    }
                }
            }
            return true;
        };
        MapGenerator.prototype.roomAtStart = function (hallway, size, offsetFromEnd) {
            if (offsetFromEnd) {
                if (hallway.direction == "x") {
                    this.roomAt(hallway.startY - Math.floor(size / 2), hallway.startX - size, hallway.startY + Math.ceil(size / 2), hallway.startX);
                }
                else {
                    this.roomAt(hallway.startY - size, hallway.startX - Math.floor(size / 2), hallway.startY, hallway.startX + Math.ceil(size / 2));
                }
            }
            else {
                this.roomAt(hallway.startY - Math.floor(size / 2), hallway.startX - Math.floor(size / 2), hallway.startY + Math.ceil(size / 2), hallway.startX + Math.ceil(size / 2));
            }
        };
        MapGenerator.prototype.addFloorDecor = function (images, count) {
            for (var i = 0; i < count; i++) {
                var image = RogueRpg.Helpers.SelectOne(images);
                var object = new RogueRpg.WorldObject(image.name);
                object.drawHeight = image.size;
                object.drawWidth = image.size;
                while (!object.drawX) {
                    var x = randIntMap(0, this._mapInProgress.mapWidth - 1);
                    var y = randIntMap(0, this._mapInProgress.mapHeight - 1);
                    if (this._mapInProgress.tiles[y][x].walkable) {
                        object.drawX = x * this._mapInProgress.tileSize;
                        object.drawY = y * this._mapInProgress.tileSize;
                    }
                }
                // Insert at the front so it draws underneath other world objects
                this._mapInProgress.specialWorldObjects.splice(0, 0, object);
            }
        };
        MapGenerator.prototype.roomAtEnd = function (hallway, size, offsetFromEnd) {
            var endX = hallway.startX + hallway.length;
            var endY = hallway.startY + hallway.length;
            if (offsetFromEnd) {
                if (hallway.direction == "x") {
                    this.roomAt(hallway.startY - Math.floor(size / 2), endX, hallway.startY + Math.ceil(size / 2), endX + size);
                }
                else {
                    this.roomAt(endY, hallway.startX - Math.floor(size / 2), endY + size, hallway.startX + Math.ceil(size / 2));
                }
            }
            else {
                if (hallway.direction == "x") {
                    this.roomAt(hallway.startY - Math.floor(size / 2), endX - Math.floor(size / 2), hallway.startY + Math.ceil(size / 2), endX + Math.ceil(size / 2));
                }
                else {
                    this.roomAt(endY - Math.floor(size / 2), hallway.startX - Math.floor(size / 2), endY + Math.ceil(size / 2), hallway.startX + Math.ceil(size / 2));
                }
            }
        };
        MapGenerator.prototype.roomAt = function (top, left, bottom, right) {
            var room = {
                left: left,
                top: top,
                right: right,
                bottom: bottom
            };
            this.drawRoomToMap(room);
            if (roll(25)) {
                this.addDecorativeWorldObjectAt(top, left);
            }
            if (roll(25)) {
                this.addDecorativeWorldObjectAt(top, right - 1);
            }
            if (roll(25)) {
                this.addDecorativeWorldObjectAt(bottom - 1, left);
            }
            if (roll(25)) {
                this.addDecorativeWorldObjectAt(bottom - 1, right - 1);
            }
            this.addRandomWorldObjectToRoom(room);
            return room;
        };
        MapGenerator.prototype.addDecorativeWorldObjectAt = function (y, x) {
            var location = new RogueRpg.Location(x, y);
            var specialRoll = Math.random();
            var object;
            if (specialRoll < 1.0 && RogueRpg.Helpers.game.floor == 1) {
                object = Math.random() > 0.2 ? new RogueRpg.WorldObject("Empty Mine Cart") : new RogueRpg.WorldObject("Filled Mine Cart");
                object.drawHeight = 80;
                object.drawWidth = 80;
            }
            // Put in center of the tile
            if (object) {
                object.drawX = location.x * this._mapInProgress.tileSize - object.drawWidth / 2 + this._mapInProgress.tileSize / 2;
                object.drawY = location.y * this._mapInProgress.tileSize - 10;
                this._mapInProgress.specialWorldObjects.push(object);
            }
        };
        MapGenerator.prototype.setupMap = function (size, tilesets, tilesetOverride) {
            var type = MapGenerator.Tilesets[tilesetOverride || RogueRpg.Helpers.SelectOne(tilesets)];
            this.tileType = type.name;
            this.decorTypes = type.decors;
            this.enemyCount = type.enemies;
            // Sets up an empty 30x30 map for the generator to use
            this._mapInProgress = new RogueRpg.Map();
            this._mapInProgress.tileSize = type.size;
            this._mapInProgress.mapHeight = size;
            this._mapInProgress.mapWidth = size;
            this._mapInProgress.tiles = [];
            for (var i = 0; i < size; i++) {
                var row = [];
                for (var j = 0; j < size; j++) {
                    var tile = new RogueRpg.Tile();
                    //if (type.backgroundType == "tiled") {
                    //    tile.type = type.name + " Background";
                    //}
                    row.push(tile);
                }
                this._mapInProgress.tiles.push(row);
            }
        };
        MapGenerator.prototype.branchOutFrom = function (point, previousDirection, scalingFactor) {
            if (this.currentLevelNumber < 5 && scalingFactor < 0.7 * 0.7 * 0.7) {
                return 0;
            }
            var distance = randIntMap(3, 5);
            var isDeadEnd = true;
            var numBranches = 0;
            if (previousDirection != "up" && Math.random() < scalingFactor) {
                var newPoint = new RogueRpg.Location(point.x, point.y + distance);
                if (this.pathTo(point, newPoint)) {
                    numBranches += this.branchOutFrom(newPoint, "down", scalingFactor * 0.7);
                    isDeadEnd = false;
                }
            }
            if (previousDirection != "down" && Math.random() < scalingFactor) {
                var newPoint = new RogueRpg.Location(point.x, point.y - distance);
                if (this.pathTo(point, newPoint)) {
                    numBranches += this.branchOutFrom(newPoint, "up", scalingFactor * 0.7);
                    isDeadEnd = false;
                }
            }
            if (previousDirection != "right" && Math.random() < scalingFactor) {
                var newPoint = new RogueRpg.Location(point.x + distance, point.y);
                if (this.pathTo(point, newPoint)) {
                    numBranches += this.branchOutFrom(newPoint, "left", scalingFactor * 0.7);
                    isDeadEnd = false;
                }
            }
            if (previousDirection != "left" && Math.random() < scalingFactor) {
                var newPoint = new RogueRpg.Location(point.x - distance, point.y);
                if (this.pathTo(point, newPoint)) {
                    numBranches += this.branchOutFrom(newPoint, "right", scalingFactor * 0.7);
                    isDeadEnd = false;
                }
            }
            if (isDeadEnd && !this.exitPlaced) {
                this.placeExit(point.x, point.y);
            }
            if (previousDirection != "none") {
                if (Math.random() < .02) {
                    this.addRandomWorldObjectToTile(point);
                }
                else if (isDeadEnd && Math.random() < 0.65 - (scalingFactor / 2) && Math.random() < 0.7) {
                    this.addRandomWorldObjectToTile(point);
                }
            }
            return numBranches;
        };
        MapGenerator.prototype.placeExitInRoom = function (room) {
            this.placeExit((room.left + room.right) / 2 - 1, (room.top + room.bottom) / 2 - 1);
        };
        MapGenerator.prototype.placeExit = function (x, y) {
            this.addWorldObjectToTile(new RogueRpg.Location(x, y), "Level Exit");
        };
        /** Generic: If start and end share a single axis, draws a hallway between the two points. Returns 'true' only if the hallway was actually added (it could have ran into the edge of the map) */
        MapGenerator.prototype.pathTo = function (start, end) {
            if (end.x < 0 || end.y < 1 || end.x >= this._mapInProgress.mapWidth || end.y >= this._mapInProgress.mapHeight ||
                start.x < 0 || start.y < 1 || start.x >= this._mapInProgress.mapWidth || start.y >= this._mapInProgress.mapHeight) {
                return false;
            }
            if (start.x == end.x && start.y < end.y) {
                for (var y = start.y; y <= end.y; y++) {
                    this.makeFloor(y, start.x);
                    this.fillWall(y - 1, start.x);
                }
            }
            else if (start.x == end.x && start.y > end.y) {
                for (var y = start.y; y >= end.y; y--) {
                    this.makeFloor(y, start.x);
                    this.fillWall(y - 1, start.x);
                }
            }
            else if (start.y == end.y && start.x < end.x) {
                for (var x = start.x; x <= end.x; x++) {
                    this.makeFloor(start.y, x);
                    this.fillWall(start.y - 1, x);
                }
            }
            else {
                for (var x = start.x; x >= end.x; x--) {
                    this.makeFloor(start.y, x);
                    this.fillWall(start.y - 1, x);
                }
            }
            return true;
        };
        MapGenerator.prototype.fillWall = function (y, x) {
            if (!this._mapInProgress.tiles[y][x].type) {
                this._mapInProgress.tiles[y][x].type = this.getWallTile();
            }
            if (roll(35)) {
                if (this.decorTypes) {
                    this._mapInProgress.tiles[y][x].decor = RogueRpg.Helpers.SelectOne(this.decorTypes);
                }
            }
        };
        MapGenerator.prototype.horizontalHallway = function (startX, startY, length) {
            var hallway = {
                startX: startX,
                startY: startY,
                direction: "x",
                length: length
            };
            this.fillHorizontalHallway(hallway);
            return hallway;
        };
        MapGenerator.prototype.mainHorizontalHallway = function () {
            var hallway = {
                startX: 0,
                startY: 15,
                direction: "x",
                length: 20
            };
            this.fillHorizontalHallway(hallway);
            this.placeExit(hallway.startX + hallway.length - 1, hallway.startY);
            return hallway;
        };
        MapGenerator.prototype.verticalHallway = function (startX, startY, length) {
            var hallway = {
                startX: startX,
                startY: startY,
                direction: "y",
                length: length
            };
            this.fillVerticalHallway(hallway);
            return hallway;
        };
        MapGenerator.prototype.crossHallway = function (mainHallway) {
            var hallway = {
                startX: randIntMap(mainHallway.startX + 3, mainHallway.startX + mainHallway.length - 2),
                startY: randIntMap(mainHallway.startY - 8, mainHallway.startY - 2),
                direction: "y",
                length: randIntMap(8, mainHallway.startY - 2)
            };
            this.fillVerticalHallway(hallway);
            return hallway;
        };
        MapGenerator.prototype.fillHorizontalHallway = function (hallway) {
            for (var i = hallway.startX; i < hallway.startX + hallway.length; i++) {
                if (i < 0 || i >= this._mapInProgress.mapWidth)
                    continue;
                this.fillWall(hallway.startY - 1, i);
                this.makeFloor(hallway.startY, i);
            }
        };
        MapGenerator.prototype.fillVerticalHallway = function (hallway) {
            for (var i = hallway.startY; i < hallway.startY + hallway.length; i++) {
                this.makeFloor(i, hallway.startX);
            }
            this.fillWall(hallway.startY - 1, hallway.startX);
        };
        MapGenerator.prototype.parallelHallway = function (mainHallway) {
            var above = Math.random() > 0.5;
            var hallway = {
                startX: randIntMap(mainHallway.startX + 3, mainHallway.startX + 7),
                startY: mainHallway.startY + (above ? 2 : -2),
                direction: "x",
                length: randIntMap(mainHallway.length - 12, mainHallway.length - 7)
            };
            for (var i = hallway.startX; i < hallway.startX + hallway.length; i++) {
                this.makeFloor(hallway.startY, i);
                this.fillWall(hallway.startY - 1, i);
            }
            var placement = above ? -1 : 1;
            this.makeFloor(hallway.startY + placement, hallway.startX);
            this.makeFloor(hallway.startY + placement, hallway.startX + hallway.length - 1);
            return hallway;
        };
        MapGenerator.prototype.createHallwayRoom = function (hallway) {
            var roomHeight = randIntMap(2, 4);
            var roomWidth = randIntMap(2, 4);
            var roomCenter = randIntMap(hallway.startX + roomWidth, hallway.startX + hallway.length - roomWidth);
            var room = {
                left: roomCenter - Math.floor(roomWidth / 2),
                top: hallway.startY - Math.floor(roomHeight / 2),
                right: roomCenter + Math.ceil(roomWidth / 2),
                bottom: hallway.startY + Math.ceil(roomHeight / 2)
            };
            this.drawRoomToMap(room);
            this.addRandomWorldObjectToRoom(room);
            return room;
        };
        MapGenerator.prototype.createSideRoom = function (hallway) {
            var above = Math.random() > 0.5;
            var roomHeight = randIntMap(2, 4);
            var roomWidth = randIntMap(2, 4);
            var roomCenterX = randIntMap(hallway.startX + roomWidth, hallway.startX + hallway.length - roomWidth);
            var room;
            if (above) {
                var room = {
                    left: roomCenterX - Math.floor(roomWidth / 2),
                    top: hallway.startY + 2,
                    right: roomCenterX + Math.ceil(roomWidth / 2),
                    bottom: hallway.startY + 2 + roomHeight
                };
                // Draw the entrace to the room
                this._mapInProgress.tiles[hallway.startY + 1][roomCenterX].type = this.getFloorTile();
                this._mapInProgress.tiles[hallway.startY + 1][roomCenterX].walkable = true;
            }
            else {
                var room = {
                    left: roomCenterX - Math.floor(roomWidth / 2),
                    top: hallway.startY - 2 - roomHeight,
                    right: roomCenterX + Math.ceil(roomWidth / 2),
                    bottom: hallway.startY - 1
                };
                // Draw the entrace to the room
                this._mapInProgress.tiles[hallway.startY - 1][roomCenterX].type = this.getFloorTile();
                this._mapInProgress.tiles[hallway.startY - 1][roomCenterX].walkable = true;
            }
            this.drawRoomToMap(room);
            this.addRandomWorldObjectToRoom(room);
            return room;
        };
        MapGenerator.prototype.createTopRoom = function (hallway) {
            var roomHeight = randIntMap(3, 5);
            var roomWidth = randIntMap(3, 5);
            var room = {
                left: hallway.startX - Math.floor(roomWidth / 2),
                top: hallway.startY - roomHeight,
                right: hallway.startX + Math.ceil(roomWidth / 2),
                bottom: hallway.startY
            };
            this.drawRoomToMap(room);
            this.addRandomWorldObjectToRoom(room);
            return room;
        };
        MapGenerator.prototype.createBottomRoom = function (hallway) {
            var roomHeight = randIntMap(3, 5);
            var roomWidth = randIntMap(3, 5);
            var room = {
                left: hallway.startX - Math.floor(roomWidth / 2),
                top: hallway.startY + hallway.length,
                right: hallway.startX + Math.ceil(roomWidth / 2),
                bottom: hallway.startY + hallway.length + roomHeight
            };
            this.drawRoomToMap(room);
            this.addRandomWorldObjectToRoom(room);
            return room;
        };
        MapGenerator.prototype.drawRoomToMap = function (room) {
            for (var i = room.top; i < room.bottom; i++) {
                if (!this._mapInProgress.tiles[i])
                    continue;
                for (var j = room.left; j < room.right; j++) {
                    this._mapInProgress.tiles[i][j].type = this.getFloorTile();
                    this._mapInProgress.tiles[i][j].walkable = true;
                }
            }
            for (var j = room.left; j < room.right; j++) {
                this.fillWall(room.top - 1, j);
            }
        };
        // Expects a location where x and y are room offsets, not absolute pixel values
        MapGenerator.prototype.addWorldObjectToTile = function (location, type, parameter) {
            var object = new RogueRpg.WorldObject(type);
            if (type.indexOf(" World") > 0 || type.indexOf("Scumbasher") >= 0 || type === "Potion Slot") {
                object.drawHeight = 70;
                object.drawWidth = 40;
            }
            else if (type == "Healing Shrine") {
                object.drawHeight = 80;
                object.drawWidth = 40;
            }
            else if (type == "Item Upgrade Shrine") {
                object.drawHeight = 80;
                object.drawWidth = 40;
            }
            else if (type == "Treasure Chest" || type == "Dark Treasure Chest") {
                object.drawHeight = 100;
                object.drawWidth = 100;
            }
            else if (type.indexOf(" Item") > 0) {
                object.drawHeight = 100;
                object.drawWidth = 100;
            }
            else if (type == "Level Exit") {
                object.drawHeight = 80;
                object.drawWidth = 80;
                this.exitPlaced = true;
            }
            else if (type == "Sitting Wizard") {
                object.drawHeight = 80;
                object.drawWidth = 80;
            }
            else if (type == "Frozen Monk") {
                object.drawHeight = 130;
                object.drawWidth = 130;
            }
            else if (type == "Ghost Girl") {
                object.drawHeight = 80;
                object.drawWidth = 40;
            }
            else if (type == "Alchemy Table") {
                object.drawHeight = 75;
                object.drawWidth = 75;
            }
            else if (type == "Prisma Found") {
                object.drawHeight = 50;
                object.drawWidth = 50;
            }
            else if (type == "TestMonster") {
                var monster = new RogueRpg.Monster(parameter, null, 1);
                object = new RogueRpg.TestMonsterWorldObject(monster.type);
                object.drawHeight = monster.drawHeight;
                object.drawWidth = monster.drawWidth;
                object.data = parameter;
            }
            // Put in center of the tile
            if (object) {
                try {
                    if (MapGenerator.itemsForTiles[location.x + "x" + location.y]) {
                        if (object.type === "Level Exit") {
                            // Blast away any other object on this tile. The exit is #1
                            var index = this._mapInProgress.specialWorldObjects.indexOf(MapGenerator.itemsForTiles[location.x + "x" + location.y]);
                            if (index >= 0) {
                                this._mapInProgress.specialWorldObjects.splice(index, 1);
                            }
                        }
                        else {
                            // Don't add another object to this space
                            return;
                        }
                    }
                }
                catch (ex) {
                    // do nothing
                }
                MapGenerator.itemsForTiles[location.x + "x" + location.y] = object;
                object.drawX = location.x * this._mapInProgress.tileSize - object.drawWidth / 2 + this._mapInProgress.tileSize / 2;
                object.drawY = location.y * this._mapInProgress.tileSize - 10;
                this._mapInProgress.specialWorldObjects.push(object);
                if (type == "TestMonster") {
                    var obj = object;
                    obj.monster.x = object.drawX;
                    obj.monster.y = object.drawY;
                    obj.monster.drawX = obj.monster.x - obj.monster.drawWidth / 2;
                    obj.monster.drawY = obj.monster.y - obj.monster.drawHeight;
                }
            }
        };
        MapGenerator.prototype.addWorldObjectToRoom = function (room, type) {
            var location = new RogueRpg.Location(Math.floor((room.right + room.left) / 2), Math.floor((room.bottom + room.top) / 2));
            this.addWorldObjectToTile(location, type);
        };
        MapGenerator.prototype.addRandomWorldObjectToRoom = function (room) {
            var special = this.tryAddSpecialReward(room);
            if (special) {
                return;
            }
            var location = new RogueRpg.Location(Math.floor((room.right + room.left) / 2), Math.floor((room.bottom + room.top) / 2));
            this.addRandomWorldObjectToTile(location);
        };
        MapGenerator.prototype.addRandomWorldObjectToTile = function (location) {
            var specialRoll = Math.random(); // Not affected by luck - thats too OP
            if (specialRoll > 0.96) {
                this.addWorldObjectToTile(location, "Item Upgrade Shrine");
            }
            else if (specialRoll > 0.72) {
                this.addWorldObjectToTile(location, "Treasure Chest");
            }
            else if (specialRoll > 0.58) {
                this.addWorldObjectToTile(location, "Healing Shrine");
            }
            else if (specialRoll > 0.55) {
                this.addWorldObjectToTile(location, "Sitting Wizard");
            }
            else if (specialRoll > 0.4) {
                this.spawnRandomPotionAt(location.x, location.y);
            }
            else if (specialRoll > 0.36) {
                this.addWorldObjectToTile(location, "Ghost Girl");
            }
            else if (specialRoll > 0.1) {
                // this.addWorldObjectToTile(location, "Alchemy Table");
                // TODO: Not releasing alchemy
            }
        };
        MapGenerator.prototype.spawnRandomPotionAt = function (tileX, tileY) {
            var potion = RogueRpg.Game.randomPotion();
            var dp = new RogueRpg.DroppedPotion(potion, new RogueRpg.Location((tileX + 0.5) * this._mapInProgress.tileSize - 15, (tileY + 0.5) * this._mapInProgress.tileSize - 15));
            this._mapInProgress.potionsOnGround.push(dp);
        };
        MapGenerator.AvailableTilesets = ["Mine", "Orc Cave", "Ice Tunnel", "Small", "Yellow Brick",
            "Cave", "Cosmic", "Dungeon"];
        MapGenerator.Tilesets = {
            "Mine": { name: "Mine", size: 130, enemies: 35, decors: ["Mine Wall Pickaxe", "Mine Wall Torch"] },
            "Orc Cave": { name: "Orc Cave", size: 130, enemies: 35, decors: ["Mine Wall Torch"] },
            "Ice Tunnel": { name: "Ice Tunnel", size: 60, enemies: 20 },
            "Small": { name: "Small", size: 80, enemies: 30 },
            "Yellow Brick": { name: "Yellow Brick", size: 120, enemies: 40 },
            "Cave": { name: "Cave", size: 130, enemies: 30 },
            "Cosmic": { name: "Cosmic", size: 100, enemies: 30 },
            "Dungeon": { name: "Dungeon", size: 160, enemies: 50 },
            "Magma": { name: "Magma", size: 130, enemies: 35 },
            "Earth": { name: "Earth", size: 130, enemies: 35 }
        };
        MapGenerator.addedRelicThisStage = false;
        MapGenerator.addedClassThisStage = false;
        MapGenerator.itemsForTiles = {}; // Associative Array
        MapGenerator.RoomTypeEnemies = [
            "Empty",
            "Normal",
            "Dense",
            "StrongSingle",
            "WeakHorde"
        ];
        MapGenerator.RoomTypeReward = [
            "Empty",
            "Healing",
            "Shrine",
            "Treasure",
            "TreasureBig"
        ];
        MapGenerator.UpRoomOptions = { minWidth: 2, maxWidth: 3, minHeight: 3, maxHeight: 4 };
        MapGenerator.UpRoomOptionsNarrowing = { minWidth: 3, maxWidth: 4, minHeight: 3, maxHeight: 4 };
        return MapGenerator;
    }());
    RogueRpg.MapGenerator = MapGenerator;
    function FindRoomLocation(room) {
        return new RogueRpg.Location(Math.floor((room.right + room.left) / 2), Math.floor((room.bottom + room.top) / 2));
    }
})(RogueRpg || (RogueRpg = {}));
function roll(percent) {
    return Math.random() * 100 < percent;
}
//# sourceMappingURL=MapGenerator.js.map