/// <reference path="game.ts" />
var RogueRpg;
(function (RogueRpg) {
    "use strict";
    // The Panels class handles UI based on HTML elements
    var Panels = /** @class */ (function () {
        function Panels() {
        }
        Panels.showSecondaryConfirmPrompt = function (message, confirmText, confirmCallback, cancelText, cancelCallback) {
            by.id('extraDetailsDiv2').style.display = "block";
            by.id("extraDetailsText2").innerHTML = message;
            by.id("extraDetailsCloseButton2").innerText = cancelText || "返回";
            by.id("extraDetailsConfirmButton2").innerText = confirmText;
            by.id('extraDetailsCloseButton2').onclick = function () {
                by.id('extraDetailsDiv2').style.display = "none";
                if (cancelCallback) {
                    cancelCallback();
                }
            };
            by.id('extraDetailsConfirmButton2').onclick = function () {
                by.id("extraDetailsDiv2").style.display = "none";
                confirmCallback();
            };
        };
        return Panels;
    }());
    RogueRpg.Panels = Panels;
    var ListItem = /** @class */ (function () {
        function ListItem() {
        }
        ListItem.buildPrisma = function (container, name, description, cost, callback) {
            return ListItem.build(container, undefined, name, description, { prismaPrice: cost, prismaCallback: callback });
        };
        ListItem.buildGold = function (container, name, description, cost, callback) {
            return ListItem.build(container, undefined, name, description, { goldPrice: cost, goldCallback: callback });
        };
        // if price is present, displays an upgrade button
        ListItem.build = function (container, imgName, name, description, cc) {
            container.innerText = '';
            container.classList.add('listItem');
            if (imgName) {
                var imgHolder = div('', 'listItemImageHolder');
                imgHolder.appendChild(img(imgName, undefined, 'listItemImage'));
                container.appendChild(imgHolder);
            }
            var holder = div(undefined, 'listItemTextHolder');
            container.appendChild(holder);
            var title = div(undefined, 'listItemName');
            title.innerText = name;
            holder.appendChild(title);
            var descriptionDiv = div(undefined, 'listItemDescription');
            descriptionDiv.innerText = description;
            holder.appendChild(descriptionDiv);
            var buttonContainer = div(undefined, 'listItemButtonHolder');
            container.appendChild(buttonContainer);
            if (cc.shardPrice && cc.shardCallback) {
                var button = div(undefined, 'listItemButton');
                buttonContainer.appendChild(button);
                button.classList.add('listItemShardButton');
                if ((cc.useStoredShards && RogueRpg.Helpers.game.storedShards < cc.shardPrice) ||
                    (!cc.useStoredShards && RogueRpg.Helpers.game.shards < cc.shardPrice)) {
                    button.classList.add('cannotAfford');
                }
                button.appendChild(span('', 'shard-count'));
                button.appendChild(span(RogueRpg.Helpers.formatNumber(cc.shardPrice)));
                button.onclick = cc.shardCallback;
                if (cc.gemstonePrice && cc.gemstoneCallback) {
                    buttonContainer.appendChild(document.createElement("br"));
                    container.classList.add('listItemLarge');
                    buttonContainer.appendChild(div("", "listItemOrText", "or"));
                }
            }
            if (cc.gemstonePrice && cc.gemstoneCallback) {
                var button = div(undefined, 'listItemButton');
                buttonContainer.appendChild(button);
                button.classList.add('listItemGemstoneButton');
                if (RogueRpg.Helpers.game.gemstones < cc.gemstonePrice) {
                    button.classList.add('cannotAfford');
                }
                button.appendChild(span('', 'gemstone-count'));
                button.appendChild(span("" + cc.gemstonePrice));
                button.onclick = cc.gemstoneCallback;
            }
            if (cc.prismaPrice && cc.prismaCallback) {
                var button = div(undefined, 'listItemButton');
                buttonContainer.appendChild(button);
                button.classList.add('listItemPrismaButton');
                if (RogueRpg.Helpers.game.prismaState.prisma < cc.prismaPrice) {
                    button.classList.add('cannotAfford');
                }
                button.appendChild(span('', 'prisma-count'));
                button.appendChild(span(RogueRpg.Helpers.formatNumber(cc.prismaPrice)));
                button.onclick = cc.prismaCallback;
            }
            if (cc.goldPrice && cc.goldCallback) {
                var button = div(undefined, 'listItemButton');
                buttonContainer.appendChild(button);
                button.classList.add('listItemGoldButton');
                if (RogueRpg.Helpers.game.prismaState.gold < cc.goldPrice) {
                    button.classList.add('cannotAfford');
                }
                button.appendChild(span('', 'gold-count'));
                button.appendChild(span(RogueRpg.Helpers.formatNumber(cc.goldPrice)));
                button.onclick = cc.goldCallback;
            }
            if (cc.iapPrice && cc.iapCallback) {
                var button = div(undefined, 'listItemButton');
                buttonContainer.appendChild(button);
                button.classList.add('listItemIapButton');
                button.appendChild(span(cc.iapPrice));
                button.onclick = cc.iapCallback;
            }
            return container;
        };
        return ListItem;
    }());
    RogueRpg.ListItem = ListItem;
    var InGameStoreListItem = /** @class */ (function () {
        function InGameStoreListItem(item, callback, duringGame) {
            this.item = item;
            this.callback = callback;
            this.duringGame = duringGame;
            this.container = div(undefined, 'inGameStoreListItem');
            this.initialize();
        }
        InGameStoreListItem.prototype.initialize = function () {
            var _this = this;
            var shardCallback = function () {
                if (RogueRpg.Helpers.game.shards >= _this.item.shardsPrice && _this.item.quantity > 0) {
                    RogueRpg.Helpers.game.itemBought(_this.item);
                    RogueRpg.Helpers.game.shards -= _this.item.shardsPrice;
                    _this.item.quantity--;
                    _this.callback();
                    _this.initialize();
                }
            };
            var gemstoneCallback = function () {
                if (RogueRpg.Helpers.game.gemstones >= _this.item.gemstonePrice && _this.item.quantity > 0) {
                    RogueRpg.Helpers.game.itemBought(_this.item);
                    RogueRpg.Helpers.game.gemstones -= _this.item.gemstonePrice;
                    _this.item.quantity--;
                    _this.callback();
                    _this.initialize();
                }
            };
            var listItem = ListItem.build(this.container, this.item.imageName, this.item.name, "剩余: " + this.item.quantity, {
                shardPrice: this.item.shardsPrice, shardCallback: shardCallback, useStoredShards: !this.duringGame, gemstonePrice: this.item.gemstonePrice, gemstoneCallback: gemstoneCallback
            });
            if (this.item.quantity === 0) {
                this.container.classList.add("unowned");
            }
            else {
                this.container.classList.remove("unowned");
            }
        };
        return InGameStoreListItem;
    }());
    RogueRpg.InGameStoreListItem = InGameStoreListItem;
    var UpgradeListItem = /** @class */ (function () {
        function UpgradeListItem(upgrade, callback) {
            this.upgrade = upgrade;
            this.callback = callback;
            this.container = div(undefined, 'upgradeListItem');
            this.initialize();
        }
        UpgradeListItem.prototype.initialize = function () {
            var _this = this;
            var currentRank = RogueRpg.Helpers.game.prismaState.levels[this.upgrade.id] || 0;
            var maxRank = this.upgrade.ranks;
            var description = this.upgrade.description
                .replace("{0}", (currentRank * this.upgrade.gainPerRank) + '')
                .replace("{1}", '' + this.upgrade.gainPerRank) + "\n" +
                " [" + currentRank + " / " + this.upgrade.ranks + "]\n";
            var cost = this.upgrade.initialCost + currentRank * this.upgrade.costPerRank;
            var name = this.upgrade.name;
            if (this.upgrade.resource === "prisma") {
                var callback = function () {
                    if (cost <= RogueRpg.Helpers.game.prismaState.prisma) {
                        RogueRpg.Helpers.game.prismaState.prisma -= cost;
                        RogueRpg.Helpers.game.prismaState.levels[_this.upgrade.id] =
                            (RogueRpg.Helpers.game.prismaState.levels[_this.upgrade.id] || 0) + 1;
                        _this.initialize();
                        _this.callback();
                        RogueRpg.Helpers.game.persistData();
                    }
                };
                if (currentRank >= maxRank) {
                    cost = null;
                    callback = null;
                }
                var listItem = ListItem.buildPrisma(this.container, name, description, cost, callback);
            }
            else if (this.upgrade.resource === "gold") {
                var callback = function () {
                    if (cost <= RogueRpg.Helpers.game.prismaState.gold) {
                        RogueRpg.Helpers.game.prismaState.gold -= cost;
                        RogueRpg.Helpers.game.prismaState.levels[_this.upgrade.id] =
                            (RogueRpg.Helpers.game.prismaState.levels[_this.upgrade.id] || 0) + 1;
                        if (_this.upgrade.id === RogueRpg.PermanentUpgrade.Upgrades.BuyPrisma.id) {
                            RogueRpg.Helpers.game.prismaState.prisma += RogueRpg.PermanentUpgrade.Upgrades.BuyPrisma.gainPerRank;
                        }
                        _this.initialize();
                        _this.callback();
                        RogueRpg.Helpers.game.persistData();
                    }
                };
                if (currentRank >= maxRank) {
                    cost = null;
                    callback = null;
                }
                var listItem = ListItem.buildGold(this.container, name, description, cost, callback);
            }
            else if (this.upgrade.resource === "iap" && REAL_MONEY_ENABLED) {
                //alert("Test: found IAP. Browsing products: " + Helpers.game.upgradeProducts.join(","))
                for (var i = 0; i < RogueRpg.Helpers.game.upgradeProducts.length; i++) {
                    var product = RogueRpg.Helpers.game.upgradeProducts[i];
                    if (product.productId === this.upgrade.inAppPurchaseId) {
                        var callback = function () {
                            inAppPurchase.buy(product.productId)
                                .then(function (data) {
                                RogueRpg.Helpers.game.activatePurchase(_this.upgrade.inAppPurchaseId);
                                _this.initialize();
                                _this.callback();
                                RogueRpg.Helpers.game.persistData();
                            })
                                .catch(function (err) {
                                //alert("Error making purchase:" + err);    
                            });
                        };
                        // TODO: Also disable if you've already bought the most expensive of the above
                        if (currentRank >= maxRank) {
                            cost = null;
                            callback = null;
                        }
                        //alert("Found product " + product.productId + ". Building now");
                        var listItem = ListItem.build(this.container, null, product.title, product.description, { iapPrice: product.price, iapCallback: callback });
                    }
                }
            }
            else {
                //alert("Type not found: " + this.upgrade.resource + " or RME: " + REAL_MONEY_ENABLED)
            }
        };
        return UpgradeListItem;
    }());
    RogueRpg.UpgradeListItem = UpgradeListItem;
    var RelicListItem = /** @class */ (function () {
        function RelicListItem(index, canUpgrade, callback) {
            this.index = index;
            this.canUpgrade = canUpgrade;
            this.callback = callback;
            this.container = div(undefined, 'relicListItem');
            this.initialize();
        }
        RelicListItem.prototype.initialize = function () {
            var _this = this;
            var relic = RogueRpg.Relic.Relics[this.index];
            var relicLevel = RogueRpg.Relic.RelicLevels[this.index];
            var owned = relicLevel > 0;
            var price = owned ? RogueRpg.Relic.getExpNeededAtLevel(relicLevel, relic) : relic.price;
            var text = "Level " + relicLevel + ": ";
            // There is only one
            for (var name in relic.statsPerLevel) {
                var sign = name === "percent_cooldowns" ? '-' : '+';
                var str = RogueRpg.Relic.StatToName(name);
                text += sign + relic.statsPerLevel[name] * relicLevel + (name.indexOf("percent_") == 0 ? "%" : "") + " " + str;
            }
            var callback = function () {
                if (_this.canUpgrade) {
                    if (RogueRpg.Helpers.game.storedShards >= price) {
                        RogueRpg.Helpers.game.storedShards -= price;
                        RogueRpg.Relic.RelicLevels[_this.index]++;
                        _this.initialize();
                        _this.callback();
                        RogueRpg.Helpers.game.persistData();
                    }
                }
            };
            var listItem = ListItem.build(this.container, relic.name, relic.name, text, { shardPrice: this.canUpgrade ? price : 0, shardCallback: callback, useStoredShards: this.canUpgrade });
            this.container = listItem;
            if (relicLevel === 0) {
                this.container.classList.add("unowned");
            }
            else {
                this.container.classList.remove("unowned");
            }
        };
        return RelicListItem;
    }());
    RogueRpg.RelicListItem = RelicListItem;
})(RogueRpg || (RogueRpg = {}));
//# sourceMappingURL=Panels.js.map