/// <reference path="GameUtils.ts" />
/// <reference path="MapGenerator.ts" />
/// <reference path="Levels.ts" />
/// <reference path="Items.ts" />
var inAppPurchase = window.inAppPurchase; // in app purchase plugin
var REAL_MONEY_ENABLED = true;
var ENABLE_ART_REVIEW = false;
var RogueRpg;
(function (RogueRpg) {
    "use strict";
    var LevelPointState = /** @class */ (function () {
        function LevelPointState() {
        }
        return LevelPointState;
    }());
    RogueRpg.LevelPointState = LevelPointState;
    var Game = /** @class */ (function () {
        function Game() {
            this.callbacks = [];
            this.goldAccumulation = 0;
            this.storedShards = 0;
            this._mouseDown = false;
            this._mouseX = -1;
            this._mouseY = -1;
            this._tappedThisCycle = true; // True for one frame - to pick up items and attack monsters
            this._doubleTappedThisCycle = false;
            this._doubleTapLoopTimer = 0;
            this.activeAnimations = [];
            this.spriteAnimations = [];
            this.bridgedAnimations = [];
            this.floatingText = [];
            this.losingLifeChunks = [];
            this.projectiles = [];
            this.flyingKickCount = 0;
            this.flyingKickTimer = 0;
            this.maxHealingPotionCount = 3;
            this.movingCyclesSinceLastAttack = 0;
            this.movedThisCycle = false;
            this.needsToPersist = false;
            this.lastDestroyedItem = null;
            // TODO: Cleanup inventory seems to work. Need to add the flying-clear system later
            // Fly around the level killing all enemies and collecting all loot
            this.clearingCurrentLevel = false;
            this.clearingCurrentLevelAngle = 0;
            this.maxMultiplierUnlocked = 1;
            this.savedValues = {};
            this.loopPaused = false;
            this.previousMapStack = [];
            this.classSelectionPaneVisible = false;
            this.levelUpPaneVisible = false;
            this.ratePromptHasBeenShown = false;
            this.exceptionLogged = false;
            this.startTime = Date.now();
            this.deltaTimes = [];
            this.loopTimes = [];
            this.loopAverage = 0;
            this.average = 0;
            this.recordTiming = false;
            this.tutorialState = 0;
            this.lockHeroMovement = false;
            this.tutorialCompleted = false;
            this.tutorialGoblins = []; // TODO: Persist. What if you quit during the tutorial?
            // TODO: persist these
            this.lastCompletedGameTime = null;
            this.bonusShardsThisGame = false;
            this.advancingLevel = false;
            // Called instead of gameLoop
            this.white = 0;
            this.originalRelicLevel = 1;
            this.relicAnimationPhase = "";
            this.relicStateTimer = 0;
            this.shardGainShowing = false;
            this.bonusShards = 0;
            this.movingShards = [];
            this.highestClassUnlocked = 0;
            this.lastRelicIndex = 0;
            this.shardTimer = 0;
            this.shownRelicPrompt = false;
            this.gemstoneSectionDown = false;
            this.shardSectionDown = false;
            this.inventoryAccessDown = false;
            this.healingPotionButtonDown = false;
        }
        Game.prototype.setGemstones = function (count) {
            this.gemstones = count;
            this.renderer.renderGemstones = true;
        };
        Game.prototype.setShards = function (count) {
            this.shards = count;
            this.renderer.renderShardsSection = true;
        };
        Object.defineProperty(Game.prototype, "floor", {
            get: function () {
                return this.hero.floor;
            },
            enumerable: true,
            configurable: true
        });
        Game.prototype.initialize = function () {
            var _this = this;
            //ga_storage._setAccount('UA-67414544-1'); //Replace with your own
            //ga_storage._trackPageview('/index.html');
            RogueRpg.Helpers.game = this;
            this.renderer = new RogueRpg.Renderer(this);
            RogueRpg.Helpers.renderer = this.renderer;
            if (REAL_MONEY_ENABLED) {
                try {
                    var ids = [];
                    for (var i = 0; i < RogueRpg.InAppPurchase.Purchases.length; i++) {
                        var p = RogueRpg.InAppPurchase.Purchases[i];
                        ids.push(p.id);
                    }
                    if (inAppPurchase) {
                        inAppPurchase.getProducts(ids).then(function (products) {
                            _this.gemstoneProducts = products;
                            var upgradePurchaseIds = ["unlock_auto_items"];
                            inAppPurchase.getProducts(upgradePurchaseIds).then(function (products) {
                                //alert("onload, tried to get unlock_auto_items: " + JSON.stringify(products));
                                _this.upgradeProducts = products;
                                inAppPurchase.restorePurchases().then(function (items) {
                                    //alert("restored purchases: " + JSON.stringify(items));
                                    try {
                                        for (var i = 0; i < items.length; i++) {
                                            for (var j = 0; j < RogueRpg.InAppPurchase.Purchases.length; j++) {
                                                var p = RogueRpg.InAppPurchase.Purchases[j];
                                                if (p.id === items[i].productId) {
                                                    _this.gemstones += p.baseGems;
                                                    inAppPurchase.consume(items[i].productType, items[i].receipt, items[i].signature);
                                                    //alert("A purchase completed after the game closed. Restoring purchase of " + p.baseGems + " gemstones");
                                                    continue;
                                                }
                                            }
                                            for (var j = 0; j < upgradePurchaseIds.length; j++) {
                                                if (items[i].productId === upgradePurchaseIds[j]) {
                                                    //alert("restoring IAP: " + items[i].productId);
                                                    _this.activatePurchase(upgradePurchaseIds[j]);
                                                }
                                            }
                                        }
                                    }
                                    catch (err) {
                                        //alert("Wow error: " + JSON.stringify(err));
                                    }
                                }).catch(function (err) {
                                    REAL_MONEY_ENABLED = false;
                                    //alert("Failed to restore products " + JSON.stringify(err));
                                });
                            }).catch(function (err) {
                                REAL_MONEY_ENABLED = false;
                                //alert("Failed to load auto-unlock product " + JSON.stringify(err));
                            });
                        }).catch(function (err) {
                            REAL_MONEY_ENABLED = false;
                            //alert("Error loading in-app purchases: " + JSON.stringify(err));
                        });
                    }
                }
                catch (ex) {
                    REAL_MONEY_ENABLED = false;
                    // alert("Error trying to load in-app purchases: " + ex.message);
                }
            }
            var levelUpNames = Object.getOwnPropertyNames(RogueRpg.L);
            for (var i = 0; i < levelUpNames.length; i++) {
                RogueRpg.Level[RogueRpg.L[levelUpNames[i]]] = 0; // set all to 0, later on values can be added by persistence
            }
            this.loadOrCreatePersistedState();
            this.screenCaptureAfterDeath = by.id("screenshotCanvas");
            this.screenCaptureAfterDeath.height = this.renderer.wh;
            this.screenCaptureAfterDeath.width = this.renderer.ww;
            this._hideItemDisplayHandler = this.hideItemDisplay.bind(this);
            this._hideExtraDetailsButtonHandler = this.hideExtraDetails.bind(this);
            // this._hideHatInfoHandler = this.hideHatInfo.bind(this);
            by.id("extraDetailsCloseButton").onclick = function () { return _this._hideExtraDetailsButtonHandler(); };
            //this.renderer.canvas.addEventListener("mousedown", this.onMouseDown.bind(this));
            //this.renderer.canvas.addEventListener("mouseup", this.onMouseUp.bind(this));
            //this.renderer.canvas.addEventListener("mousemove", this.onMouseMove.bind(this));
            this.renderer.canvas.addEventListener("touchstart", this.onTouchStart.bind(this));
            this.renderer.canvas.addEventListener("touchend", this.onTouchEnd.bind(this));
            this.renderer.canvas.addEventListener("touchmove", this.onTouchMove.bind(this));
            var inventoryButton = by.id("inventoryAccess");
            inventoryButton.onclick = function () { return _this.showInventory(); };
            var inventoryCloseButton = by.id("inventoryClose");
            inventoryCloseButton.onclick = function () { return _this.closeInventory(); };
            var gemstoneCloseButton = by.id("gemstoneCloseButton");
            gemstoneCloseButton.onclick = function () { return _this.hideGemstoneMenu(); };
            var relicListCloseButton = by.id("relicListCloseButton");
            relicListCloseButton.onclick = function () { return _this.hideRelicsOrUpgradesMenu(); };
            var upgradeListCloseButton = by.id("upgradeListCloseButton");
            upgradeListCloseButton.onclick = function () { return _this.hideRelicsOrUpgradesMenu(); };
            var inGameStoreCloseButton = by.id("inGameStoreCloseButton");
            inGameStoreCloseButton.onclick = function () { return _this.hideInGameStore(); };
            var gemstoneInfo = by.id("gemstoneInfo");
            var levelInfoCloseButton = by.id("levelInfoCloseButton");
            levelInfoCloseButton.addEventListener("click", function () { return _this.hideLevelInfoPane(); });
            var alchemyCloseButton = by.id("alchemyCloseButton");
            alchemyCloseButton.addEventListener("click", function () { return _this.hideAlchemyPane(); });
            var levelUpInfoAccessButton = by.id("levelUpInfoAccessButton");
            levelUpInfoAccessButton.addEventListener("click", function () { return _this.showLevelInfoPane(); });
            var restartButton = by.id("restartButton");
            restartButton.addEventListener("click", function () {
                if (!_this.tutorialCompleted)
                    return;
                by.id("reallyRestartButton").style.display = "inline-block";
                setTimeout(function () {
                    by.id("reallyRestartButton").style.display = "none";
                }, 1000);
            });
            var reallyRestartButton = by.id("reallyRestartButton");
            reallyRestartButton.addEventListener("click", function () { return _this.handleHeroDeath(RogueRpg.Monster.ResetButtonMonster); });
            var settingsButton = by.id("settingsButton");
            settingsButton.onclick = function () {
                _this.showSettingsPane();
            };
            var settingsCloseButton = by.id("settingsCloseButton");
            settingsCloseButton.onclick = function () {
                _this.closeSettingsPane();
            };
            var dprButton = by.id("dprButton");
            if (dprButton) {
                dprButton.innerText = "Dpr: " + this.renderer.dpr;
                dprButton.onclick = function () {
                    _this.renderer.changeDpr();
                    dprButton.innerText = "Dpr: " + _this.renderer.dpr;
                };
            }
            var aaButton = by.id("aaButton");
            if (aaButton) {
                aaButton.innerText = this.renderer.antialias ? "AA: On" : "AA: Off";
                aaButton.onclick = function () {
                    _this.renderer.changeAA();
                    aaButton.innerText = _this.renderer.antialias ? "AA: On" : "AA: Off";
                    _this.persistData();
                };
            }
            var graphicsButton = by.id('graphicsButton');
            if (graphicsButton) {
                graphicsButton.innerText = "Gfx: " + (this.renderer.useSimpleGraphics ? "low" : "high");
                graphicsButton.onclick = function () {
                    _this.renderer.useSimpleGraphics = !_this.renderer.useSimpleGraphics;
                    _this.renderer.setUseSimpleGraphics(_this.renderer.useSimpleGraphics);
                    graphicsButton.innerText = "Gfx: " + (_this.renderer.useSimpleGraphics ? "low" : "high");
                    _this.persistData();
                };
            }
            var damageTextButton = by.id("damageTextButton");
            if (damageTextButton) {
                damageTextButton.innerText = "伤害: " + (this.renderer.showDamageText ? "show" : "hide");
                damageTextButton.onclick = function () {
                    _this.renderer.showDamageText = !_this.renderer.showDamageText;
                    damageTextButton.innerText = "伤害: " + (_this.renderer.showDamageText ? "show" : "hide");
                    _this.persistData();
                };
            }
            // TODO: This can be removed after 1.0.9, it's a temporary measure to make sure all monsters on the 
            // loaded game have an attackRange, which is a new property
            if (this.map && this.map.monsters) {
                for (var i = 0; i < this.map.monsters.length; i++) {
                    if (!this.map.monsters[i].attackRange) {
                        this.map.monsters[i].attackRange = 50;
                    }
                }
            }
            this.activeAnimations = [];
            this.spriteAnimations = [];
            this.floatingText = [];
            this._mouseDown = false;
            this._tappedThisCycle = false;
            // this.particleEmitter = new ParticleEmitter({ image: "sparkle", timeRemaining: 500 }, 100, this.heroWeaponCenter);
            // this.particleEmitter = new ParticleEmitter({ image: "PurpleCloud", timeRemaining: 800 }, 100, this.heroWeaponCenter);
            var emitter = new RogueRpg.ParticleEmitter({ image: "TinySparkle", timeRemaining: 500 }, 40, function () {
                var img = RogueRpg.Images.get(_this.hero.equippedWeapon.image);
                var canvas = RogueRpg.ImageToCanvas(img);
                var loc = new RogueRpg.Location(_this.hero.x + (_this.hero.facingRight ? 22 : -22), _this.hero.y - 50);
                for (var i = 0; i < 1000; i++) {
                    var x = randIntMap(0, img.width - 1);
                    var y = randIntMap(0, img.height - 1);
                    if (canvas.getContext("2d").getImageData(x, y, 1, 1).data[3] === 255) {
                        loc = new RogueRpg.Location(_this.hero.x + (60 - x) * (_this.hero.facingRight ? 1 : -1), _this.hero.y - 85 + y);
                        break;
                    }
                }
                return loc;
            }, function () {
                return _this.hero.equippedWeapon && _this.hero.equippedWeapon.rarity > 6;
            });
            var slowEmitter = new RogueRpg.ParticleEmitter({ animationKey: "Slow" }, 150, function () {
                return new RogueRpg.Location(_this.hero.x, _this.hero.y);
            }, function () { return false; });
            RogueRpg.ParticleEmitter.Emitters.push(slowEmitter);
            RogueRpg.ParticleEmitter.Emitters.push(emitter);
        };
        // Given a permanent in-app-purchase, activate it's results.
        Game.prototype.activatePurchase = function (id) {
            if (id === "unlock_auto_items") {
                RogueRpg.Helpers.game.prismaState.levels[RogueRpg.PermanentUpgrade.Upgrades.UnlockAuto.id] = 1;
                RogueRpg.Helpers.game.prismaState.levels[RogueRpg.PermanentUpgrade.Upgrades.AutoOpenChests.id] = 1;
                RogueRpg.Helpers.game.prismaState.levels[RogueRpg.PermanentUpgrade.Upgrades.AutoItemSell.id] = 1;
                RogueRpg.Helpers.game.prismaState.levels[RogueRpg.PermanentUpgrade.Upgrades.AutoItemPickup.id] = 1;
            }
        };
        // Generates a random number using the player's luck value
        // Rolls an average of 1 + luck/100 dice to return the max random number
        Game.prototype.playerRandom = function (itemRelated, limitLuckTo) {
            if (itemRelated && RogueRpg.Level[RogueRpg.L.Greedless]) {
                return Math.random();
            }
            var luck = Math.abs(this.hero.luck);
            if (typeof limitLuckTo === "number") {
                luck = Math.min(luck, limitLuckTo);
            }
            var rands = [Math.random()];
            for (var i = 100; i <= luck; i += 100) {
                rands.push(Math.random());
            }
            if (Math.random() < (luck % 100) / 100.0) {
                rands.push(Math.random());
            }
            if (this.hero.luck >= 0) {
                return Math.max.apply(null, rands);
            }
            else {
                return Math.min.apply(null, rands);
            }
        };
        Game.prototype.expectedValueFromCurrentLuck = function () {
            var luck = Math.abs(this.hero.luck);
            var delta = 0.5 - (1 / (2 + luck / 100));
            if (this.hero.luck >= 0) {
                return 0.5 + delta;
            }
            else {
                return 0.5 - delta;
            }
        };
        Game.prototype.getUnlockableCounts = function (mode) {
            var count = 0;
            var max = 0;
            for (var i = 0; i < RogueRpg.Relic.Unlockables.length; i++) {
                var u = RogueRpg.Relic.Unlockables[i];
                if (u.difficulty === mode) {
                    max++;
                    if (RogueRpg.Relic.RelicLevels[u.relicIndex] > 0) {
                        count++;
                    }
                }
            }
            switch (mode) {
                case 1:
                    max += 3; // Classes
                    count += Math.min(this.highestClassUnlocked, 3);
                    break;
                case 2:
                    max += 2; // Potion Slot, Starting Weapon
                    if (RogueRpg.Helpers.game.maxPotions >= 6)
                        count++;
                    if (!!RogueRpg.Helpers.game.startingWeapon)
                        count++;
                    break;
                case 4:
                    max += 2; // Potion Slot, Starting Weapon
                    if (RogueRpg.Helpers.game.maxPotions >= 7)
                        count++;
                    if (!!RogueRpg.Helpers.game.startingWeapon)
                        count++;
                    break;
                case 8:
                    max += 1; // Potion Slot
                    if (RogueRpg.Helpers.game.maxPotions >= 8)
                        count++;
                    break;
                case 16:
                    break;
            }
            return { count: count, max: max };
        };
        // Goes through all your items, keeping only the strongest ones you own
        Game.prototype.cleanupInventory = function () {
            var topItems = [this.hero.equippedWeapon, this.hero.equippedShield, this.hero.equippedArmor, this.hero.equippedAccessory];
            for (var i = 0; i < this.hero.items.length; i++) {
                var item = this.hero.items[i];
                var j = Game.indexForType(item);
                if (this.isSecondItemBetter(topItems[j], item)) {
                    topItems[j] = this.hero.items[i];
                }
                else {
                    this.hero.items.splice(i, 1);
                    i--;
                }
            }
        };
        Game.prototype.getItemSellValue = function (item) {
            var value = item.level / 2;
            if (item instanceof RogueRpg.Weapon) {
                var w = item;
                value += w._baseDamage / w._baseAttackSpeed * 1000;
            }
            else if (item instanceof RogueRpg.Armor) {
                var a = item;
                value += a.baseDefense;
            }
            else if (item instanceof RogueRpg.Shield) {
                var s = item;
                value += s.baseReduction;
            }
            else if (item instanceof RogueRpg.Accessory) {
                var r = item;
                value += 5 + r.level + r.attacksPerSecond * 2;
            }
            // This shouldn't ever go past 7 but whatever
            // n/a, white, blue, yellow, orange, green, black, purple
            var tierMap = [0.1, 0.4, 1, 1.6, 2.8, 5, 10, 19, 25, 27, 28, 29, 30];
            var tierMultiplier = tierMap[item.rarity];
            value *= tierMultiplier;
            if (item.fixedFullName) {
                value = value * 1.4;
            }
            value += randInt(1, 5);
            value = value * RogueRpg.Helpers.game.hero.state.multiplier;
            return Math.ceil(value);
        };
        Game.prototype.isSecondItemBetter = function (first, second) {
            if (!first) {
                return true;
            }
            var e1 = this.calculateItemEffect(first);
            var e2 = this.calculateItemEffect(second);
            return (e2.defRatio + e2.dpsRatio > e1.defRatio + e2.dpsRatio);
        };
        Game.indexForType = function (item) {
            if (item instanceof RogueRpg.Weapon) {
                return 0;
            }
            else if (item instanceof RogueRpg.Shield) {
                return 1;
            }
            else if (item instanceof RogueRpg.Armor) {
                return 2;
            }
            else {
                return 3;
            }
        };
        Game.prototype.clearCurrentLevel = function () {
            this.clearingCurrentLevel = true;
            this.clearedChests = false;
            for (var i = 0; i < this.map.specialWorldObjects.length; i++) {
                if (this.map.specialWorldObjects[i].type == "Level Exit") {
                    this.currentMapExitCrystal = this.map.specialWorldObjects[i];
                }
            }
            this.sortMonstersByDistance();
        };
        Game.prototype.updateCycleClearCurrentLevel = function () {
            if (!this.currentMapExitCrystal.x) {
                this.currentMapExitCrystal.x = this.currentMapExitCrystal.drawX + this.currentMapExitCrystal.drawWidth / 2;
            }
            if (!this.currentMapExitCrystal.y) {
                this.currentMapExitCrystal.y = this.currentMapExitCrystal.drawY + this.currentMapExitCrystal.drawHeight;
            }
            if (this.map.gemstonesOnGround.length > 0) {
                var g = this.map.gemstonesOnGround[0];
                this.map.gemstonesOnGround.splice(0, 1);
                this.gemstones += g.value;
                this.renderer.resetStaticContext();
            }
            else if (this.map.itemsOnGround.length > 0) {
                this.pickupItem(0);
            }
            else if (this.map.monsters.length > 0) {
                // go to the first monster
                var mon = this.map.monsters[0];
                if (RogueRpg.Helpers.dist(this.hero, mon) < 150) {
                    this.damageMonster(mon, mon.maxLife * 2, true);
                    this.bridgedAnimations.push(new RogueRpg.BridgedImage(Math.random() > 0.5 ? "LightningWhite" : "LightningPurple", this.hero.x, this.hero.y - this.hero.drawHeight / 2, mon.x, mon.y, 700));
                    this.sortMonstersByDistance();
                }
                else {
                    this.clearingCurrentLevelAngle = RogueRpg.Helpers.angle(this.hero, mon);
                    this.moveUnitTo(this.hero, mon.x, mon.y, 300, true);
                }
            }
            else if (RogueRpg.Helpers.dist(this.hero, this.currentMapExitCrystal) > 1500) {
                this.moveUnitTo(this.hero, this.currentMapExitCrystal.x, this.currentMapExitCrystal.y, 300, true);
                this.clearingCurrentLevelAngle = RogueRpg.Helpers.angle(this.hero, this.currentMapExitCrystal);
            }
            else if (RogueRpg.Helpers.dist(this.hero, this.currentMapExitCrystal) > 500) {
                this.moveUnitTo(this.hero, this.currentMapExitCrystal.x, this.currentMapExitCrystal.y, 150, true);
                this.clearingCurrentLevelAngle = RogueRpg.Helpers.angle(this.hero, this.currentMapExitCrystal);
            }
            else if (RogueRpg.Helpers.dist(this.hero, this.currentMapExitCrystal) > 200) {
                this.moveUnitTo(this.hero, this.currentMapExitCrystal.x, this.currentMapExitCrystal.y, 60, true);
                this.clearingCurrentLevelAngle = RogueRpg.Helpers.angle(this.hero, this.currentMapExitCrystal);
            }
            else if (RogueRpg.Helpers.dist(this.hero, this.currentMapExitCrystal) > 10) {
                this.moveUnitTo(this.hero, this.currentMapExitCrystal.x, this.currentMapExitCrystal.y, 10, true);
                this.clearingCurrentLevelAngle = RogueRpg.Helpers.angle(this.hero, this.currentMapExitCrystal);
            }
            else if (!this.clearedChests) {
                for (var i = 0; i < this.map.specialWorldObjects.length; i++) {
                    var ob = this.map.specialWorldObjects[i];
                    if (ob.type == "Treasure Chest" || ob.type == "Dark Treasure Chest") {
                        this.openChest(ob);
                    }
                }
                var count = this.map.potionsOnGround.length;
                while (this.map.potionsOnGround.length > 0 && count >= 0) {
                    count--;
                    this.pickupPotion(0);
                }
                this.clearedChests = true;
            }
            else {
                // this.cleanupInventory();
                this.clearingCurrentLevel = false;
                this.healHeroFor(this.hero.maxLife);
                if (!RogueRpg.Helpers.IsWalkable(new RogueRpg.Location(this.hero.x, this.hero.y), this.map)) {
                    this.hero.x = this.currentMapExitCrystal.x;
                    this.hero.y = this.currentMapExitCrystal.y - 5;
                }
                // Make the map fully visible on this floor
                for (var i = 0; i < this.map.mapWidth; i++) {
                    for (var j = 0; j < this.map.mapHeight; j++) {
                        if (this.map.tiles[i] && this.map.tiles[i][j]) {
                            this.map.tiles[i][j].seen = this.map.tiles[i][j].seen || this.map.tiles[i][j].walkable;
                        }
                    }
                }
                this.needsToPersist = true;
                this.renderer.resetStaticContext();
            }
        };
        Game.prototype.findNearestUnit = function (from, list, maxDist) {
            if (maxDist === void 0) { maxDist = 10000000; }
            var nearest = null;
            var lowestDistance = maxDist;
            for (var i = 0; i < list.length; i++) {
                var u = list[i];
                var dist = RogueRpg.Helpers.dist(from, u);
                if (dist < lowestDistance && dist <= maxDist) {
                    lowestDistance = dist;
                    nearest = u;
                }
            }
            return nearest;
        };
        Game.prototype.sortMonstersByDistance = function () {
            var _this = this;
            this.map.monsters.sort(function (a, b) {
                return RogueRpg.Helpers.dist(_this.hero, a) - RogueRpg.Helpers.dist(_this.hero, b);
            });
        };
        Game.prototype.setupNewGameMap = function () {
            this.renderer.resetStaticContext();
            this.map = new RogueRpg.MapGenerator().createLevel(1);
            this.previousMapStack = [];
            this.hero.x = this.map.heroStartX;
            this.hero.y = this.map.heroStartY;
            this.setupInGameStoreItems();
        };
        // There is a fixed set of items available for purchase each game, although i could add
        // extra items at later times
        Game.prototype.setupInGameStoreItems = function () {
            this.inGameStoreItems = [];
            this.inGameStoreItems.push(new RogueRpg.InGameStoreItem("PotionGrey", "英雄药剂", 1, 500, 40));
            this.inGameStoreItems.push(new RogueRpg.InGameStoreItem("PotionRed", "治疗药剂 (40%)", 2, 100, 25));
            this.inGameStoreItems.push(new RogueRpg.InGameStoreItem("Portal", "深林传送门", 1, 300, 30));
            this.inGameStoreItems.push(new RogueRpg.InGameStoreItem("Item Upgrade Shrine", "物品升级圣殿", 1, 900, 70));
            this.inGameStoreItems.push(new RogueRpg.InGameStoreItem("PotionRed", "治疗药剂 (40%)", 2, 600, 25));
            this.inGameStoreItems.push(new RogueRpg.InGameStoreItem("PotionGrey", "英雄药剂", 1, 1500, 40));
            if (this.maxMultiplierUnlocked >= 8) {
                this.inGameStoreItems.push(new RogueRpg.InGameStoreItem("PotionRainbow", "经验药剂 +60", 1, 300, 25));
            }
            else if (this.maxMultiplierUnlocked >= 4) {
                this.inGameStoreItems.push(new RogueRpg.InGameStoreItem("PotionRainbow", "经验药剂 +30", 1, 300, 20));
            }
            else if (this.maxMultiplierUnlocked >= 2) {
                this.inGameStoreItems.push(new RogueRpg.InGameStoreItem("PotionRainbow", "经验药剂 +10", 1, 300, 15));
            }
            if (this.maxMultiplierUnlocked >= 8) {
                this.inGameStoreItems.push(new RogueRpg.InGameStoreItem("PotionGreen", "经验药剂 +80", 1, 500, 35));
            }
            else if (this.maxMultiplierUnlocked >= 4) {
                this.inGameStoreItems.push(new RogueRpg.InGameStoreItem("PotionGreen", "经验药剂 +50", 1, 400, 30));
            }
            else if (this.maxMultiplierUnlocked >= 2) {
                this.inGameStoreItems.push(new RogueRpg.InGameStoreItem("PotionGreen", "经验药剂 +30", 1, 300, 25));
            }
            else {
                this.inGameStoreItems.push(new RogueRpg.InGameStoreItem("PotionGreen", "经验药剂 +20", 1, 200, 20));
            }
            this.inGameStoreItems.push(new RogueRpg.InGameStoreItem("PotionRed", "英雄药剂 (40%)", 1, 1500, 25));
        };
        Game.prototype.loadOrCreatePersistedState = function () {
            var _this = this;
            var heroState;
            var heroStateString = localStorage.getItem("heroState");
            if (heroStateString) {
                try {
                    heroState = JSON.parse(heroStateString);
                }
                catch (ex) { }
            }
            if (heroState) {
                // Keep this here indefinitely for people who upgrade versions
                heroState.multiplier = heroState.multiplier || 1;
                heroState.baseDefense = heroState.baseDefense || 0;
                heroState.baseDamage = heroState.baseDamage || 0;
                heroState.baseReduction = heroState.baseReduction || 0;
                this.hero = new RogueRpg.Hero(heroState);
            }
            else {
                this.hero = new RogueRpg.Hero();
            }
            RogueRpg.Helpers.hero = this.hero;
            var previousMapString = localStorage.getItem("previousMap");
            if (previousMapString) {
                try {
                    var previousMap = JSON.parse(previousMapString);
                    if (!Array.isArray(previousMap)) {
                        if (previousMap) {
                            RogueRpg.Map.HydrateDroppedItems(previousMap);
                            this.previousMapStack = [previousMap];
                        }
                        else {
                            this.previousMapStack = [];
                        }
                    }
                }
                catch (ex) {
                }
            }
            var map;
            var mapString = localStorage.getItem("map");
            if (mapString) {
                try {
                    map = JSON.parse(mapString);
                }
                catch (ex) { }
            }
            if (map) {
                this.map = map;
                if (!this.map.potionsOnGround) {
                    this.map.potionsOnGround = [];
                }
                this.beginLoop();
                if (this.map.monsters) {
                    for (var i = 0; i < this.map.monsters.length; i++) {
                        this.map.monsters[i].activeEffects = this.map.monsters[i].activeEffects || [];
                    }
                }
                // Final challenge for ISM: handle the need to persist dropped items via a State
                RogueRpg.Map.HydrateDroppedItems(this.map);
            }
            // TODO(ItemState Migration): This code needs to remain in place for players who upgrade game versions
            // But at some point before too long I can remove it because no one will really remember what dropped months+ ago
            if (this.hero.state && this.hero.state.itemsOnGround && this.hero.state.itemsOnGround.length > 0 && !Array.isArray(this.map.itemsOnGround)) {
                this.hero.state.itemsOnGround.forEach(function (itemState) {
                    _this.map.itemsOnGround.push(new RogueRpg.DroppedItem(itemState));
                });
            }
            var gemstones = 0;
            var gemstonesString = localStorage.getItem("gemstones");
            if (gemstonesString) {
                try {
                    gemstones = JSON.parse(gemstonesString);
                }
                catch (ex) { }
            }
            this.setGemstones(gemstones || 0);
            var shards;
            var shardsString = localStorage.getItem("shards");
            if (shardsString) {
                try {
                    shards = JSON.parse(shardsString);
                }
                catch (ex) { }
            }
            this.setShards(shards || 0);
            this.storedShards = 0;
            var storedShards = localStorage.getItem("storedShards");
            if (storedShards) {
                try {
                    this.storedShards = JSON.parse(storedShards);
                }
                catch (ex) { }
            }
            var useSimple = localStorage.getItem("useSimpleGraphics");
            if (useSimple === "true") {
                this.renderer.useSimpleGraphics = true;
            }
            var startingWeapon = localStorage.getItem("startingWeapon");
            if (startingWeapon && startingWeapon !== "undefined") {
                this.startingWeapon = startingWeapon;
            }
            var antialias = localStorage.getItem("antialias");
            if (antialias === "true") {
                this.renderer.antialias = true;
            }
            var showDamageText = localStorage.getItem("showDamageText");
            if (showDamageText === "false") {
                this.renderer.showDamageText = false;
            }
            else {
                this.renderer.showDamageText = true;
            }
            var levelUpStateString = localStorage.getItem("LevelUpStates");
            if (levelUpStateString) {
                try {
                    RogueRpg.Level = JSON.parse(levelUpStateString);
                }
                catch (ex) {
                }
            }
            var relicLevelString = localStorage.getItem("relicLevels");
            var relicExpString = localStorage.getItem("relicExp");
            if (relicLevelString) {
                try {
                    RogueRpg.Relic.RelicLevels = JSON.parse(relicLevelString);
                    RogueRpg.Relic.RelicExperience = JSON.parse(relicExpString);
                    if (RogueRpg.Relic.RelicLevels.length == 4) {
                        // I went from 4 to 8 relics and re-ordered them for version 1.1.5, so this
                        // needs to stay in place forever to handle conversion for upgraders (currently 2000 installs)
                        // Old: Dex/Cdr, dmg/str, def/life, luck/spd
                        // New: Life, Dmg, Def, Str, Dex, luck, Cdr, spd, 
                        // Note that Cdr and speed can be purchased separately
                        RogueRpg.Relic.RelicLevels = [
                            Math.max(1, RogueRpg.Relic.RelicLevels[2]),
                            RogueRpg.Relic.RelicLevels[1],
                            RogueRpg.Relic.RelicLevels[2],
                            RogueRpg.Relic.RelicLevels[1],
                            RogueRpg.Relic.RelicLevels[0],
                            RogueRpg.Relic.RelicLevels[3],
                            RogueRpg.Relic.RelicLevels[0],
                            RogueRpg.Relic.RelicLevels[3]
                        ];
                        RogueRpg.Relic.RelicExperience = [0, 0, 0, 0, 0, 0, 0, 0];
                        // Experience has changed pretty significantly so zero it out.
                    }
                    if (RogueRpg.Relic.RelicLevels.length == 8) {
                        RogueRpg.Relic.RelicLevels.push(0, 0, 0); // Defense Gem, Keen Spectacles, Bloody Fang
                    }
                }
                catch (ex) {
                }
            }
            this.prismaState = { prisma: 0, gold: 0, levels: {} };
            var prismaString = localStorage.getItem("prismaState");
            if (prismaString) {
                try {
                    var state = JSON.parse(prismaString);
                    this.prismaState = state;
                }
                catch (ex) {
                    // No need
                }
                // Make sure both are set
                this.prismaState.prisma = this.prismaState.prisma || 0;
                this.prismaState.gold = this.prismaState.gold || 0;
                this.prismaState.levels = this.prismaState.levels || {};
            }
            // TODO: if movingShards ever has items, then show the relic screen
            var msString = localStorage.getItem("movingShards");
            if (msString) {
                try {
                    this.movingShards = JSON.parse(msString);
                }
                catch (ex) {
                    this.movingShards = [];
                }
            }
            var highestClassUnlocked = localStorage.getItem("highestClassUnlocked");
            this.highestClassUnlocked = 0;
            if (highestClassUnlocked == "1") {
                this.highestClassUnlocked = 1;
            }
            else if (highestClassUnlocked == "2") {
                this.highestClassUnlocked = 2;
            }
            else if (highestClassUnlocked == "3") {
                this.highestClassUnlocked = 3;
            }
            else if (highestClassUnlocked == "4") {
                this.highestClassUnlocked = 4;
            }
            else if (highestClassUnlocked == "5") {
                this.highestClassUnlocked = 5;
            }
            var multUnlockedString = localStorage.getItem("maxMultiplierUnlocked");
            if (multUnlockedString) {
                this.maxMultiplierUnlocked = 1;
                try {
                    this.maxMultiplierUnlocked = JSON.parse(multUnlockedString);
                }
                catch (ex) { }
            }
            var ratePromptString = localStorage.getItem("ratePromptHasBeenShown");
            this.ratePromptHasBeenShown = ratePromptString === "true";
            var maxPotions = localStorage.getItem("maxPotions");
            this.maxPotions = maxPotions && maxPotions != "undefined" ? JSON.parse(maxPotions) : 5;
            var inGameStoreItems = localStorage.getItem("inGameStoreItems");
            if (inGameStoreItems) {
                try {
                    this.inGameStoreItems = JSON.parse(inGameStoreItems);
                }
                catch (ex) { }
            }
            if (!this.inGameStoreItems || this.inGameStoreItems.length == 0) {
                this.setupInGameStoreItems();
            }
            var tutorialCompleted = localStorage.getItem("tutorialCompleted");
            this.tutorialCompleted = tutorialCompleted !== null;
            if (!map) {
                // New game!
                this.setupNewGameMap();
                this.showClassSelectionPane(function () {
                    setTimeout(function () {
                        if (!_this.tutorialCompleted) {
                            _this.setupTutorial();
                        }
                        _this.beginLoop();
                    }, 10);
                });
                return;
            }
            // For the latest version update, convert all healing potions into 'healing potion count'
            if (typeof this.hero.state.healingPotionCount != "number") {
                this.hero.state.healingPotionCount = 0;
            }
            for (var i = 0; i < this.hero.state.potions.length; i++) {
                var p = this.hero.state.potions[i];
                if (p.name.substr(0, 7) === "Healing") {
                    this.hero.state.healingPotionCount++;
                    this.hero.state.potions.splice(i, 1);
                    i--;
                }
            }
            if (this.hero.getHeroState().life <= 0) {
                this.beginNewGameAfterDeath();
            }
            else if (!this.hero.state.classChosen && this.hero.level == 1) {
                this.showClassSelectionPane(function () {
                    _this.beginLoop();
                });
            }
        };
        Game.prototype.persistData = function () {
            localStorage.setItem("heroState", JSON.stringify(this.hero.getHeroState()));
            localStorage.setItem("map", JSON.stringify(this.map));
            if (this.previousMapStack) {
                localStorage.setItem("previousMap", JSON.stringify(this.previousMapStack));
            }
            else {
                localStorage.removeItem("previousMap");
            }
            this.persistItemComplex("LevelUpStates", RogueRpg.Level);
            this.persistItemComplex("prismaState", this.prismaState);
            this.persistItemComplex("inGameStoreItems", this.inGameStoreItems);
            this.persistItem("relicLevels", JSON.stringify(RogueRpg.Relic.RelicLevels));
            this.persistItem("relicExp", JSON.stringify(RogueRpg.Relic.RelicExperience));
            this.persistItem("movingShards", JSON.stringify(this.movingShards));
            this.persistItem("gemstones", JSON.stringify(this.gemstones));
            this.persistItem("shards", JSON.stringify(this.shards));
            this.persistItem("highestClassUnlocked", "" + this.highestClassUnlocked);
            this.persistItem("ratePromptHasBeenShown", "" + this.ratePromptHasBeenShown);
            this.persistItem("tutorialCompleted", "" + this.tutorialCompleted);
            this.persistItem("maxMultiplierUnlocked", "" + this.maxMultiplierUnlocked);
            this.persistItem("storedShards", "" + this.storedShards);
            this.persistItem("antialias", "" + this.renderer.antialias);
            this.persistItem("useSimpleGraphics", "" + this.renderer.useSimpleGraphics);
            this.persistItem("startingWeapon", "" + this.startingWeapon);
            this.persistItem("maxPotions", "" + this.maxPotions);
        };
        // Persists a primitive if it has changed
        Game.prototype.persistItem = function (key, value) {
            if (!this.savedValues.hasOwnProperty(key) || this.savedValues[key] !== value) {
                this.savedValues[key] = value;
                localStorage.setItem(key, value);
            }
        };
        // Persists an object if any part has changed, based on JSON.stringify()
        Game.prototype.persistItemComplex = function (key, value) {
            var json = JSON.stringify(value);
            if (!this.savedValues.hasOwnProperty(key) || this.savedValues[key] !== json) {
                this.savedValues[key] = json;
                localStorage.setItem(key, json);
            }
        };
        Game.prototype.beginLoop = function () {
            this.loopPaused = false;
            //by.id("staticCanvas").style.visibility = "visible";
            //if (this._loopId) {
            //    clearInterval(this._loopId);
            //}
            if (!this._loopId) {
                this.startTime = this.time();
                if (window.requestAnimationFrame) {
                    this._loopId = requestAnimationFrame(this.gameLoop.bind(this));
                }
                else {
                    this._loopId = window.setInterval(this.gameLoop.bind(this), RogueRpg.Helpers.LoopTime / 2);
                }
                // this._loopId = setInterval(() => this.gameLoop(), Game.deltaTime);
            }
        };
        Game.prototype.time = function () {
            return (window.performance && window.performance.now && window.performance.now()) || Date.now();
        };
        Game.prototype.onResize = function () {
            // TODO: Remove onResize if not needed, logic moved to wh === window.innerHeight in loop
            //this.resizeExistingCanvas(this.canvas, this.context);
            //this.resetStaticContext();
        };
        Game.prototype.pauseLoop = function () {
            this.loopPaused = true;
            // by.id("staticCanvas").style.visibility = "hidden";
            //if (this._loopId) {
            //    clearInterval(this._loopId);
            //}
        };
        Game.prototype.beginSubLevel = function (tileType) {
            var subMap = new RogueRpg.MapGenerator().createSpecialLevel(tileType, this.floor);
            // Need to cache the old map and the old hero position
            var previous = this.map;
            previous.heroStartX = this.hero.x;
            previous.heroStartY = this.hero.y;
            this.previousMapStack.push(previous);
            this.map = subMap;
            this.hero.x = this.map.heroStartX;
            this.hero.y = this.map.heroStartY;
        };
        Game.prototype.startNewLevel = function () {
            this.setupNewGameMap();
            this.activeAnimations = [];
            this.spriteAnimations = [];
            this.losingLifeChunks = [];
            this.floatingText = [];
            this._mouseDown = false;
            this._tappedThisCycle = false;
        };
        Game.prototype.showClassSelectionPane = function (callback) {
            var _this = this;
            if (this.classSelectionPaneVisible) {
                return;
            }
            this.classSelectionPaneVisible = true;
            this._mouseDown = false;
            this.pauseLoop();
            this.refreshHomeScreenResources();
            by.id("classPane").style.display = "block";
            by.id("classList").innerText = "";
            by.id("homeRelicsButton").onclick = function () {
                _this.showRelicListMenu(true, function () { return _this.refreshHomeScreenResources(); });
            };
            by.id("homeUpgradesButton").onclick = function () {
                _this.showUpgradeListMenu(function () { return _this.refreshHomeScreenResources(); });
            };
            by.id("confirmClassSelection").disabled = true;
            for (var i = 0; i < RogueRpg.HeroClass.Classes.length; i++) {
                var item = RogueRpg.HeroClass.Classes[i];
                var unlocked = i <= this.highestClassUnlocked;
                var div = document.createElement("div");
                if (!unlocked) {
                    div.className = "locked";
                }
                div.innerText = item.name;
                div.onclick = this.showDetailsForClass.bind(this, div, item, unlocked, callback);
                by.id("classList").appendChild(div);
            }
            if (this.maxMultiplierUnlocked > 1) {
                by.id("multiplierSection").style.display = "block";
                by.id("multiplierButton").disabled = false;
                var multiplierButton = by.id("multiplierButton");
                multiplierButton.innerText = this.hero.state.multiplier + "x";
                multiplierButton.onclick = function () {
                    var index = RogueRpg.MultiplierModes.Modes.indexOf(_this.hero.state.multiplier);
                    index++;
                    if (index >= RogueRpg.MultiplierModes.Modes.length || RogueRpg.MultiplierModes.Modes[index] > _this.maxMultiplierUnlocked) {
                        index = 0;
                    }
                    _this.hero.state.multiplier = RogueRpg.MultiplierModes.Modes[index];
                    multiplierButton.innerText = _this.hero.state.multiplier + "x";
                    changeUnlockableCount();
                };
            }
            else {
                // by.id("multiplierSection").style.display = "none";
                by.id("multiplierButton").disabled = true;
            }
            var changeUnlockableCount = function () {
                var stats = _this.getUnlockableCounts(_this.hero.state.multiplier);
                by.id("unlockedCount").innerText = stats.count;
                by.id("unlockedMax").innerText = stats.max;
                if (_this.hero.state.multiplier >= 50) {
                    by.id("prismaSection").style.display = "block";
                }
                else {
                    by.id("prismaSection").style.display = "none";
                }
            };
            changeUnlockableCount();
        };
        Game.prototype.refreshHomeScreenResources = function () {
            by.id("numShards").innerText = RogueRpg.Helpers.formatNumber(this.storedShards);
            by.id("numGemstones").innerText = RogueRpg.Helpers.formatNumber(this.gemstones);
            by.id("numPrisma").innerText = RogueRpg.Helpers.formatNumber(this.prismaState.prisma);
            by.id("numGold").innerText = RogueRpg.Helpers.formatNumber(this.prismaState.gold);
        };
        Game.prototype.showDetailsForClass = function (div, item, unlocked, callback) {
            var _this = this;
            var oldPoint = document.querySelector(".selectedPoint");
            if (oldPoint) {
                oldPoint.classList.remove("selectedPoint");
            }
            div.classList.add("selectedPoint");
            if (unlocked) {
                by.id("confirmClassSelection").innerText = "现在开始";
                by.id("confirmClassSelection").disabled = false;
            }
            else {
                by.id("confirmClassSelection").innerText = "人物未解锁";
                by.id("confirmClassSelection").disabled = true;
            }
            by.id("classDescription").innerText = item.description;
            by.id("confirmClassSelection").onclick = function () {
                _this.hero.state.classIndex = RogueRpg.HeroClass.Classes.indexOf(item);
                _this.hero.state.classChosen = true;
                if (RogueRpg.HeroClass.Classes[_this.hero.state.classIndex].name == "Monk") {
                    _this.hero.equippedWeapon = new RogueRpg.Weapon(0, 1, "Wraps"); // Start off with a Club
                }
                else if (RogueRpg.HeroClass.Classes[_this.hero.state.classIndex].name == "Hunter") {
                    _this.hero.equippedWeapon = new RogueRpg.Weapon(0, 1, "Simple Bow");
                }
                // TODO: Change how you acquire pets
                /*window.setTimeout(() => {
                    var pet = new Pet("Sparks1", Helpers.game.hero.x, Helpers.game.hero.y);
                    pet.drawX = Helpers.hero.drawX;
                    pet.drawY = Helpers.hero.drawY;
                    this.hero.state.pets.push(pet);
                }, 1000);*/
                _this.setupNewGameMap();
                _this.persistData();
                by.id("classPane").style.display = "none";
                _this.classSelectionPaneVisible = false;
                callback();
            };
        };
        Game.prototype.showLevelUpPaneButtonPressed = function () {
            // this.showInventory(); // Side effect of pausing the game.
            this.pauseLoop();
            this.showLevelUpPane(); // Assumes inventory is open
        };
        Game.prototype.showLevelUpPane = function () {
            var _this = this;
            if (this.hero.state.levelUpPoints <= 0) {
                return;
            }
            this._mouseDown = false;
            this.pauseLoop();
            by.id("levelUpPane").style.display = "block";
            this.levelUpPaneVisible = true;
            by.id("levelUpTitlePane").innerText = "技能 " + (this.hero.getHeroState().lastLevelBought + 1) + "!";
            by.id("levelUpPaneCloseButton").onclick = function () { return _this.hideLevelUpPane(); };
            by.id("levelUpDescription").innerText = "";
            by.id("levelUpPurchaseList").innerText = "";
            var pointSelections = this.getCurrentLevelUpPoints();
            if (pointSelections[0].isSeries) {
                by.id("levelUpPurchaseList").innerText = "选择一种技能:";
                by.id("confirmLevelUp").innerText = "选择";
            }
            else {
                by.id("levelUpPurchaseList").innerText = "选择技能:";
                by.id("confirmLevelUp").innerText = "确认";
            }
            if (pointSelections.length == 0) {
                pointSelections = RogueRpg.LevelUpPoint.LargeLevelUpPoints;
            }
            for (var i = 0; i < pointSelections.length; i++) {
                var item = pointSelections[i];
                var div = document.createElement("div");
                div.innerHTML = item.name;
                div.onclick = this.showDetailsForLevelUpPoint.bind(this, div, item);
                by.id("levelUpPurchaseList").appendChild(div);
            }
        };
        // Accepts selectedDiv so it can animate if needed
        Game.prototype.selectAndShowSeries = function (selectedDiv, boundSeries) {
            var allItems = boundSeries.items.slice(0);
            // Remove items that are maxxed out
            for (var i = 0; i < allItems.length; i++) {
                var item = allItems[i];
                if (RogueRpg.Level[item.name] >= item.ranks) {
                    allItems.splice(i, 1);
                    i--;
                }
            }
            if (allItems.length == 0) {
                allItems = RogueRpg.LevelUpPoint.LargeLevelUpPoints;
            }
            var count = 3;
            if (!allItems[0].isSeries) {
                count += (this.prismaState.levels[RogueRpg.PermanentUpgrade.Upgrades.SkillChoice.id] || 0);
            }
            var selectedItems = this.shuffleArray(allItems).splice(0, count);
            allItems = selectedItems; //this.shuffleArray(allItems).splice(0, 3);
            this.hero.state.currentLevelUpPointOptions = allItems;
            this.persistData();
            by.id("levelUpDescription").innerText = "";
            by.id("confirmLevelUp").style.display = "none";
            by.id("confirmLevelUp").innerText = "确认这个技能";
            by.id("levelUpPurchaseList").innerText = "选择一个技能:";
            for (var i = 0; i < allItems.length; i++) {
                var item = allItems[i];
                var div = document.createElement("div");
                div.innerText = item.name;
                div.onclick = this.showDetailsForLevelUpPoint.bind(this, div, item);
                by.id("levelUpPurchaseList").appendChild(div);
            }
        };
        Game.prototype.showDetailsForLevelUpPoint = function (div, point) {
            var _this = this;
            var oldPoint = document.querySelector(".selectedPoint");
            if (oldPoint) {
                oldPoint.classList.remove("selectedPoint");
            }
            div.classList.add("selectedPoint");
            var isSeries = !!point.items;
            by.id("confirmLevelUp").style.display = isSeries ? "Select this path" : "Learn this skill";
            by.id("confirmLevelUp").style.display = "inline-block";
            by.id("levelUpDescription").innerText = point.description;
            if (!point.isSeries) {
                by.id("levelUpDescription").innerText += "\n\n [" + (RogueRpg.Level[point.name] || 0) + "/" + point.ranks + "]";
            }
            by.id("confirmLevelUp").onclick = function () {
                if (isSeries) {
                    _this.selectAndShowSeries(div, point);
                }
                else {
                    if (_this.hero.state.levelUpPoints > 0) {
                        _this.hero.state.levelUpPoints--;
                        _this.hero.clearStatCaches();
                        _this.renderer.renderLevelBar = true;
                        _this.hero.getHeroState().lastLevelBought += 1;
                        RogueRpg.Level[point.name] = RogueRpg.Level[point.name] ? RogueRpg.Level[point.name] + 1 : 1;
                        _this.hero.state.currentLevelUpPointOptions = null;
                        _this.persistData();
                        _this.hideLevelUpPane();
                    }
                }
            };
        };
        Game.prototype.hideLevelUpPane = function () {
            by.id("levelUpPane").style.display = "none";
            by.id("confirmLevelUp").style.display = "none";
            this.levelUpPaneVisible = false;
            this.beginLoop();
        };
        Game.prototype.getCurrentLevelUpPoints = function () {
            if (this.hero.state.currentLevelUpPointOptions && this.hero.state.currentLevelUpPointOptions.length > 0) {
                return this.hero.state.currentLevelUpPointOptions;
            }
            var allItems = this.getAvailablePointsForCurrentLevel().slice(0);
            // Remove items that are maxxed out
            for (var i = 0; i < allItems.length; i++) {
                var item = allItems[i];
                if (!item.isSeries && RogueRpg.Level[item.name] >= item.ranks) {
                    allItems.splice(i, 1);
                    i--;
                }
            }
            var count = 3;
            if (!allItems[0].isSeries) {
                count += (this.prismaState.levels[RogueRpg.PermanentUpgrade.Upgrades.SkillChoice.id] || 0);
            }
            allItems = this.shuffleArray(allItems).splice(0, count);
            this.hero.state.currentLevelUpPointOptions = allItems;
            this.persistData();
            return this.hero.state.currentLevelUpPointOptions;
        };
        /**
         * Randomize array element order in-place.
         * Using Fisher-Yates shuffle algorithm.
         */
        Game.prototype.shuffleArray = function (array) {
            for (var i = array.length - 1; i > 0; i--) {
                var j = Math.floor(Math.random() * (i + 1));
                var temp = array[i];
                array[i] = array[j];
                array[j] = temp;
            }
            return array;
        };
        // TOOD: add support for series selection, don't show series without available points
        // TODO: select up to three random items
        // TODO: Maximize anticipation: Short reveal animation. Notify at 75% to series?
        Game.prototype.getAvailablePointsForCurrentLevel = function () {
            var points = [];
            var level = (this.hero.getHeroState().lastLevelBought + 1);
            if (level == 2 || (level <= 30 && level % 3 == 0) || (level > 30 && level % 2 == 0)) {
                // TODO: Return 3 series and allow a selection
                var allSeries = [
                    { isSeries: true, name: "勇士", items: RogueRpg.LevelUpPoint.Series.Berserker.items, description: RogueRpg.LevelUpPoint.Series.Berserker.description },
                    { isSeries: true, name: "学者", items: RogueRpg.LevelUpPoint.Series.Scholar.items, description: RogueRpg.LevelUpPoint.Series.Scholar.description },
                    { isSeries: true, name: "打击", items: RogueRpg.LevelUpPoint.Series.Crusher.items, description: RogueRpg.LevelUpPoint.Series.Crusher.description },
                    { isSeries: true, name: "飓风", items: RogueRpg.LevelUpPoint.Series.Flurry.items, description: RogueRpg.LevelUpPoint.Series.Flurry.description },
                    { isSeries: true, name: "运气", items: RogueRpg.LevelUpPoint.Series.Luck.items, description: RogueRpg.LevelUpPoint.Series.Luck.description },
                    { isSeries: true, name: "竞技", items: RogueRpg.LevelUpPoint.Series.Athletics.items, description: RogueRpg.LevelUpPoint.Series.Athletics.description }
                ];
                return allSeries;
                // this.shuffleArray(allItems).splice(0, 3)
            }
            else if (level == 5) {
                // Show the level up point for your class
                var className = RogueRpg.HeroClass.Classes[this.hero.state.classIndex].name;
                var series = RogueRpg.LevelUpPoint.Series[className];
                return [
                    { isSeries: true, name: className, description: series.description, items: series.items }
                ];
            }
            else if (level == 20 || level == 26 || level == 31) {
                return RogueRpg.LevelUpPoint.EmpowerSeries;
            }
            else if (level == 8 || level == 11 || level == 17 || level == 23) {
                return RogueRpg.LevelUpPoint.DropChanceModifiers;
            }
            else if (level == 4 || level == 7 || level == 14) {
                return RogueRpg.LevelUpPoint.DualLevelUpPoints;
            }
            else {
                return RogueRpg.LevelUpPoint.MultiplierBonuses;
            }
        };
        Game.prototype.showSettingsPane = function () {
            by.id('settingsPane').style.display = "block";
        };
        Game.prototype.closeSettingsPane = function () {
            by.id('settingsPane').style.display = "none";
        };
        Game.prototype.showInventory = function () {
            var _this = this;
            if (!this.tutorialCompleted && this.tutorialState < 5)
                return;
            this._mouseDown = false;
            this.inventoryVisible = true;
            this.pauseLoop();
            by.id("inventory").style.display = "block";
            var slots = by.id("bagSlots");
            slots.innerText = "";
            for (var i = 0; i < this.hero.items.length; i++) {
                var itemDiv = div();
                var item = this.hero.items[i];
                itemDiv.innerText = item.fullName;
                itemDiv.classList.add("itemRarity" + Math.min(item.rarity, 7));
                itemDiv.classList.add("itemListing");
                itemDiv.addEventListener("click", this.showItemDisplay.bind(this, item, /*isEquipped=*/ false));
                slots.appendChild(itemDiv);
            }
            var equipmentContainer = by.id("equipment");
            equipmentContainer.innerHTML = "";
            var weaponSlot = div("weaponSlot", "equipmentSlot");
            weaponSlot.style.background = this.hero.equippedWeapon ? "url(\"images/" + this.hero.equippedWeapon.image + ".png\")" : "";
            weaponSlot.addEventListener("click", this.showItemDisplay.bind(this, this.hero.equippedWeapon, /*isEquipped=*/ true));
            if (this.hero.equippedWeapon && this.hero.equippedWeapon.rarity == 5) {
                weaponSlot.style.background += ", url('images/InconceivableGlow.png')";
            }
            else if (this.hero.equippedWeapon && this.hero.equippedWeapon.rarity == 6) {
                weaponSlot.style.background += ", url('images/AbsurdGlow.png')";
            }
            else if (this.hero.equippedWeapon && this.hero.equippedWeapon.rarity >= 7) {
                weaponSlot.style.background += ", url('images/ImpossibleGlow.png')";
            }
            equipmentContainer.appendChild(weaponSlot);
            var shieldSlot = div("shieldSlot", "equipmentSlot");
            shieldSlot.style.background = this.hero.equippedShield ? "url(\"images/" + this.hero.equippedShield.image + ".png\")" : "";
            shieldSlot.addEventListener("click", this.showItemDisplay.bind(this, this.hero.equippedShield, /*isEquipped=*/ true));
            if (this.hero.equippedShield && this.hero.equippedShield.rarity == 5) {
                shieldSlot.style.background += ", url('images/InconceivableGlow.png')";
            }
            else if (this.hero.equippedShield && this.hero.equippedShield.rarity == 6) {
                shieldSlot.style.background += ", url('images/AbsurdGlow.png')";
            }
            else if (this.hero.equippedShield && this.hero.equippedShield.rarity >= 7) {
                shieldSlot.style.background += ", url('images/ImpossibleGlow.png')";
            }
            equipmentContainer.appendChild(shieldSlot);
            var armorSlot = div("armorSlot", "equipmentSlot");
            armorSlot.style.background = this.hero.equippedArmor ? "url(\"images/" + this.hero.equippedArmor.image + ".png\")" : "";
            armorSlot.addEventListener("click", this.showItemDisplay.bind(this, this.hero.equippedArmor, /*isEquipped=*/ true));
            if (this.hero.equippedArmor && this.hero.equippedArmor.rarity == 5) {
                armorSlot.style.background += ", url('images/InconceivableGlow.png')";
            }
            else if (this.hero.equippedArmor && this.hero.equippedArmor.rarity == 6) {
                armorSlot.style.background += ", url('images/AbsurdGlow.png')";
            }
            else if (this.hero.equippedArmor && this.hero.equippedArmor.rarity >= 7) {
                armorSlot.style.background += ", url('images/ImpossibleGlow.png')";
            }
            equipmentContainer.appendChild(armorSlot);
            var accessorySlot = div("accessorySlot", "equipmentSlot");
            accessorySlot.style.background = this.hero.equippedAccessory ? "url(\"images/" + this.hero.equippedAccessory.image + ".png\")" : "";
            accessorySlot.addEventListener("click", this.showItemDisplay.bind(this, this.hero.equippedAccessory, /*isEquipped=*/ true));
            if (this.hero.equippedAccessory && this.hero.equippedAccessory.rarity == 5) {
                accessorySlot.style.background += ", url('images/InconceivableGlow.png')";
            }
            else if (this.hero.equippedAccessory && this.hero.equippedAccessory.rarity == 6) {
                accessorySlot.style.background += ", url('images/AbsurdGlow.png')";
            }
            else if (this.hero.equippedAccessory && this.hero.equippedAccessory.rarity >= 7) {
                accessorySlot.style.background += ", url('images/ImpossibleGlow.png')";
            }
            equipmentContainer.appendChild(accessorySlot);
            var stats = by.id("characterStats");
            stats.innerHTML = "";
            var level = div();
            level.innerHTML = "等级: " + this.hero.level;
            level.addEventListener("click", function () { return _this.showExtraDetailsDiv(RogueRpg.Messages.exp(_this.hero)); });
            stats.appendChild(level);
            var exp = div();
            exp.innerHTML = "经验: " + Math.floor(this.hero.currentExperience) + " / " + Math.ceil(this.hero.neededExperience);
            exp.addEventListener("click", function () { return _this.showExtraDetailsDiv(RogueRpg.Messages.exp(_this.hero)); });
            stats.appendChild(exp);
            stats.appendChild(br());
            var life = div();
            life.innerHTML = "生命: " + green(Math.round(this.hero.life)) + " / " + green(this.hero.maxLife);
            life.addEventListener("click", function () { return _this.showExtraDetailsDiv(RogueRpg.Messages.life(_this.hero)); });
            stats.appendChild(life);
            var lifeSteal = div();
            if (RogueRpg.Level[RogueRpg.L.NoLifesteal]) {
                lifeSteal.innerHTML = "生命偷取: +" + green(Math.round(this.hero.maxLife * 0.008));
            }
            else {
                lifeSteal.innerHTML = "生命偷取: " + green(Math.round(this.hero.getLifestealPercent())) + "%";
            }
            stats.appendChild(lifeSteal);
            var damage = div();
            damage.innerHTML = "伤害: " + bold(this.hero.weaponDamage);
            damage.addEventListener("click", function () { return _this.showExtraDetailsDiv(RogueRpg.Messages.damage(_this.hero)); });
            stats.appendChild(damage);
            var delay = div();
            delay.innerHTML = "技能伤害: " + bold(this.hero.weaponDelay);
            delay.addEventListener("click", function () { return _this.showExtraDetailsDiv(RogueRpg.Messages.damage(_this.hero)); });
            stats.appendChild(delay);
            stats.appendChild(br());
            var reduction = div();
            reduction.innerHTML = "技能CD: " + bold(this.hero.reduction);
            reduction.addEventListener("click", function () { return _this.showExtraDetailsDiv(RogueRpg.Messages.defense(_this.hero)); });
            stats.appendChild(reduction);
            var defense = div();
            defense.innerHTML = "防御力: " + bold(this.hero.defense);
            defense.addEventListener("click", function () { return _this.showExtraDetailsDiv(RogueRpg.Messages.defense(_this.hero)); });
            stats.appendChild(defense);
            stats.appendChild(br());
            var strength = div();
            strength.innerHTML = "物理攻击: " + red(this.hero.strength);
            stats.appendChild(strength);
            strength.addEventListener("click", function () { return _this.showExtraDetailsDiv(RogueRpg.Messages.strength(_this.hero)); });
            var dexterity = div();
            dexterity.innerHTML = "命中: " + blue(this.hero.dexterity);
            dexterity.addEventListener("click", function () { return _this.showExtraDetailsDiv(RogueRpg.Messages.dexterity(_this.hero)); });
            stats.appendChild(dexterity);
            var luck = div();
            luck.innerHTML = "幸运值: " + yellow(this.hero.luck);
            stats.appendChild(luck);
            luck.addEventListener("click", function () { return _this.showExtraDetailsDiv(RogueRpg.Messages.luck(_this.hero)); });
            var speed = div();
            speed.innerHTML = "速度: " + this.hero.speed;
            speed.addEventListener("click", function () { return _this.showExtraDetailsDiv(RogueRpg.Messages.speed(_this.hero)); });
            stats.appendChild(speed);
            stats.appendChild(br());
            //var gemstoneAccess = document.createElement("button");
            //gemstoneAccess.innerText = "Gemstones";
            //gemstoneAccess.addEventListener("click",() => this.showGemstoneMenu());
            //stats.appendChild(gemstoneAccess);
            var accessRow = document.createElement("div");
            var levelInfoAccess = document.createElement("div");
            levelInfoAccess.id = "levelInfoAccessButton";
            levelInfoAccess.className = "touchableButton";
            levelInfoAccess.addEventListener("click", function () { return _this.showLevelInfoPane(); });
            accessRow.appendChild(levelInfoAccess);
            var relicAccess = document.createElement("div");
            relicAccess.id = "relicAccessButton";
            relicAccess.className = "touchableButton";
            relicAccess.addEventListener("click", function () { return _this.showRelicMenuFromInventory(); });
            relicAccess.appendChild(RogueRpg.Relic.Relics[this.lastRelicIndex].image);
            accessRow.appendChild(relicAccess);
            stats.appendChild(accessRow);
            var btn = div("inGameStoreInventoryButton", "", "增加");
            btn.onclick = function () {
                _this.showInGameStore();
            };
            stats.appendChild(btn);
            var potionSlots = by.id("potionSlots");
            potionSlots.innerText = "";
            for (var i = 0; i < this.maxPotions; i++) {
                var potionContainer = div(null, "potionContainer");
                var size = potionSlots.getBoundingClientRect().width / this.maxPotions - 5;
                potionContainer.style.width = size + "px";
                potionContainer.style.height = size + "px";
                var potion = this.hero.state.potions[i];
                if (potion) {
                    potionContainer.style.backgroundImage = "url(\"images/" + potion.image + ".png\")";
                    potionContainer.addEventListener("touchend", function (pot) {
                        this.renderPotionDisplay(pot);
                    }.bind(this, potion));
                }
                else {
                    potionContainer.style.backgroundImage = "url(\"images/PotionPlaceholder.png\")";
                }
                potionSlots.appendChild(potionContainer);
            }
            //var levelUpAccess = document.createElement("button");
            //levelUpAccess.innerText = "Level Up (" + this.hero.state.levelUpPoints + ")";
            //levelUpAccess.addEventListener("click",() => this.showLevelUpPane());
            //stats.appendChild(levelUpAccess);
        };
        Game.prototype.hideAlchemyPane = function () {
            this.alchemyPaneVisible = false;
            by.id("alchemyPane").style.display = "none";
            this.beginLoop();
        };
        Game.prototype.showAlchemyPane = function () {
            this.alchemyPaneVisible = true;
            this.pauseLoop();
            by.id("alchemyPane").style.display = "block";
            this.renderAlchemyPane();
        };
        Game.prototype.renderAlchemyPane = function () {
            var potionSlots = by.id("alchemyPotionSlots");
            potionSlots.innerText = "";
            for (var i = 0; i < 5; i++) {
                var potionContainer = div(null, "potionContainer");
                var size = (window.innerWidth - 30 - 60) / 5;
                potionContainer.style.width = size + "px";
                potionContainer.style.height = size + "px";
                var potion = this.hero.state.potions[i];
                if (potion) {
                    potionContainer.style.backgroundImage = "url(\"images/" + potion.image + ".png\")";
                    potionContainer.addEventListener("touchend", function (pot) {
                        this.renderPotionDisplay(pot);
                    }.bind(this, potion));
                }
                else {
                    potionContainer.style.backgroundImage = "url(\"images/PotionPlaceholder.png\")";
                }
                potionSlots.appendChild(potionContainer);
            }
        };
        Game.prototype.hideLevelInfoPane = function () {
            this.levelUpInfoPaneVisible = false;
            by.id("levelInfoPane").style.display = "none";
        };
        Game.prototype.showLevelInfoPane = function () {
            if (!this.tutorialCompleted)
                return;
            this.levelUpInfoPaneVisible = true;
            by.id("levelInfoPane").style.display = "block";
            var cont = by.id("levelInfoContent");
            cont.innerText = "";
            for (var i in RogueRpg.L) {
                if (RogueRpg.Level[RogueRpg.L[i]] == 1) {
                    var ob = document.createElement("div");
                    ob.innerText = RogueRpg.LevelUpPoint.all[RogueRpg.L[i]].description;
                    cont.appendChild(ob);
                }
                else if (RogueRpg.Level[RogueRpg.L[i]] > 1) {
                    var ob = document.createElement("div");
                    ob.innerText = RogueRpg.LevelUpPoint.all[RogueRpg.L[i]].description + " - " + RogueRpg.Level[RogueRpg.L[i]] + " ranks";
                    cont.appendChild(ob);
                }
            }
        };
        // Called whenever an item is purchased in the store, adds item to inventory, potions slot
        // boost or portal used
        Game.prototype.itemBought = function (item) {
            if (item.name === "Heroic Potion") {
                var potion = new RogueRpg.Potion(RogueRpg.PotionEffect.Clear);
                this.grantPotion(potion);
            }
            else if (item.name.substr(0, 7) == "Healing") {
                // Note: This catches the old, deprecated potions, which might be persisted?
                this.hero.state.healingPotionCount++;
            }
            else if (item.name === "Experience Potion +10") {
                var potion = new RogueRpg.Potion(RogueRpg.PotionEffect.Experience, 10);
                this.grantPotion(potion);
            }
            else if (item.name === "Experience Potion +30") {
                var potion = new RogueRpg.Potion(RogueRpg.PotionEffect.Experience, 30);
                this.grantPotion(potion);
            }
            else if (item.name === "Experience Potion +60") {
                var potion = new RogueRpg.Potion(RogueRpg.PotionEffect.Experience, 60);
                this.grantPotion(potion);
            }
            else if (item.name === "Max Life Potion +20") {
                var potion = new RogueRpg.Potion(RogueRpg.PotionEffect.Maxlife, 20);
                this.grantPotion(potion);
            }
            else if (item.name === "Max Life Potion +30") {
                var potion = new RogueRpg.Potion(RogueRpg.PotionEffect.Maxlife, 30);
                this.grantPotion(potion);
            }
            else if (item.name === "Max Life Potion +50") {
                var potion = new RogueRpg.Potion(RogueRpg.PotionEffect.Maxlife, 50);
                this.grantPotion(potion);
            }
            else if (item.name === "Max Life Potion +80") {
                var potion = new RogueRpg.Potion(RogueRpg.PotionEffect.Maxlife, 80);
                this.grantPotion(potion);
            }
            else if (item.name === "Forest Portal") {
                var obj = new RogueRpg.WorldObject("Portal");
                obj.drawX = this.hero.drawX;
                obj.drawY = this.hero.drawY;
                this.map.specialWorldObjects.push(obj);
            }
            else if (item.name === "Item Upgrade Shrine") {
                var obj = new RogueRpg.WorldObject("Item Upgrade Shrine");
                obj.drawX = this.hero.drawX;
                obj.drawY = this.hero.drawY;
                this.map.specialWorldObjects.push(obj);
            }
            // Force the inventory to reload to update display of items
            this.persistData();
            this.showInventory();
        };
        Game.prototype.showInGameStore = function (onCloseCallback) {
            var _this = this;
            if (!this.tutorialCompleted) {
                onCloseCallback();
                return;
            }
            if (!this.inGameStoreItems) {
                this.setupInGameStoreItems();
            }
            this.inGameStoreOnCloseCallback = onCloseCallback;
            this.inGameStoreVisible = true;
            by.id('inGameStore').style.display = "block";
            var shards = by.id('inGameStoreShards');
            shards.innerText = RogueRpg.Helpers.formatNumber(this.shards);
            var gemstones = by.id('inGameStoreGemstones');
            gemstones.innerText = RogueRpg.Helpers.formatNumber(this.gemstones);
            var list = by.id('inGameStoreList');
            list.innerText = "";
            var items = [];
            for (var i = 0; i < this.inGameStoreItems.length; i++) {
                if (this.inGameStoreItems[i].quantity > 0) {
                    var item = new RogueRpg.InGameStoreListItem(this.inGameStoreItems[i], function () {
                        shards.innerText = RogueRpg.Helpers.formatNumber(_this.shards);
                        gemstones.innerText = _this.gemstones + '';
                        for (var k = 0; k < items.length; k++) {
                            items[k].initialize();
                        }
                    }, true);
                    items.push(item);
                    list.appendChild(item.container);
                }
            }
            // TODO: Handle disabling REAL MONEY
            if (REAL_MONEY_ENABLED) {
                by.id('getMoreGemstonesButton').onclick = function () { return _this.showGemstoneMenu(); };
            }
            else {
                by.id('getMoreGemstonesButton').style.display = "none";
            }
        };
        Game.prototype.hideInGameStore = function () {
            this.inGameStoreVisible = false;
            by.id('inGameStore').style.display = "none";
            if (this.inGameStoreOnCloseCallback) {
                this.inGameStoreOnCloseCallback();
            }
        };
        Game.prototype.showUpgradeListMenu = function (upgradeListOnClose) {
            var _this = this;
            this.closeInventory();
            this.pauseLoop();
            this.relicListOnClose = upgradeListOnClose;
            by.id("upgradeListMenu").style.display = "block";
            var goldDiv = by.id("upgradeListMenuGold");
            var prismaDiv = by.id("upgradeListMenuPrisma");
            var list = by.id("upgradeList");
            list.innerText = "";
            var items = [];
            for (var k in RogueRpg.PermanentUpgrade.Upgrades) {
                var upgrade = RogueRpg.PermanentUpgrade.Upgrades[k];
                var item = new RogueRpg.UpgradeListItem(upgrade, function () {
                    goldDiv.innerText = RogueRpg.Helpers.formatNumber(_this.prismaState.gold, true);
                    prismaDiv.innerText = RogueRpg.Helpers.formatNumber(_this.prismaState.prisma, true);
                    refresh();
                });
                items.push(item);
                list.appendChild(item.container);
            }
            var refresh = function () {
                for (var k = 0; k < items.length; k++) {
                    items[k].initialize();
                }
            };
            goldDiv.innerText = RogueRpg.Helpers.formatNumber(this.prismaState.gold, true);
            prismaDiv.innerText = RogueRpg.Helpers.formatNumber(this.prismaState.prisma, true);
        };
        Game.prototype.showRelicListMenu = function (editAllowed, relicListOnClose) {
            var _this = this;
            if (editAllowed === void 0) { editAllowed = false; }
            this.closeInventory();
            this.pauseLoop();
            this.relicListOnClose = relicListOnClose;
            by.id("relicListMenu").style.display = "block";
            var shards = by.id('relicListMenuShards');
            var list = by.id('relicList');
            list.innerText = "";
            var ownedRelicIndices = [];
            var availableRelicIndices = [];
            for (var i = 0; i < RogueRpg.Relic.Relics.length; i++) {
                if (RogueRpg.Relic.RelicLevels[i] > 0) {
                    ownedRelicIndices.push(i);
                }
                else if (RogueRpg.Relic.Relics[i].price) {
                    availableRelicIndices.push(i);
                }
            }
            var items = [];
            for (var i = 0; i < ownedRelicIndices.length; i++) {
                var item = new RogueRpg.RelicListItem(ownedRelicIndices[i], editAllowed, function () {
                    shards.innerText = RogueRpg.Helpers.formatNumber(_this.storedShards, true);
                    refresh();
                });
                items.push(item);
                list.appendChild(item.container);
            }
            for (var i = 0; i < availableRelicIndices.length; i++) {
                var item = new RogueRpg.RelicListItem(availableRelicIndices[i], editAllowed, function () {
                    shards.innerText = RogueRpg.Helpers.formatNumber(_this.storedShards, true);
                    refresh();
                });
                items.push(item);
                list.appendChild(item.container);
            }
            var refresh = function () {
                for (var k = 0; k < items.length; k++) {
                    items[k].initialize();
                }
            };
            var scrollIndex = 0;
            var upButton = by.id("relicUpButton");
            var downButton = by.id("relicDownButton");
            if (ownedRelicIndices.length + availableRelicIndices.length > (window.innerHeight - 200) / 70) {
                upButton.ontouchend = function () {
                    scrollIndex = Math.max(0, scrollIndex - 1);
                    if (items[scrollIndex]) {
                        items[scrollIndex].container.scrollIntoView();
                    }
                };
                downButton.ontouchend = function () {
                    scrollIndex = Math.min(items.length - 1, scrollIndex + 1);
                    if (items[scrollIndex]) {
                        items[scrollIndex].container.scrollIntoView();
                    }
                };
                upButton.style.visibility = "visible";
                downButton.style.visibility = "visible";
            }
            else {
                upButton.style.visibility = "hidden";
                downButton.style.visibility = "hidden";
            }
            shards.innerText = RogueRpg.Helpers.formatNumber(this.storedShards, true);
        };
        // Show relics and allow the player to select the one to wear
        Game.prototype.showRelicMenuFromInventory = function () {
            if (!this.tutorialCompleted)
                return;
            this.showRelicListMenu();
        };
        Game.prototype.hideRelicsOrUpgradesMenu = function () {
            by.id("relicListMenu").style.display = "none";
            by.id("upgradeListMenu").style.display = "none";
            if (this.relicListOnClose) {
                this.relicListOnClose();
            }
            else {
                this.beginLoop();
            }
        };
        // Notes: A long time ago, Gemstones could be used to straight buy upgrades. I don't want the game
        // to literally be: buy gemstones, get marginal power. That isn't fun. Anything you spend gemstones
        // on needs to be either a percentage boost, or a random bonus (and one that persists in a tangible way).
        // So the 'gemstone menu' is now for in app purchases
        Game.prototype.showGemstoneMenu = function () {
            this.gemstoneStoreVisible = true;
            by.id("gemstoneStore").style.display = "block";
            by.id("gemstoneStoreCount").innerText = RogueRpg.Helpers.formatNumber(this.gemstones);
            var listing = by.id("gemstoneStoreListing");
            listing.innerHTML = "";
            for (var i in this.gemstoneProducts) {
                var product = this.gemstoneProducts[i];
                var purchase = RogueRpg.InAppPurchase.Purchases[i];
                var container = document.createElement("div");
                container.className = "inAppPurchase";
                var img = document.createElement("img");
                var text = document.createElement("span");
                container.appendChild(img);
                container.appendChild(text);
                img.src = purchase.image;
                img.className = "gemstoneImage";
                text.className = "purchaseCount";
                text.innerText = "Get " + purchase.baseGems.toLocaleString();
                var button = document.createElement("div");
                button.className = "gemstonePurchaseButton";
                if (purchase.mockPriceUsd) {
                    button.innerText = "$" + purchase.mockPriceUsd;
                }
                else {
                    button.classList.add('watchAVideo');
                    button.innerText = "watch a video";
                }
                button.onclick = function (product, purchase) {
                    var _this = this;
                    inAppPurchase.buy(product.productId)
                        .then(function (data) {
                        _this.gemstones += purchase.baseGems;
                        by.id("gemstoneStoreCount").innerText = _this.gemstones + "";
                        by.id('inGameStoreGemstones').innerText = '' + _this.gemstones;
                        by.id("numGemstones").innerText = '' + _this.gemstones;
                        _this.renderer.resetStaticContext();
                        inAppPurchase.consume(data.productType, data.receipt, data.signature);
                    })
                        .catch(function (err) {
                        //alert("Error making purchase:" + err);    
                    });
                }.bind(this, product, purchase);
                container.appendChild(button);
                listing.appendChild(container);
            }
        };
        Game.prototype.hideGemstoneMenu = function () {
            this.gemstoneStoreVisible = false;
            by.id("gemstoneStore").style.display = "none";
            if (this.gemstoneMenuOnClose) {
                var callback = this.gemstoneMenuOnClose;
                this.gemstoneMenuOnClose = null;
                callback();
            }
        };
        Game.prototype.unequipItem = function (item, event) {
            if (item == this.hero.equippedWeapon) {
                this.hero.items.push(item);
                this.hero.equippedWeapon = null;
            }
            if (item == this.hero.equippedShield) {
                this.hero.items.push(item);
                this.hero.equippedShield = null;
            }
            if (item == this.hero.equippedArmor) {
                this.hero.items.push(item);
                this.hero.equippedArmor = null;
            }
            if (item == this.hero.equippedAccessory) {
                this.hero.items.push(item);
                this.hero.equippedAccessory = null;
            }
            if (event) {
                this.closeInventory();
                this.showInventory();
                event.stopImmediatePropagation();
                event.cancelBubble = true;
            }
            this.hero.clearStatCaches();
        };
        // Only works to equip an item currently in the inventory
        Game.prototype.equipItem = function (item, event) {
            this.renderer.renderItemCount = true;
            var oldMaxLife = this.hero.maxLife;
            var itemIndex = this.hero.items.indexOf(item);
            if (item instanceof RogueRpg.Weapon) {
                if (this.hero.equippedWeapon) {
                    this.hero.items[itemIndex] = this.hero.equippedWeapon;
                }
                else {
                    this.hero.items.splice(itemIndex, 1);
                }
                this.hero.equippedWeapon = item;
            }
            else if (item instanceof RogueRpg.Shield) {
                if (this.hero.equippedShield) {
                    this.hero.items[itemIndex] = this.hero.equippedShield;
                }
                else {
                    this.hero.items.splice(itemIndex, 1);
                }
                this.hero.equippedShield = item;
            }
            else if (item instanceof RogueRpg.Armor) {
                if (this.hero.equippedArmor) {
                    this.hero.items[itemIndex] = this.hero.equippedArmor;
                }
                else {
                    this.hero.items.splice(itemIndex, 1);
                }
                this.hero.equippedArmor = item;
            }
            else if (item instanceof RogueRpg.Accessory) {
                if (this.hero.equippedAccessory) {
                    this.hero.items[itemIndex] = this.hero.equippedAccessory;
                }
                else {
                    this.hero.items.splice(itemIndex, 1);
                }
                this.hero.equippedAccessory = item;
            }
            this.hero.clearStatCaches();
            var newMaxLife = this.hero.maxLife;
            var lifeChangeRatio = newMaxLife / oldMaxLife;
            this.hero.life *= lifeChangeRatio;
            // Used to stop clicking an an item from also showing it.
            // event is not passed if the item is being autoequipped, to avoid opening inventory
            if (event) {
                // Refresh the inventory, which includes hero stats
                this.closeInventory();
                this.showInventory();
                event.stopImmediatePropagation();
                event.cancelBubble = true;
            }
            this.hero.clearStatCaches();
        };
        Game.prototype.showItemDisplay = function (item, isEquipped) {
            if ((this.tutorialState == 9 && item.fullName != "Sword") || (this.tutorialState == 6 && item.fullName != "Longsword")) {
                return;
            }
            this._itemDisplayVisible = true;
            var equippedItem;
            var inventoryItem;
            if (isEquipped) {
                equippedItem = item;
            }
            else {
                if (item instanceof RogueRpg.Weapon && this.hero.equippedWeapon) {
                    equippedItem = this.hero.equippedWeapon;
                }
                else if (item instanceof RogueRpg.Shield && this.hero.equippedShield) {
                    equippedItem = this.hero.equippedShield;
                }
                else if (item instanceof RogueRpg.Armor && this.hero.equippedArmor) {
                    equippedItem = this.hero.equippedArmor;
                }
                else if (item instanceof RogueRpg.Accessory && this.hero.equippedAccessory) {
                    equippedItem = this.hero.equippedAccessory;
                }
                inventoryItem = item;
            }
            if (equippedItem) {
                this.renderItemDisplay(equippedItem, "itemDisplayEquipped");
            }
            if (inventoryItem) {
                this.renderItemDisplay(inventoryItem, "itemDisplayInventory");
            }
        };
        Game.prototype.drinkPotion = function (potion) {
            switch (potion.effects[0]) {
                case RogueRpg.PotionEffect.Experience:
                    this.grantExperience(potion.values[0], true);
                    break;
                case RogueRpg.PotionEffect.Maxlife:
                    this.hero.baseMaxLife += potion.values[0];
                    this.hero.clearStatCaches();
                    this.healHeroFor(potion.values[0]);
                    break;
                case RogueRpg.PotionEffect.Dex:
                    this.hero.baseDexterity += potion.values[0];
                    break;
                case RogueRpg.PotionEffect.Healing:
                    this.healHeroFor(this.hero.maxLife * (potion.values[0] / 100));
                    break;
                case RogueRpg.PotionEffect.Luck:
                    this.hero.state.luck += potion.values[0];
                    break;
                case RogueRpg.PotionEffect.SkillPoint:
                    this.hero.state.levelUpPoints += potion.values[0];
                    this.renderer.renderLevelBar = true;
                    break;
                case RogueRpg.PotionEffect.Speed:
                    this.hero.baseSpeed += potion.values[0];
                    break;
                case RogueRpg.PotionEffect.Strength:
                    this.hero.baseStrength += potion.values[0];
                    break;
                case RogueRpg.PotionEffect.Clear:
                    this.closeInventory();
                    this.clearCurrentLevel();
                    break;
                case RogueRpg.PotionEffect.Damage:
                    this.hero.state.baseDamage += potion.values[0];
                    break;
                case RogueRpg.PotionEffect.Reduction:
                    this.hero.state.baseReduction += potion.values[0];
                    break;
                case RogueRpg.PotionEffect.Defense:
                    this.hero.state.baseDefense += potion.values[0];
                    break;
                default:
                    return;
            }
            this.hero.clearStatCaches();
        };
        Game.prototype.renderPotionDisplay = function (item) {
            var _this = this;
            var container = by.id("itemDisplayPotion");
            container.style.display = "block";
            var iconDiv = container.querySelector(".itemDisplayIcon");
            iconDiv.style.backgroundImage = "url(\"images/" + item.image + ".png\")";
            iconDiv.innerHTML = "";
            var clickTarget = div();
            clickTarget.style.width = "70px";
            clickTarget.style.height = "70px";
            clickTarget.style.margin = "0";
            var drinkFuction = function (e) {
                var oldIndex = _this.hero.state.potions.indexOf(item);
                _this.hero.state.potions.splice(oldIndex, 1);
                _this.drinkPotion(item);
                _this.hideItemDisplay(e);
                _this.persistData();
                _this.renderer.drawHeroLifeBar(_this.hero);
                if (item.effects[0] !== RogueRpg.PotionEffect.Clear) {
                    _this.showInventory();
                }
            };
            by.id("potionDisplayDrink").onclick = drinkFuction;
            iconDiv.appendChild(clickTarget); // At the moment, clicking on the potion image wont drink it
            var textDiv = container.querySelector(".itemDisplayName");
            textDiv.innerHTML = "";
            var span = document.createElement("span");
            span.className = "itemRarity3";
            span.innerText = item.name;
            textDiv.appendChild(span);
            var description = container.querySelector(".itemDisplayFlavorText");
            description.innerText = item.description;
            var stats = container.querySelector(".itemDisplayMainStatText");
            stats.innerHTML = RogueRpg.Potion.getStatsString(item);
            document.removeEventListener("click", this._hideItemDisplayHandler, false);
            setTimeout(function () {
                document.addEventListener("click", _this._hideItemDisplayHandler, false);
            }, 100);
        };
        Game.prototype.renderItemDisplay = function (item, divName) {
            var _this = this;
            this.displayedItem = item;
            var container = by.id(divName);
            container.style.display = "block";
            var iconDiv = container.querySelector(".itemDisplayIcon");
            iconDiv.style.backgroundImage = "url(\"images/" + item.image + ".png\")";
            iconDiv.innerHTML = "";
            var clickTarget = div();
            clickTarget.style.width = "70px";
            clickTarget.style.height = "70px";
            clickTarget.style.margin = "0";
            if (divName == "itemDisplayInventory") {
                var equipFunction = function (e) {
                    var oldIndex = _this.hero.items.indexOf(item);
                    _this.equipItem(item, e);
                    _this.hideItemDisplay(e);
                    if (_this.hero.items[oldIndex]) {
                        _this.showItemDisplay(_this.hero.items[oldIndex], false);
                    }
                    _this.persistData();
                };
                by.id("itemDisplayEquip").onclick = equipFunction;
                clickTarget.addEventListener("click", equipFunction, /*useCapture=*/ true);
            }
            else {
                /*clickTarget.addEventListener("click",(e: MouseEvent) => {
                    this.unequipItem(item, e);
                    this.hideItemDisplay(e);
                    this.persistData();
                }, true);*/
            }
            iconDiv.appendChild(clickTarget);
            var textDiv = container.querySelector(".itemDisplayName");
            textDiv.innerHTML = "";
            var span = document.createElement("span");
            span.className = "itemRarity" + Math.min(item.rarity, 7);
            span.innerText = item.fullName + " (" + item.level + ")";
            textDiv.appendChild(span);
            var textDiv = container.querySelector(".itemDisplayMainStatText");
            textDiv.innerHTML = item.getMainStatText();
            var list = container.querySelector(".itemDisplayProperties");
            list.innerHTML = "";
            for (var i = 0; i < item.magicProperties.length; i++) {
                var prop = div();
                prop.innerHTML = RogueRpg.MagicProperty.getDisplayString(item.magicProperties[i]);
                list.appendChild(prop);
            }
            var flavor = container.querySelector(".itemDisplayFlavorText");
            flavor.innerText = item.getFlavorText();
            if (divName == "itemDisplayInventory") {
                var effect = this.calculateItemEffect(item);
                var itemRatios = by.id("itemDisplayApproxRatios");
                itemRatios.innerHTML = RogueRpg.Messages.approxRatios(effect);
                var trash = by.id("itemDisplayTrash");
                var sellText = by.id("itemDisplaySellText");
                sellText.innerText = RogueRpg.Helpers.formatNumber(this.getItemSellValue(item));
                trash.onclick = function (event) {
                    var index = _this.hero.items.indexOf(item);
                    _this.renderer.renderItemCount = true;
                    _this.lastDestroyedItem = item;
                    // TODO: Revert gold grant on UNDO
                    _this.grantGold(_this.getItemSellValue(item));
                    _this.addUndoDiv(item);
                    // TODO: Animate this
                    _this.hero.items.splice(index, 1);
                    _this.hideItemDisplay(null);
                    // Remove the inventory item
                    var items = document.querySelectorAll('#bagSlots > .itemListing');
                    var listing = items.item(index);
                    listing.parentElement.removeChild(listing);
                    // Prevent the click from also opening the item behind it
                    event.stopImmediatePropagation();
                    event.cancelBubble = true;
                    if (_this.hero.items[index]) {
                        _this.showItemDisplay(_this.hero.items[index], false);
                    }
                    else if (index > 0 && _this.hero.items[index - 1]) {
                        _this.showItemDisplay(_this.hero.items[index - 1], false);
                    }
                    _this.persistData();
                };
            }
            // Might accidentally have one still added
            document.removeEventListener("click", this._hideItemDisplayHandler, false);
            setTimeout(function () {
                document.addEventListener("click", _this._hideItemDisplayHandler, false);
            }, 100);
        };
        Game.prototype.closeUndoDiv = function () {
            var oldDiv = document.querySelector(".undoDiv");
            if (oldDiv) {
                oldDiv.parentElement.removeChild(oldDiv);
            }
        };
        Game.prototype.addUndoDiv = function (item) {
            var _this = this;
            if (!this.tutorialCompleted && this.tutorialState < 14)
                return;
            this.closeUndoDiv();
            var span = document.createElement("span");
            span.innerText = item.fullName;
            span.className = "itemRarity" + Math.min(item.rarity, 7);
            var div = document.createElement("div");
            div.className = "undoDiv";
            div.innerText = "- ";
            div.appendChild(span);
            var button = document.createElement("button");
            button.innerText = "Undo";
            button.onclick = function () {
                _this.hero.items.push(item);
                _this.grantGold(-1 * _this.getItemSellValue(item));
                _this.closeInventory();
                _this.showInventory();
                _this.showItemDisplay(item, false);
            };
            div.appendChild(button);
            document.body.appendChild(div);
        };
        Game.prototype.hideItemDisplay = function (event) {
            if (this.tutorialState == 7 || this.tutorialState == 10) {
                return;
            }
            this.displayedItem = null;
            this._itemDisplayVisible = false;
            document.removeEventListener("click", this._hideItemDisplayHandler, false);
            by.id("itemDisplayInventory").style.display = "none";
            by.id("itemDisplayEquipped").style.display = "none";
            by.id("itemDisplayPotion").style.display = "none";
            if (event) {
                event.stopImmediatePropagation();
                event.preventDefault();
            }
        };
        Game.prototype.showDialogPrompt = function (textHtml, confirmHtml, callbackOnOk, cancelHtml) {
            var _this = this;
            by.id("extraDetailsConfirmButton").style.display = "block";
            by.id("extraDetailsConfirmButton").disabled = false;
            by.id("extraDetailsConfirmButton").innerHTML = confirmHtml;
            by.id("extraDetailsCloseButton").style.display = "block";
            by.id("extraDetailsCloseButton").disabled = false;
            by.id("extraDetailsCloseButton").innerHTML = cancelHtml;
            this.promptCallbackOnCancel = null;
            by.id("extraDetailsConfirmButton").onclick = function (e) {
                callbackOnOk();
                _this.hideExtraDetails(e);
            };
            this._mouseDown = false;
            this.pauseLoop();
            this.extraDetailsDivVisible = true;
            by.id("extraDetailsDiv").style.display = "block";
            by.id("extraDetailsText").innerHTML = textHtml;
        };
        // if buttonClose is true, this prevent all interaction with the game until the box is closed;
        Game.prototype.showExtraDetailsDiv = function (textHtml, callbackOnClose) {
            if (!this.tutorialCompleted && this.tutorialState < 12)
                return;
            this._mouseDown = false;
            this.pauseLoop();
            this.promptCallbackOnCancel = callbackOnClose;
            by.id("extraDetailsCloseButton").innerText = "好";
            by.id("extraDetailsConfirmButton").style.display = "none";
            this.extraDetailsDivVisible = true;
            by.id("extraDetailsDiv").style.display = "block";
            by.id("extraDetailsText").innerHTML = textHtml;
        };
        Game.prototype.blast = function () {
            this.spriteAnimations.push(RogueRpg.Images.Sprites.Explosion.getInstance(new RogueRpg.Location(this.hero.drawX + this.hero.drawWidth / 1.5, this.hero.drawY - 10)));
            this.damageArea(this.hero.x, this.hero.y, 5 * this.hero.weaponDamage * (1000 / this.hero.weaponDelay), 200, false);
            for (var i = 0; i < this.map.specialWorldObjects.length; i++) {
                var obj = this.map.specialWorldObjects[i];
                if (obj.type == "Frozen Monk") {
                    if (RogueRpg.Helpers.distance(obj.drawX + obj.drawWidth / 2, this.hero.x, obj.drawY + obj.drawHeight / 2, this.hero.y) < 400) {
                        this.map.specialWorldObjects[i].type = "Unfrozen Monk";
                    }
                }
            }
        };
        Game.prototype.resurrect = function () {
            //ga_storage._trackEvent('Resurrect', "Floor " + this.hero.floor);
            var _this = this;
            this.hero.life = this.hero.maxLife;
            this.beginLoop();
            this.blast();
            setTimeout(function () { return _this.blast(); }, 1000);
            setTimeout(function () { return _this.blast(); }, 2000);
            setTimeout(function () { return _this.blast(); }, 3000);
        };
        Game.prototype.showPleaseRatePrompt = function () {
            var _this = this;
            if (this.ratePromptHasBeenShown) {
                return;
            }
            this.ratePromptHasBeenShown = true;
            this.closeInventory();
            this.extraDetailsDivVisible = true;
            this._mouseDown = false;
            this.pauseLoop();
            by.id("extraDetailsDiv").style.display = "block";
            by.id("extraDetailsText").innerHTML = "<b>请评价!</b><br /><br />你喜欢恶棍角色扮演吗？对应用程序进行评级也将帮助其他人找到并享受游戏!<br /><br /><i>不管怎样，谢谢你玩！这是50英镑<img src='images/gemstones.png' class='tinyGemstones' /> 用你喜欢的方式.</i>";
            by.id("extraDetailsCloseButton").onclick = function () { return _this._hideExtraDetailsButtonHandler(); };
            this.setGemstones(this.gemstones + 50);
            this.persistData();
            by.id("extraDetailsCloseButton").innerText = "不";
            by.id("extraDetailsConfirmButton").style.display = "block";
            by.id("extraDetailsConfirmButton").innerHTML = "现在";
            by.id("extraDetailsConfirmButton").disabled = false;
            by.id("extraDetailsConfirmButton").onclick = function (e) {
                _this.hideExtraDetails(e);
                _this.showInventory();
                setTimeout(function () {
                    cordova.plugins.market.open('io.cordova.RogueRpg');
                }, RogueRpg.Helpers.LoopTime * 3);
            };
        };
        Game.prototype.showDeathPrompt = function (killer) {
            var _this = this;
            this.closeInventory();
            this.extraDetailsDivVisible = true;
            this._mouseDown = false;
            this.pauseLoop();
            // TODO: Experiment with removing the death screen for now 
            //this.promptCallbackOnCancel = () => this.handleRelicExpAfterDeath();
            this.persistData();
            var gemstonesToContinue = 25 + this.floor * 5;
            var message = RogueRpg.Messages.death(this.hero, killer);
            by.id("extraDetailsCloseButton").onclick = function () {
                _this.confirmDefeat(function () {
                    by.id("extraDetailsCloseButton").onclick = function () { return _this._hideExtraDetailsButtonHandler(); };
                    by.id("extraDetailsDiv").style.display = "none";
                    by.id("staticCanvas").style.visibility = "visible";
                    _this.handleRelicExpAfterDeath();
                });
            };
            if (gemstonesToContinue <= this.gemstones) {
                message += "<div>支付 " + gemstonesToContinue + " 宝石继续吗? (你有 " + this.gemstones + ")</div>";
                by.id("extraDetailsCloseButton").innerText = "不";
                by.id("extraDetailsConfirmButton").style.display = "block";
                by.id("extraDetailsConfirmButton").innerText = "复活 (" + gemstonesToContinue + ")";
                by.id("extraDetailsConfirmButton").disabled = false;
                by.id("extraDetailsConfirmButton").onclick = function () {
                    _this.confirmResurrection(function () {
                        by.id("extraDetailsDiv").style.display = "none";
                        _this.extraDetailsDivVisible = false;
                        _this.setGemstones(_this.gemstones - gemstonesToContinue);
                        _this.resurrect();
                        by.id("extraDetailsCloseButton").onclick = function () { return _this._hideExtraDetailsButtonHandler(); };
                    });
                };
            }
            else {
                message += "<div>你需要 " + gemstonesToContinue + " 宝石要继续，但你只有。 " + this.gemstones + "</div>";
                by.id("extraDetailsCloseButton").innerText = "失败";
                by.id("extraDetailsConfirmButton").style.display = "block";
                if (REAL_MONEY_ENABLED) {
                    by.id("extraDetailsConfirmButton").innerText = "购买宝石。(需要 " + gemstonesToContinue + ")";
                    by.id("extraDetailsConfirmButton").disabled = false;
                    by.id("extraDetailsConfirmButton").onclick = function () {
                        _this.showGemstoneMenu();
                        _this.gemstoneMenuOnClose = function () {
                            _this.showDeathPrompt(killer);
                        };
                    };
                }
                else {
                    by.id("extraDetailsConfirmButton").innerText = "不能复活。 (need " + gemstonesToContinue + ")";
                    by.id("extraDetailsConfirmButton").disabled = true;
                }
            }
            by.id("extraDetailsDiv").style.display = "block";
            by.id("extraDetailsText").innerHTML = message;
        };
        Game.prototype.confirmDefeat = function (callback) {
            var message = "你确定要接受失败吗?<br />" +
                "<br />你可以用你的碎片在死后升级文物。<br /><br /><br />";
            RogueRpg.Panels.showSecondaryConfirmPrompt(message, "失败", callback);
        };
        Game.prototype.confirmResurrection = function (callback) {
            var message = "你确定要用宝石来复活吗?<br />" +
                "<br />每一层都变得更加困难，所以接受死亡和保存你的宝石通常是有意义的。<br />";
            RogueRpg.Panels.showSecondaryConfirmPrompt(message, "确定", callback);
        };
        Game.prototype.unlockMonkPrompt = function () {
            var _this = this;
            this.closeInventory();
            this.extraDetailsDivVisible = true;
            this._mouseDown = false;
            this.pauseLoop();
            var gemstonesToUnlock = 50;
            var message = "这个和尚被冻在一块冰里。!<br /><br />也许有某种方法可以融化冰，用宝石的力量复活和尚。";
            by.id("extraDetailsCloseButton").onclick = function () { return _this._hideExtraDetailsButtonHandler(); };
            if (gemstonesToUnlock <= this.gemstones) {
                message += "<div>Pay " + gemstonesToUnlock + " 宝石解锁和尚吗? (你有 " + this.gemstones + ")</div>";
                by.id("extraDetailsCloseButton").innerText = "不";
                by.id("extraDetailsConfirmButton").style.display = "block";
                by.id("extraDetailsConfirmButton").innerText = "营救 (" + gemstonesToUnlock + ")";
                by.id("extraDetailsConfirmButton").disabled = false;
                by.id("extraDetailsConfirmButton").onclick = function () {
                    _this.setGemstones(_this.gemstones - gemstonesToUnlock);
                    by.id("extraDetailsDiv").style.display = "none";
                    _this.extraDetailsDivVisible = false;
                    _this.resurrect();
                };
            }
            else {
                message += "<div>You need " + gemstonesToUnlock + " 宝石去救和尚，但你只有。" + this.gemstones + "</div>";
                by.id("extraDetailsCloseButton").innerText = "不";
                by.id("extraDetailsConfirmButton").style.display = "block";
                by.id("extraDetailsConfirmButton").disabled = true;
                by.id("extraDetailsConfirmButton").innerText = "不拯救 (" + gemstonesToUnlock + ")";
                by.id("extraDetailsConfirmButton").onclick = null;
            }
            by.id("extraDetailsDiv").style.display = "block";
            by.id("extraDetailsText").innerHTML = message;
        };
        Game.prototype.hideExtraDetails = function (event) {
            this.extraDetailsDivVisible = false;
            if (!this.inventoryVisible && !this.classSelectionPaneVisible) {
                this.beginLoop();
            }
            by.id("extraDetailsDiv").style.display = "none";
            this._mouseX = null;
            this._mouseY = null;
            this._mouseDown = false;
            if (event) {
                event.stopImmediatePropagation();
                event.preventDefault();
                event.cancelBubble = true;
            }
            if (this.promptCallbackOnCancel) {
                this.promptCallbackOnCancel();
                this.promptCallbackOnCancel = null;
            }
        };
        Game.prototype.closeInventory = function () {
            if (this.tutorialState > 0 && this.tutorialState < 11) {
                return;
            }
            this.closeUndoDiv();
            by.id("extraDetailsDiv").style.display = "none";
            this.inventoryVisible = false;
            by.id("inventory").style.display = "none";
            this.beginLoop();
        };
        Game.prototype.gameLoop = function () {
            Game.deltaTime = this.time() - this.startTime;
            if (Game.deltaTime > 100) {
                Game.deltaTime = 100;
            }
            if (this.recordTiming) {
                this.deltaTimes.push(Game.deltaTime);
                if (this.deltaTimes.length > 100) {
                    this.deltaTimes.splice(0, 1);
                }
                this.average = this.deltaTimes.reduce(function (p, c) { return p + c; }, 0) / this.deltaTimes.length;
            }
            this.startTime = this.time();
            try {
                this.hero.cachedHeroState = false; // Clear at the beginning of the loop all the cached character data we might need to calculate
                if (this.advancingLevel) {
                    this.advanceLevelLoop();
                }
                else if (this.shardGainShowing) {
                    this.shardGainLoop();
                }
                else if (!this.loopPaused) {
                    var timings = [];
                    //timings.push(performance.now()); // A
                    if (this.hero.attackingTime > 0) {
                        this.hero.attackingTime -= Game.deltaTime;
                    }
                    if (!this.clearingCurrentLevel) {
                        this.handleClickOrTouch();
                    }
                    else {
                        this.updateCycleClearCurrentLevel();
                    }
                    //timings.push(performance.now()); // B
                    this.doAggregations();
                    this.updateProjectiles();
                    this.checkForNearbyHealing();
                    this.checkForNearbyGemstonesAndShards();
                    this.updateActiveEffects();
                    this.handleCallbacks();
                    //this._updatePets();
                    this._updateMonsters();
                    this.updateOtherUnits();
                    this.renderer.render(this);
                    //timings.push(performance.now());
                    this.hero.specialDelay -= Game.deltaTime; //Helpers.LoopTime;
                    this.hero.remainingDelay -= Game.deltaTime; // Helpers.LoopTime;
                    this._tappedThisCycle = false;
                    this.checkForTimeBasedState();
                    //timings.push(performance.now());
                    var tim = timings.join(", ");
                    //console.log(tim);
                }
                if (!this.tutorialCompleted && this.tutorialState > 0) {
                    this.drawTutorial();
                }
                if (this.needsToPersist) {
                    this.needsToPersist = false;
                    this.persistData();
                }
            }
            catch (ex) {
                if (!this.exceptionLogged) {
                    this.exceptionLogged = true;
                    //ga_storage._trackEvent('Exception', ex.message);
                    alert("Error during game. Please report this to the developer:\n\n" + ex.message + "\n -- \n" + ex.stack);
                }
                else {
                }
            }
            if (window.requestAnimationFrame) {
                this._loopId = requestAnimationFrame(this.gameLoop.bind(this));
            }
            var loopTime = this.time() - this.startTime;
            if (this.recordTiming) {
                this.loopTimes.push(loopTime);
                if (this.loopTimes.length > 100) {
                    this.loopTimes.splice(0, 1);
                }
                this.loopAverage = this.loopTimes.reduce(function (p, c) { return p + c; }, 0) / this.loopTimes.length;
            }
        };
        Game.prototype.doAggregations = function () {
            if (this.goldAccumulation > 0) {
                this.floatingText.push(new RogueRpg.FloatingText("+" + RogueRpg.Helpers.formatNumber(this.goldAccumulation) + " gold", new RogueRpg.Rgba(230, 210, 70), this.hero.drawX - 5, this.hero.drawY - 20, false, RogueRpg.FloatingText.DEFAULT_DURATION * 2));
                this.goldAccumulation = 0;
            }
            if (this.prismaState.levels[RogueRpg.PermanentUpgrade.Upgrades.AutoOpenChests.id]) {
                for (var i = 0; i < this.map.specialWorldObjects.length; i++) {
                    var obj = this.map.specialWorldObjects[i];
                    if (obj.type == "Treasure Chest" || obj.type == "Dark Treasure Chest") {
                        if (RogueRpg.Helpers.dist(obj, this.hero) < 90) {
                            this.openChest(obj);
                        }
                    }
                }
            }
        };
        Game.prototype.updateProjectiles = function () {
            for (var i = 0; i < this.projectiles.length; i++) {
                var p = this.projectiles[i];
                if (p.dead) {
                    this.projectiles.splice(i, 1);
                    i--;
                    continue;
                }
                p.x += p.xVelocity;
                p.y += p.yVelocity;
                if (p.source == this.hero) {
                    // Check for monster intersection
                    for (var j = 0; j < this.map.monsters.length; j++) {
                        var monster = this.map.monsters[j];
                        if (monster && RogueRpg.Dude.containsX(p.x, monster, 8) && RogueRpg.Dude.containsY(p.y, monster, 8)) {
                            this.damageMonster(monster, p.damage, true);
                            p.dead = true;
                            break;
                        }
                    }
                }
                else {
                    // Check for hero intersection
                }
                if (!RogueRpg.Helpers.IsWalkable(new RogueRpg.Location(p.x, p.y), this.map)) {
                    // TODO: Handle projectile death - explosion, or stick-to-wall, for example
                    p.dead = true;
                }
            }
        };
        Game.prototype.handleCallbacks = function () {
            for (var i = 0; i < this.callbacks.length; i++) {
                var c = this.callbacks[i];
                c.timeLeft -= Game.deltaTime;
                if (c.timeLeft <= 0) {
                    this.callbacks.splice(i, 1);
                    i--;
                    c.callback();
                }
            }
        };
        // Trigger any active effects and remove old ones if needed
        Game.prototype.updateActiveEffects = function () {
            var effects = this.hero.state.activeEffects;
            if (!effects)
                return;
            for (var i = 0; i < effects.length; i++) {
                var e = effects[i];
                if (e.name === "Stab") {
                    while (e.state.delayRemaining <= 0) {
                        e.state.delayRemaining += this.hero.weaponDelay / 2;
                        this.doStab();
                    }
                    e.state.delayRemaining -= Game.deltaTime;
                }
                e.timeLeft -= Game.deltaTime;
                if (e.timeLeft <= 0) {
                    effects.splice(i, 1);
                    i--;
                }
            }
        };
        Game.prototype.doStab = function () {
            var _this = this;
            var numTargets = 2; // TODO: Add customization to special abilities
            if (RogueRpg.Level[RogueRpg.L.StabTargets]) {
                numTargets++;
            }
            var range = 95;
            // First, find up to 2 enemies in range
            // For each, deal 2x damage and trigger a stab animation
            var monstersInRange = this.map.monsters.filter(function (monster) { return RogueRpg.Helpers.dist(monster, _this.hero) <= range; });
            this.shuffleArray(monstersInRange);
            while (numTargets > 0 && monstersInRange.length > 0) {
                var target = monstersInRange.pop();
                numTargets--;
                var damage = this.applyRange(this.hero.weaponDamage);
                this.damageMonster(target, damage, true);
                this.animateAttack(target, "Stab"); // TODO: Rotation
            }
        };
        Game.prototype.setupTutorial = function () {
            this.tutorialState = 1;
            by.id("inventoryAccess").style.visibility = "hidden";
            by.id("itemDisplayTrash").style.visibility = "hidden";
            // Remove any monsters from before the tutorial to avoid conflicts
            for (var i = this.map.monsters.length - 1; i >= 0; i--) {
                var m = this.map.monsters[i];
                if (m.drawX <= 1200) {
                    this.map.monsters.splice(i, 1);
                }
            }
        };
        Game.prototype.tut = function (t, isLower) {
            if (!this.renderer.tutorialContext) {
                return;
            }
            if (isLower) {
                this.renderer.tutorialContext.strokeText(t, 20, 240);
                this.renderer.tutorialContext.fillText(t, 20, 240);
            }
            else {
                this.renderer.tutorialContext.strokeText(t, 20, 100);
                this.renderer.tutorialContext.fillText(t, 20, 100);
            }
        };
        Game.prototype.drawTutorial = function () {
            var _this = this;
            if (this.renderer.tutorialContext) {
                this.renderer.tutorialContext.strokeStyle = "rgb(0, 0, 0)";
                this.renderer.tutorialContext.fillStyle = "rgb(255, 255, 255)";
                this.renderer.tutorialContext.lineWidth = 2;
                this.renderer.tutorialContext.font = "bold 19px 'Lucida Sans'";
                this.renderer.tutorialContext.clearRect(0, 0, this.renderer.ww, this.renderer.wh);
            }
            if (this.tutorialState == 1) {
                this.tut("触碰任何地方移动");
                this.hero.state.healingPotionCount = 2;
                this.pointerLocation = new RogueRpg.Location(this.renderer.ww - 65, this.renderer.wh / 2);
                if (this.hero.x >= 700) {
                    var goblin = this.spawnSpecificMonster(this.map, new RogueRpg.Location(this.hero.x - 50, this.hero.y), "Goblin", true);
                    goblin.maxLife = 7;
                    goblin.life = 7;
                    goblin.skipLoot = true;
                    this.tutorialGoblins.push(goblin);
                    this.tutorialState++;
                    this.pointerLocation = new RogueRpg.Location(goblin.drawX - RogueRpg.Helpers.OriginX + 40, goblin.drawY - RogueRpg.Helpers.OriginY + 40);
                }
            }
            else if (this.tutorialState == 2) {
                this.tut("接触敌人进攻。");
                this.lockHeroMovement = true;
                if (this.tutorialGoblins.length > 0 && this.tutorialGoblins[0].life <= 3) {
                    var goblin = this.spawnSpecificMonster(this.map, new RogueRpg.Location(this.hero.x + 50, this.hero.y), "Goblin");
                    this.pointerLocation = new RogueRpg.Location(goblin.drawX - RogueRpg.Helpers.OriginX, goblin.drawY - RogueRpg.Helpers.OriginY);
                    goblin.skipLoot = true;
                    this.tutorialGoblins.push(goblin);
                    // Damage the hero so a potion can be used
                    var damageAmount = Math.floor(this.hero.maxLife * 0.4);
                    this.animateDamageTaken(damageAmount, this.hero.life);
                    this.hero.life -= damageAmount;
                    this.tutorialState++;
                    this.pointerLocation = new RogueRpg.Location(goblin.drawX - RogueRpg.Helpers.OriginX + 20, goblin.drawY - RogueRpg.Helpers.OriginY + 40);
                }
            }
            else if (this.tutorialState == 3) {
                this.tut("双击特殊攻击。");
                if (this.tutorialGoblins.length > 1 && this.tutorialGoblins[0].life <= 0 && this.tutorialGoblins[1].life <= 0) {
                    var sword = new RogueRpg.Weapon(1, 1, "Sword");
                    sword._baseDamage = 2;
                    var swordPlus = new RogueRpg.Weapon(1, 1, "Longsword");
                    swordPlus._baseDamage = 3;
                    this.dropItem(sword, this.tutorialGoblins[0].drawX, this.tutorialGoblins[0].drawY);
                    this.dropItem(swordPlus, this.tutorialGoblins[1].drawX, this.tutorialGoblins[1].drawY);
                    this.tutorialState++;
                }
            }
            else if (this.tutorialState == 4) {
                this.tut("点击掉落的物品就可以捡起来。");
                if (this.map.itemsOnGround.length > 0) {
                    this.pointerLocation = new RogueRpg.Location(this.map.itemsOnGround[0].drawX - RogueRpg.Helpers.OriginX + 30, this.map.itemsOnGround[0].drawY - RogueRpg.Helpers.OriginY);
                }
                if (this.map.itemsOnGround.length == 0) {
                    this.tutorialState++;
                }
            }
            else if (this.tutorialState == 5) {
                this.tut("打开库存");
                by.id("inventoryAccess").style.visibility = "visible";
                this.renderer.tutorialContext.strokeStyle = "red";
                this.renderer.tutorialContext.lineWidth = 4;
                this.renderer.tutorialContext.strokeRect(this.renderer.ww - 10 - 55, this.renderer.wh - 50 - 55, 60, 60);
                this.pointerLocation = new RogueRpg.Location(this.renderer.ww - 65, this.renderer.wh - 105);
                if (this.inventoryVisible) {
                    this.tutorialState++;
                }
            }
            else if (this.tutorialState == 6) {
                this.tut("点击长剑", true);
                this.renderer.tutorialContext.strokeStyle = "red";
                this.renderer.tutorialContext.lineWidth = 3;
                var bagNodes = by.id("bagSlots").childNodes;
                var swordNode = null;
                for (var i = 0; i < bagNodes.length; i++) {
                    if (bagNodes[i].textContent == "Longsword") {
                        swordNode = bagNodes.item(i);
                    }
                }
                if (swordNode) {
                    var size = swordNode.getBoundingClientRect();
                    this.renderer.tutorialContext.strokeRect(size.left - 4, size.top - 4, (size.width * 0.66) + 8, size.height + 8);
                    this.pointerLocation = new RogueRpg.Location(size.left + 30, size.top);
                }
                else {
                    this.tutorialState = 0;
                }
                if (this.displayedItem) {
                    if (this.displayedItem.fullName == "Longsword") {
                        this.tutorialState++;
                    }
                    else {
                        this.hideItemDisplay(null);
                    }
                }
            }
            else if (this.tutorialState == 7) {
                this.tut("非常好！ 点“装备”", true);
                var r = by.id("itemDisplayEquip").getBoundingClientRect();
                this.renderer.tutorialContext.strokeStyle = "red";
                this.renderer.tutorialContext.lineWidth = 3;
                this.renderer.tutorialContext.strokeRect(r.left - 4, r.top - 4, r.width + 8, r.height + 8);
                this.pointerLocation = new RogueRpg.Location(r.left, r.top);
                if (this.hero.equippedWeapon.fullName == "Longsword") {
                    this.tutorialState++;
                }
            }
            else if (this.tutorialState == 8) {
                this.tut("点击任何地方返回", true);
                this.pointerLocation = null;
                if (this.displayedItem == null) {
                    this.tutorialState++;
                }
            }
            else if (this.tutorialState == 9) {
                this.tut("点击剑", true);
                var bagNodes = by.id("bagSlots").childNodes;
                var swordNode = null;
                for (var i = 0; i < bagNodes.length; i++) {
                    if (bagNodes[i].textContent == "Sword") {
                        swordNode = bagNodes.item(i);
                    }
                }
                if (swordNode) {
                    var size = swordNode.getBoundingClientRect();
                    this.renderer.tutorialContext.strokeStyle = "red";
                    this.renderer.tutorialContext.lineWidth = 3;
                    this.renderer.tutorialContext.strokeRect(size.left - 4, size.top - 4, (size.width * 0.6) + 8, size.height + 8);
                    this.pointerLocation = new RogueRpg.Location(size.left + 30, size.top);
                }
                else {
                    this.tutorialState = 0;
                }
                if (this.displayedItem) {
                    if (this.displayedItem.fullName == "Sword") {
                        this.tutorialState++;
                    }
                    else {
                        this.hideItemDisplay(null);
                    }
                }
            }
            else if (this.tutorialState == 10) {
                this.tut("不好！ 卖它。", true);
                by.id("itemDisplayTrash").style.visibility = "visible";
                var r = by.id("itemDisplayTrash").getBoundingClientRect();
                this.renderer.tutorialContext.strokeStyle = "red";
                this.renderer.tutorialContext.lineWidth = 3;
                this.renderer.tutorialContext.strokeRect(r.left - 4, r.top - 4, r.width + 8, r.height + 8);
                this.pointerLocation = new RogueRpg.Location(r.left, r.top);
                var hasSword = false;
                for (var i = 0; i < this.hero.items.length; i++) {
                    if (this.hero.items[i].fullName == "Sword") {
                        hasSword = true;
                    }
                }
                if (!hasSword) {
                    this.tutorialState++;
                }
            }
            else if (this.tutorialState == 11) {
                this.tut("按“返回”来保留库存。", true);
                this.pointerLocation = new RogueRpg.Location(this.renderer.ww - 40, this.renderer.wh - 40);
                if (!this.inventoryVisible) {
                    this.tutorialState++;
                }
            }
            else if (this.tutorialState == 12) {
                this.tut("在战斗结束后用药水治疗。");
                this.pointerLocation = new RogueRpg.Location(this.renderer.ww - RogueRpg.Renderer.HealingPotionRightOffset + 20, this.renderer.wh - RogueRpg.Renderer.HealingPotionBottomOffset + 20);
            }
            else if (this.tutorialState == 13) {
                this.pointerLocation = null;
                if (this.renderer.tutorialCanvas) {
                    this.renderer.tutorialCanvas.parentElement.removeChild(this.renderer.tutorialCanvas);
                    this.renderer.tutorialCanvas = null;
                    this.renderer.tutorialContext = null;
                }
                //if (!this.extraDetailsDivVisible) {
                this.showExtraDetailsDiv("要到下一层楼，找这个水晶。<img src='images/Level Exit.png' />", function () {
                    _this.tutorialState++;
                });
                //}
            }
            else if (this.tutorialState == 14) {
                if (!this.extraDetailsDivVisible) {
                    this.showExtraDetailsDiv("你可以在地牢深处找到碎片。 <span class='shard-count'></span> 你会发现，在探索的过程中，你的遗物将会更加坚固。", function () {
                        _this.tutorialState++;
                    });
                }
            }
            else if (this.tutorialState == 15) {
                if (!this.extraDetailsDivVisible) {
                    this.showExtraDetailsDiv("宝石 <span class='gemstone-count'></span> 可以在地牢中找到。 使用它们在死后继续或购买强大的升级。", function () {
                        _this.tutorialState++;
                    });
                }
            }
            else if (this.tutorialState == 16) {
                if (!this.extraDetailsDivVisible) {
                    this.showExtraDetailsDiv("祝你好运。 在强大到足以深入地下城之前，您需要重复走多次。 这里有10个碎片开始！", function () {
                        _this.dropShards(10, new RogueRpg.Location(_this.hero.x, _this.hero.y), 450);
                        _this.tutorialState++;
                        _this.lockHeroMovement = false;
                        _this.tutorialCompleted = true;
                        if (_this.renderer.tutorialCanvas) {
                            _this.renderer.tutorialCanvas.parentElement.removeChild(_this.renderer.tutorialCanvas);
                            _this.renderer.tutorialCanvas = null;
                            _this.renderer.tutorialContext = null;
                        }
                        //ga_storage._trackEvent('Tutorial', "Complete");
                    });
                }
            }
            if (this.pointerLocation && this.renderer.tutorialContext) {
                var y = this.pointerLocation.y - 40 - 20 + Math.min(400, (this.time() % 1000)) / 20;
                this.renderer.tutorialContext.drawImage(RogueRpg.Images.get("Pointer"), this.pointerLocation.x, y, 40, 40);
            }
        };
        Game.prototype.checkForTimeBasedState = function () {
            var now = new Date();
            // Once per day, enable an 'update'
            if (!this.lastCompletedGameTime || this.lastCompletedGameTime.getDate() != now.getDate()) {
                this.lastCompletedGameTime = now;
                this.bonusShardsThisGame = false; // TODO: Always false true; // gain a 30% bonus on your first game of the day
            }
            if (this.flyingKickCount > 0) {
                this.flyingKickTimer += Game.deltaTime;
                if (this.flyingKickTimer > 3000) {
                    this.flyingKickTimer = 0;
                    this.flyingKickCount = 0;
                }
            }
        };
        Game.prototype.spawnSpecificMonster = function (map, location, type, skipLoot) {
            if (skipLoot === void 0) { skipLoot = false; }
            return this.spawnMonster(map, location, 0, type);
        };
        Object.defineProperty(Game, "NumLevels", {
            get: function () {
                return Game.MonstersByFloor.length;
            },
            enumerable: true,
            configurable: true
        });
        // Spawns a monster in a random location on the assigned map
        Game.prototype.spawnMonster = function (map, locationOverride, levelAdjustment, specificType, skipLoot, multiplier) {
            if (skipLoot === void 0) { skipLoot = false; }
            if (multiplier === void 0) { multiplier = 1.0; }
            var effectiveFloor = Math.max(0, (this.hero.floor - 1) + (levelAdjustment || 0));
            var mon = this.getAMonsterForFloor(effectiveFloor, map, locationOverride, specificType, multiplier);
            map.monsters.push(mon);
            mon.skipLoot = skipLoot;
            return mon;
        };
        Game.prototype.getAMonsterForFloor = function (effectiveFloor, map, locationOverride, specificType, multiplier) {
            if (multiplier === void 0) { multiplier = 1.0; }
            var monsterType = specificType || Game.MonstersByFloor[effectiveFloor] || Game.MonstersByFloor[effectiveFloor % Game.MonstersByFloor.length];
            var multiplierTier = effectiveFloor < Game.MonstersByFloor.length ? 0 : Math.floor(effectiveFloor / Game.MonstersByFloor.length);
            var mon = new RogueRpg.Monster(monsterType, map, multiplierTier, locationOverride, multiplier);
            return mon;
        };
        // Returns a location that is inside a walkable tile
        Game.prototype.getRandomWalkableTilePosition = function (map) {
            var location = new RogueRpg.Location(Math.random() * map.mapWidth * map.tileSize, Math.random() * map.mapHeight * map.tileSize);
            while (!RogueRpg.Helpers.IsWalkable(location, map, 20)) {
                var location = new RogueRpg.Location(Math.random() * map.mapWidth * map.tileSize, Math.random() * map.mapHeight * map.tileSize);
            }
            return location;
        };
        Game.prototype.handleClickOrTouch = function () {
            this._doubleTappedThisCycle = false;
            this._doubleTapLoopTimer += Game.deltaTime;
            var gameX = RogueRpg.Helpers.OriginX + this._mouseX;
            var gameY = RogueRpg.Helpers.OriginY + this._mouseY;
            if (this._tappedThisCycle) {
                this.hero.facingRight = gameX > this.hero.x;
                if (this.handleIfMouseClickedOnButton(this._mouseX, this._mouseY)) {
                    return;
                }
                else if (this.handleIfMouseClickedOnItem(gameX, gameY)) {
                    // The item has been picked up
                    this.renderer.renderItemCount = true;
                    this.needsToPersist = true;
                    return;
                }
                if (this._doubleTapLoopTimer > 0 && this._doubleTapLoopTimer < 250) {
                    this._doubleTappedThisCycle = true;
                }
                this.handleSpecialDoubleTap(gameX, gameY);
                this._doubleTapLoopTimer = 0; // Restart from this tap
            }
            this.movedThisCycle = false;
            // Note that mouse handlers are already added in initialize
            if (this._mouseDown) {
                this.hero.facingRight = gameX > this.hero.x;
                // Currently only moving is supported
                if (this.handleIfMouseIsOverUnit(gameX, gameY)) {
                    this.movingCyclesSinceLastAttack = 0;
                    return;
                }
                this.movingCyclesSinceLastAttack++;
                this.movedThisCycle = true;
                if (!this.lockHeroMovement) {
                    var speed = this.hero.speed;
                    if (RogueRpg.Level[RogueRpg.L.SpeedLimit] && speed > 35) {
                        speed = 35;
                    }
                    this.moveUnitTo(this.hero, gameX, gameY, speed);
                }
            }
        };
        // Checks to see if a straight line can be drawn from one point to another without
        // passing through any walls
        Game.prototype.wayIsClear = function (from, to) {
            var dx = (to.x - from.x);
            var dy = (to.y - from.y);
            if (dx == 0 || dy == 0) {
                return false; // Default to pathfinding, don't handle this case
            }
            // Compute the nearest map gridline intersection following the line from monster to hero
            // See if that grid block is passable. If so recurse from that point, otherwise return false
            var distToNearestX = from.x % this.map.tileSize;
            var distToNearestY = from.y % this.map.tileSize;
            if (from.x < to.x) {
                distToNearestX = this.map.tileSize - distToNearestX;
            }
            if (from.y < to.y) {
                distToNearestY = this.map.tileSize - distToNearestY;
            }
            var percentMovedX = distToNearestX / dx;
            var percentMovedY = distToNearestY / dy;
            var percentMovedToNearest = Math.min(Math.abs(percentMovedX), Math.abs(percentMovedY));
            var distanceMovedX = percentMovedToNearest * dx * 1.01;
            var distanceMovedY = percentMovedToNearest * dy * 1.01;
            var newLocation = new RogueRpg.Location(from.x + distanceMovedX, from.y + distanceMovedY);
            // If we've passed the target, return true
            // TODO: If I hide in a corner, then the monsters still overshoot me, but can't pass me so they stop moving
            var oldDistance = RogueRpg.Helpers.distance(to.x, from.x, to.y, from.y);
            var newDistance = RogueRpg.Helpers.distance(newLocation.x, to.x, newLocation.y, to.y);
            var dx = newLocation.x - to.x;
            var odx = from.x - to.x;
            if (dx > 0 && odx <= 0 || dx <= 0 && odx > 0) {
                return true;
            }
            if (oldDistance <= newDistance) {
                return true;
            }
            if (RogueRpg.Helpers.IsWalkable(newLocation, this.map)) {
                return this.wayIsClear(newLocation, to);
            }
            else {
                return false;
            }
        };
        Game.prototype.findTarget = function (dude) {
            var target = null;
            if (dude.isMonster) {
                target = this.hero;
            }
            else if (dude.isPet) {
                target = this.findNearestUnit(dude, this.map.monsters, dude.agroRange); // TODO: Find nearest monster within range
                if (!target) {
                    target = this.hero;
                }
            }
            return target;
        };
        Game.prototype.getEffectAmount = function (dude, name) {
            for (var i = 0; i < dude.activeEffects.length; i++) {
                var e = dude.activeEffects[i];
                if (e.name === name) {
                    if (e.state.amount) {
                        return e.state.amount;
                    }
                    return 1;
                }
            }
            return 0;
        };
        Game.prototype.moveDude = function (dude) {
            var targetUnit = this.findTarget(dude);
            if (!targetUnit) {
                return;
            }
            // TODO: WayIsClear calculation isn't perfect
            if (this.wayIsClear(new RogueRpg.Location(dude.x, dude.y), new RogueRpg.Location(targetUnit.x, targetUnit.y))) {
                dude.path = [new RogueRpg.Location(targetUnit.x, targetUnit.y)];
            }
            else if (!dude.path || dude.path.length === 0) {
                dude.path = this.doPathFinding(dude, targetUnit);
            }
            var goal = dude.path[0];
            if (goal && RogueRpg.Helpers.distance(dude.x, goal.x, dude.y, goal.y) <= dude.speed / 5) {
                dude.path.splice(0, 1);
                goal = dude.path[0];
                dude.path = this.doPathFinding(dude, targetUnit);
            }
            if (goal) {
                var speed = dude.speed;
                if (RogueRpg.Level[RogueRpg.L.WeakenOnHit] && dude.life < dude.maxLife) {
                    speed *= 0.9;
                }
                var slowAmt = this.getEffectAmount(dude, "Slow");
                if (slowAmt) {
                    speed *= (1 - slowAmt);
                }
                var angleRadians = Math.atan2(goal.x - dude.x, goal.y - dude.y);
                var movementX = Math.sin(angleRadians) * speed / 5;
                var movementY = Math.cos(angleRadians) * speed / 5;
                var targetLocation = new RogueRpg.Location(dude.x + movementX, dude.y + movementY);
                this.moveUnitTo(dude, goal.x, goal.y, speed);
                if (RogueRpg.Helpers.dist(dude, goal) < 50) {
                    dude.path = [];
                }
            }
        };
        Game.prototype.moveUnitTo = function (unit, x, y, speed, ignoreWalls) {
            var angleRadians = Math.atan2(x - unit.x, y - unit.y);
            var movementX = Math.sin(angleRadians) * speed / 5 * (Game.deltaTime / RogueRpg.Helpers.LoopTime);
            var movementY = Math.cos(angleRadians) * speed / 5 * (Game.deltaTime / RogueRpg.Helpers.LoopTime);
            this.moveUnitX(unit, movementX, ignoreWalls);
            this.moveUnitY(unit, movementY, ignoreWalls);
        };
        Game.prototype.moveUnitX = function (unit, movement, ignoreWalls) {
            var tileX = Math.floor(unit.x / RogueRpg.Helpers.TileSize);
            var newTileX = Math.floor((unit.x + movement) / RogueRpg.Helpers.TileSize);
            var tileY = Math.floor(unit.y / RogueRpg.Helpers.TileSize);
            unit.x += movement;
            if (!ignoreWalls && (!this.map.tiles[tileY][newTileX] || !this.map.tiles[tileY][newTileX].walkable)) {
                // Do not let the hero move into the new tile, in fact stop just short
                if (movement > 0) {
                    unit.x = newTileX * RogueRpg.Helpers.TileSize - 1;
                }
                else {
                    unit.x = tileX * RogueRpg.Helpers.TileSize + 1;
                }
            }
            if (unit.isMonster || unit.isPet) {
                unit.drawX = unit.x - unit.drawWidth / 2;
            }
            var resultingTileX = Math.floor(unit.x / RogueRpg.Helpers.TileSize);
            if (unit == this.hero && resultingTileX != tileX) {
                this.renderer.renderMinimap = true;
            }
        };
        Game.prototype.moveUnitY = function (unit, movement, ignoreWalls) {
            var tileY = Math.floor(unit.y / RogueRpg.Helpers.TileSize);
            var newTileY = Math.floor((unit.y + movement) / RogueRpg.Helpers.TileSize);
            var tileX = Math.floor(unit.x / RogueRpg.Helpers.TileSize);
            unit.y += movement;
            if (!ignoreWalls && (!this.map.tiles[newTileY] || !this.map.tiles[newTileY][tileX] || !this.map.tiles[newTileY][tileX].walkable)) {
                // Do not let the hero move into the new tile, in fact stop just short
                if (movement > 0) {
                    unit.y = newTileY * RogueRpg.Helpers.TileSize - 1;
                }
                else {
                    unit.y = tileY * RogueRpg.Helpers.TileSize + 1;
                }
            }
            if (unit.isMonster || unit.isPet) {
                unit.drawY = unit.y - unit.drawHeight;
            }
            var resultingTileY = Math.floor(unit.y / RogueRpg.Helpers.TileSize);
            if (unit == this.hero && resultingTileY != tileY) {
                this.renderer.renderMinimap = true;
            }
        };
        Game.prototype.onBackButtonPressed = function () {
            if (this.levelUpInfoPaneVisible) {
                this.hideLevelInfoPane();
            }
            else if (this._itemDisplayVisible) {
                this.hideItemDisplay(null);
            }
            else if (this.extraDetailsDivVisible) {
                this.hideExtraDetails(null);
            }
            else if (this.gemstoneStoreVisible) {
                this.hideGemstoneMenu();
            }
            else if (this.inventoryVisible) {
                this.closeInventory();
            }
        };
        Game.prototype.onMenuButtonPressed = function () {
            if (!this.inventoryVisible) {
                this.showInventory();
            }
        };
        Game.prototype.checkForNearbyHealing = function () {
            if (Math.floor(this.hero.life) < this.hero.maxLife) {
                for (var i = this.map.healingOnGround.length - 1; i >= 0; i--) {
                    var item = this.map.healingOnGround[i];
                    if (RogueRpg.Helpers.distance(item.drawX + 15, this.hero.x, item.drawY + 15, this.hero.y) <= 35) {
                        this.map.healingOnGround.splice(i, 1);
                        var amountToHeal = item.value * this.hero.maxLife;
                        RogueRpg.Mood.happy(10);
                        this.healHeroFor(amountToHeal);
                    }
                }
            }
        };
        Game.prototype.healHeroFor = function (amountToHeal) {
            this.hero.life = Math.min(RogueRpg.Level[RogueRpg.L.Bloodied] ? Math.ceil(this.hero.maxLife / 2) : this.hero.maxLife, this.hero.life + amountToHeal);
            if (amountToHeal >= 1) {
                this.floatingText.push(new RogueRpg.FloatingText("+" + RogueRpg.Helpers.formatNumber(Math.floor(amountToHeal)), new RogueRpg.Rgba(10, 200, 10), this.hero.drawX + 40, this.hero.drawY - 60, true));
            }
        };
        Game.prototype.checkForNearbyGemstonesAndShards = function () {
            for (var i = this.map.gemstonesOnGround.length - 1; i >= 0; i--) {
                var item = this.map.gemstonesOnGround[i];
                item.speed = item.speed ? item.speed + 0.03 * Game.deltaTime : 0.02;
                if (RogueRpg.Helpers.distance(item.drawX + 15, this.hero.x, item.drawY + 15, this.hero.y) <= 35) {
                    this.map.gemstonesOnGround.splice(i, 1);
                    this.setGemstones(this.gemstones + item.value);
                    this.floatingText.push(new RogueRpg.FloatingText("+" + item.value + " gemstones", new RogueRpg.Rgba(180, 50, 230), this.hero.drawX + 10, this.hero.drawY - 20, false, RogueRpg.FloatingText.DEFAULT_DURATION * 2));
                }
                else {
                    var angle = Math.atan2(this.hero.y - item.drawY, this.hero.x - item.drawX);
                    item.drawX += Math.cos(angle) * item.speed;
                    item.drawY += Math.sin(angle) * item.speed;
                }
            }
            for (var i = this.map.shardsOnGround.length - 1; i >= 0; i--) {
                var loc = this.map.shardsOnGround[i];
                if (RogueRpg.Helpers.distance(loc.x, this.hero.x, loc.y, this.hero.y) <= 5 + (loc.v || 0) * Game.deltaTime / 1000) {
                    this.map.shardsOnGround.splice(i, 1);
                    this.setShards(this.shards + 5* this.hero.state.multiplier);
                    // this._floatingText.push(new FloatingText("+" + 1 + " shard", new Rgba(30, 200, 250), this.hero.drawX + 10, this.hero.drawY - 20));
                }
                else {
                    var angle = Math.atan2(this.hero.y - loc.y, this.hero.x - loc.x);
                    loc.x += Math.cos(angle) * loc.v * Game.deltaTime / 1000;
                    loc.y += Math.sin(angle) * loc.v * Game.deltaTime / 1000;
                    loc.v = loc.v ? loc.v + Game.deltaTime / 1000 * 300 : 0.1;
                }
            }
            for (var i = 0; i < this.map.goldOnGround.length; i++) {
                var g = this.map.goldOnGround[i];
                if (RogueRpg.Helpers.dist(g, this.hero) < 80) {
                    this.grantGold(g.amount, g.x, g.y);
                    this.map.goldOnGround.splice(i, 1);
                }
            }
            for (var i = 0; i < this.map.healingOnGround.length; i++) {
                var h = this.map.healingOnGround[i];
                var location = new RogueRpg.Location(h.drawX, h.drawY);
                if (!RogueRpg.Helpers.IsWalkable(location, this.map)) {
                    var angle = Math.atan2(this.hero.y - location.y, this.hero.x - location.x);
                    h.drawX += Math.cos(angle) * 2;
                    h.drawY += Math.sin(angle) * 2;
                }
            }
        };
        Game.prototype.pickupItem = function (i) {
            this.needsToPersist = true;
            var item = this.map.itemsOnGround[i];
            this.map.itemsOnGround.splice(i, 1);
            if (this.prismaState.levels[RogueRpg.PermanentUpgrade.Upgrades.AutoItemSell.id]) {
                var itemEffect = this.calculateItemEffect(item.item);
                var isSpecial = false;
                // Don't sell weapons where the delay is less or the damage is substantially higher
                // If the player gets "master of power" or "m o swiftness" then this problem is resolved
                if (item.item instanceof RogueRpg.Weapon) {
                    var wep = item.item;
                    if (wep.delay < this.hero.equippedWeapon.delay - 100 ||
                        wep.damage > this.hero.equippedWeapon.damage * 1.5) {
                        isSpecial = true;
                    }
                }
                // Do not auto-sell items with luck or speed since those don't factor into dps or def heavily
                if (item.item.getPropertyValue("flat_luck") > 9 ||
                    item.item.getPropertyValue("flat_speed") >= 1 ||
                    item.item.getPropertyValue("percent_speed") > 10) {
                    isSpecial = true;
                }
                if (!isSpecial &&
                    itemEffect.defRatio < 1 && itemEffect.dpsRatio < 1 &&
                    itemEffect.defRatio * itemEffect.dpsRatio < 0.8) {
                    var value = this.getItemSellValue(item.item);
                    this.prismaState.gold += value;
                    this.floatingText.push(new RogueRpg.FloatingText("AutoSell +" + RogueRpg.Helpers.formatNumber(value) + " gold", new RogueRpg.Rgba(210, 180, 40), item.drawX, item.drawY, false, RogueRpg.FloatingText.DEFAULT_DURATION * 2));
                    return;
                }
            }
            this.hero.items.push(item.item);
            if (item.item instanceof RogueRpg.Weapon && !this.hero.equippedWeapon) {
                this.equipItem(item.item, null);
            }
            else if (item.item instanceof RogueRpg.Shield && !this.hero.equippedShield) {
                this.equipItem(item.item, null);
            }
            else if (item.item instanceof RogueRpg.Armor && !this.hero.equippedArmor) {
                this.equipItem(item.item, null);
            }
            else if (item.item instanceof RogueRpg.Accessory && !this.hero.equippedAccessory) {
                this.equipItem(item.item, null);
            }
        };
        Game.prototype.pickupPotion = function (i) {
            var p = this.map.potionsOnGround.splice(i, 1)[0];
            this.grantPotion(p.potion, true);
        };
        Game.prototype.grantPotion = function (potion, showText) {
            if (showText === void 0) { showText = false; }
            var granted = false;
            if (potion.name.substr(0, 7) === "Healing") {
                if (this.hero.state.healingPotionCount < this.maxHealingPotionCount) {
                    this.hero.state.healingPotionCount++;
                    granted = true;
                }
            }
            else {
                if (this.hero.state.potions.length < this.maxPotions) {
                    this.hero.state.potions.push(potion);
                    granted = true;
                }
            }
            if (!granted) {
                this.map.potionsOnGround.push(new RogueRpg.DroppedPotion(potion, new RogueRpg.Location(this.hero.x - 50 + Math.random() * 100, this.hero.y - 50 + Math.random() * 100)));
                if (showText) {
                    this.floatingText.push(new RogueRpg.FloatingText("Too many potions.", new RogueRpg.Rgba(255, 0, 0), this.hero.x, this.hero.y));
                }
            }
            else if (showText) {
                this.floatingText.push(new RogueRpg.FloatingText(potion.name, new RogueRpg.Rgba(255, 255, 255), this.hero.x, this.hero.y - 50));
            }
        };
        Game.prototype.handleIfMouseClickedOnItem = function (gameX, gameY) {
            for (var i = 0; i < this.map.itemsOnGround.length; i++) {
                var item = this.map.itemsOnGround[i];
                if (gameX > item.drawX && gameX < item.drawX + item.drawWidth && gameY > item.drawY && gameY < item.drawY + item.drawHeight) {
                    this.pickupItem(i);
                    return true;
                }
            }
            for (var i = 0; i < this.map.potionsOnGround.length; i++) {
                var p = this.map.potionsOnGround[i];
                if (gameX > p.drawLocation.x - 20 && gameX < p.drawLocation.x + 20 && gameY > p.drawLocation.y - 20 && gameY < p.drawLocation.y + 20) {
                    this.pickupPotion(i);
                    break;
                }
            }
            for (var i = 0; i < this.map.specialWorldObjects.length; i++) {
                var object = this.map.specialWorldObjects[i];
                if (gameX > object.drawX && gameX < object.drawX + object.drawWidth && gameY > object.drawY && gameY < object.drawY + object.drawHeight) {
                    if (RogueRpg.Helpers.distance(this.hero.x, object.drawX + object.drawWidth / 2, gameY, this.hero.y + object.drawHeight / 2) <= 140) {
                        if (this.interactWithWorldObjectAtIndex(i)) {
                            return true;
                        }
                    }
                }
            }
            return false;
        };
        Game.prototype.isMoving = function () {
            return this.movedThisCycle;
        };
        Game.prototype.startNextLevelAnimation = function () {
            this.advancingLevel = true; // Handle in main loop
            this.screenCaptureAfterDeath.getContext("2d").drawImage(this.renderer.canvas, 0, 0, this.renderer.ww, this.renderer.wh);
            this.screenCaptureAfterDeath.style.display = "block";
            // this.renderer.screenshotContext.drawImage(this.screenCaptureAfterDeath, 0, 0, this.renderer.ww, this.renderer.wh);
            this.renderer.resetStaticContext();
            this.advancingAnimation = RogueRpg.Images.Sprites.Advance.getInstance(new RogueRpg.Location(this.hero.x, this.hero.y - this.hero.drawHeight / 2));
            this.white = 0;
        };
        Game.prototype.advanceLevelLoop = function () {
            if (this.advancingAnimation.update(Game.deltaTime)) {
                this.renderer.context.clearRect(0, 0, this.renderer.ww, this.renderer.wh);
                this.renderer.context.fillStyle = "rgba(255, 255, 255, " + (this.white * 0.15) + ")";
                this.white++;
                this.renderer.context.fillRect(0, 0, this.renderer.ww, this.renderer.wh);
                this.advancingAnimation.draw(this.renderer.context);
            }
            else {
                this.screenCaptureAfterDeath.style.display = "none";
                this.advanceToNextLevel();
            }
        };
        Game.prototype.advanceToNextLevel = function () {
            var _this = this;
            // ga_storage._trackEvent('Floor Complete', 'Floor ' + this.hero.floor);
            this.advancingLevel = false;
            var prismaToGrant = 0;
            var nextMap;
            if (this.previousMapStack && this.previousMapStack.length > 0) {
                nextMap = this.previousMapStack.pop();
            }
            else {
                if (RogueRpg.Level[RogueRpg.L.ExpOnFloor]) {
                    this.grantExperience(this.hero.neededExperience * 0.35, true);
                }
                this.hero.floor = this.hero.floor + 1;
                if (this.hero.floor > 15 && this.hero.state.multiplier === 1 && this.maxMultiplierUnlocked === 1) {
                    this.maxMultiplierUnlocked = 2;
                    this.showExtraDetailsDiv("祝贺你到达16楼。<br /><br />有更多的文物和类解锁更深的楼层。<br /><br />您还为较硬的敌人解锁了2x模式和2x碎片。你也可以在这个模式中找到新的遗迹，模式和更强的魔药。");
                }
                else if (this.hero.floor > 15 && this.hero.state.multiplier === 2 && this.maxMultiplierUnlocked === 2) {
                    this.maxMultiplierUnlocked = 4;
                    this.showExtraDetailsDiv("祝贺您以2x模式到达16楼。<br /><br />你为疯狂的敌人解锁了4个模式和4个碎片。你可以在这个模式中找到新的，更强的药剂和模式。");
                }
                else if (this.hero.floor > 15 && this.hero.state.multiplier === 4 && this.maxMultiplierUnlocked === 4) {
                    this.maxMultiplierUnlocked = 8;
                    this.showExtraDetailsDiv("祝贺您以4x模式到达第16层。<br /><br />您已为不可能的死敌解锁8x模式和8x碎片。你可以在这个模式中找到新的，更强的药剂和模式。");
                }
                else if (this.hero.floor > 15 && this.hero.state.multiplier === 8 && this.maxMultiplierUnlocked === 8) {
                    this.maxMultiplierUnlocked = 16;
                    this.showExtraDetailsDiv("祝贺您以8x模式到达第16层。<br /><br />你已经解锁了16x模式为荒谬的硬敌人和16x碎片。");
                }
                else if (this.hero.floor > 15 && this.hero.state.multiplier === 16 && this.maxMultiplierUnlocked === 16) {
                    this.maxMultiplierUnlocked = 50;
                    this.showExtraDetailsDiv("祝贺您以16x模式到达16楼。<br /><br />您已解锁50x模式。在50x模式下，您可以找到Prisma。");
                }
                else if (this.hero.floor > 15 && this.hero.state.multiplier === 50 && this.maxMultiplierUnlocked === 50) {
                    this.maxMultiplierUnlocked = 100;
                    this.showExtraDetailsDiv("祝贺您以50倍模式到达16楼。<br /><br />您已解锁100倍模式。这种模式会降低Prisma的数量。");
                }
                else if (this.hero.floor > 15 && this.hero.state.multiplier === 100 && this.maxMultiplierUnlocked === 100) {
                    this.maxMultiplierUnlocked = 200;
                    this.showExtraDetailsDiv("祝贺您以100倍模式到达16楼。<br /><br />您已解锁200倍模式。这种模式会降低Prisma的数量。");
                }
                nextMap = new RogueRpg.MapGenerator().createLevel(this.hero.floor);
                this.healHeroFor(this.hero.floor * 10);
                if (this.hero.state.multiplier >= 50 && this.hero.floor >= 3) {
                    var prismaCount = Math.round(this.hero.state.multiplier / 50);
                    // 1x/floor til 16, 2x/floor til 32, etc. So floor 16 equals 2x harder mode. roughly
                    var multiplier = 1 + Math.floor(this.hero.floor / 16);
                    var prismaToGrant = prismaCount * multiplier;
                }
            }
            this.renderer.resetStaticContext();
            setTimeout(function () {
                _this.renderer.resetStaticContext();
            }, 50);
            // TODO: Clean this up. No longer needed because itemsOnGround is now stored on map
            //this.hero.itemsOnGround = [];
            this.map = nextMap;
            this.hero.x = this.map.heroStartX;
            this.hero.y = this.map.heroStartY;
            if (prismaToGrant > 0) {
                this.grantPrisma(prismaToGrant);
            }
            // console.log("Start in new map: {" + this.map.heroStartX + ", " + this.map.heroStartY + "} with tileSize " + this.map.tileSize);
            if (this.hero.floor == 7) {
                setTimeout(function () {
                    _this.showPleaseRatePrompt(); // Only prompts if it has never been shown
                }, 350);
            }
        };
        Game.prototype.interactWithWorldObjectAtIndex = function (index) {
            var _this = this;
            this.renderer.renderMinimap = true;
            var object = this.map.specialWorldObjects[index];
            this.needsToPersist = true;
            if (object.type == "Frozen Monk") {
                this.unlockMonkPrompt();
            }
            else if (object.type == "Unfrozen Monk") {
                if (this.highestClassUnlocked < 3) {
                    this.highestClassUnlocked = 3;
                    this.showExtraDetailsDiv("谢谢你救我。<br />在这里，带上我额外的发箍，以防你想当和尚。我会留在这里热身。");
                }
                else {
                    this.showExtraDetailsDiv("谢谢你救我。<br />我会留在这里热身。");
                }
            }
            else if (object.type === "Ghost Girl") {
                if (object.data == "used") {
                    this.showExtraDetailsDiv("请为我的死报仇。这个地牢里有杀死我的罪魁祸首。");
                }
                else {
                    this.showDialogPrompt("你好。我跟你一样是个冒险家。<br /><br />哇，你有一个 <i>relic</i>? 真幸运。你可以用我的碎片。", "好", function () {
                        object.data = "used";
                        _this.setShards(_this.shards + _this.floor * 8 * _this.hero.state.multiplier);
                        _this.floatingText.push(new RogueRpg.FloatingText("+" + (_this.floor * 8 * _this.hero.state.multiplier) + " shards", new RogueRpg.Rgba(30, 200, 250), object.drawX + 10, object.drawY - 20));
                    }, "不");
                }
            }
            else if (object.type == "Portal") {
                this.beginSubLevel("Earth");
                object.type = "Portal Empty";
            }
            else if (object.type == "Sitting Wizard") {
                if (!object.data) {
                    object.data = RogueRpg.Helpers.SelectOne(["plane", "gemstones", "potions"]);
                }
                if (object.data == "plane") {
                    this.showDialogPrompt("我刚从最恐怖的地方回来！我不想谈这件事。<br /><br />如果你有死亡的愿望, 我可以送你去那里。我必须警告你，这是 <i>very</i> 危险的.", "好", function () {
                        object.data = "used";
                        _this.beginSubLevel("Magma");
                    }, "不行");
                }
                else if (object.data == "gemstones") {
                    this.showDialogPrompt("我看到了一个惊人的秘密！我不能告诉你，但也许有一天你也会学的。<br /><br />我不再需要这些了。你想要我的额外宝石吗？", "是的", function () {
                        object.data = "used";
                        _this.map.gemstonesOnGround.push(new RogueRpg.SpecialDroppedItem(randInt(5, 10), object.drawX + object.drawWidth / 2, object.drawY + object.drawHeight / 2, "gemstones", 0));
                    }, "不 谢谢");
                }
                else if (object.data == "potions") {
                    this.showExtraDetailsDiv("我来到这个洞穴寻找配料。 我发现远远超出了我的预期!<br /><br />我有太多的药剂要做。也许你会给我带一些?", function () {
                        object.data = "used";
                        var count = randInt(2, 3);
                        for (var i = 0; i < count; i++) {
                            var potion = Game.randomPotion();
                            _this.grantPotion(potion);
                        }
                    });
                }
                else {
                    this.showExtraDetailsDiv("快跑<br /><br />谢谢你的关心，但我会在这里好的。 这些怪物似乎喜欢我。");
                }
                // Ask if you want to go to another world
                // If so, then cache the current map and position
                // Generate a new map
            }
            else if (object.type === "Potion Slot") {
                this.maxPotions++;
                this.showExtraDetailsDiv("恭喜你!你已经打开了一个新的药剂槽。您现在可以保存" + this.maxPotions + " 一次药水!");
                object.type = "Pedestal";
                object.drawWidth = 40;
                object.drawHeight = 70;
                object.y = object.drawY + object.drawHeight;
                return true;
            }
            else if (object.type.indexOf("Scumbasher") >= 0) {
                this.showExtraDetailsDiv("恭喜你!你的启动武器现在更强大了。它将自动装备在所有未来的游戏中。");
                this.startingWeapon = object.type;
                object.type = "Pedestal";
                object.drawWidth = 40;
                object.drawHeight = 70;
                object.y = object.drawY + object.drawHeight; // Down from 100?
                return true;
            }
            else if (RogueRpg.Relic.FindRelicIndex(object.type.split(" World")[0]) >= 0) {
                var msg = "恭喜你!你发现了一个新的遗迹!你可以在游戏中使用碎片来改善你的遗物。";
                this.showExtraDetailsDiv(msg);
                var index = RogueRpg.Relic.FindRelicIndex(object.type.split(" World")[0]);
                RogueRpg.Relic.RelicLevels[index] = 1;
                object.type = "Pedestal";
                object.drawWidth = 40;
                object.drawHeight = 70;
                object.y = object.drawY + object.drawHeight; // Down from 100?
                return true;
            }
            else if (object.type == "Bowman Cap Item") {
                this.showExtraDetailsDiv("恭喜你!你已经解锁了恶棍!当你开始一项新游戏时，你可以选择“恶棍”");
                this.map.specialWorldObjects.splice(this.map.specialWorldObjects.indexOf(object), 1);
                this.highestClassUnlocked = 1;
                return true;
            }
            else if (object.type == "Wizard Hat Item") {
                this.showExtraDetailsDiv("恭喜你!你已经解锁了法师!当你开始一项新游戏时，你可以选择“法师”");
                this.map.specialWorldObjects.splice(this.map.specialWorldObjects.indexOf(object), 1);
                this.highestClassUnlocked = 2;
            }
            else if (object.type == "Level Exit") {
                this.startNextLevelAnimation();
                return true;
            }
            else if (object.type == "Healing Shrine") {
                this.healHeroFor(this.hero.maxLife);
                RogueRpg.Mood.happy(20);
                object.type = "Used Healing Shrine";
                return true;
            }
            else if (object.type == "Treasure Chest" || object.type == "Dark Treasure Chest") {
                this.openChest(object);
                return true;
            }
            else if (object.type == "Filled Mine Cart") {
                this.dropShards(3, new RogueRpg.Location(object.drawX + object.drawWidth / 2, object.drawY + object.drawHeight / 2), 200);
                object.type = "Empty Mine Cart";
                return true;
            }
            else if (object.type == "Item Upgrade Shrine") {
                var roll = Math.random();
                var item;
                var upgraded = false;
                RogueRpg.Mood.happy(40);
                while (!upgraded) {
                    if (roll < 0.25) {
                        item = RogueRpg.Item.generateEquipment("Weapon", this.hero.equippedWeapon.level, Math.min(7, this.hero.equippedWeapon.rarity + 1), this.hero.equippedWeapon.type);
                        upgraded = true;
                    }
                    else if (roll < 0.5) {
                        if (this.hero.equippedShield) {
                            item = RogueRpg.Item.generateEquipment("Shield", this.hero.equippedShield.level, Math.min(7, this.hero.equippedShield.rarity + 1), this.hero.equippedShield.type);
                            upgraded = true;
                        }
                    }
                    else if (roll < 0.75) {
                        if (this.hero.equippedArmor) {
                            item = RogueRpg.Item.generateEquipment("Armor", this.hero.equippedArmor.level, Math.min(7, this.hero.equippedArmor.rarity + 1), this.hero.equippedArmor.type);
                            upgraded = true;
                        }
                    }
                    else {
                        if (this.hero.equippedAccessory) {
                            item = RogueRpg.Item.generateEquipment("Accessory", this.hero.equippedAccessory.level, Math.min(7, this.hero.equippedAccessory.rarity + 1), this.hero.equippedAccessory.type);
                            upgraded = true;
                        }
                    }
                    roll = Math.random(); // In case upgraded === false
                }
                if (item) {
                    var drop = new RogueRpg.DroppedItem();
                    drop.item = item;
                    drop.drawX = object.drawX + 20;
                    drop.drawY = object.drawY + 70;
                    drop.state = drop.getItemState();
                    if (item.rarity >= 6) {
                        var width = this.renderer.context.measureText(drop.item.fullName).width;
                        this.activeAnimations.push(new RogueRpg.Animation("rarity6glow", drop.drawX + (width / 2), drop.drawY - 293, 2000));
                    }
                    this.map.itemsOnGround.push(drop);
                    object.type = "Used Item Upgrade Shrine";
                    return true;
                }
            }
            else if (object.type == "Note") {
                this.showExtraDetailsDiv(object.data);
                this.map.specialWorldObjects.splice(this.map.specialWorldObjects.indexOf(object), 1);
                return true;
            }
            else if (object.type == "Alchemy Table") {
                this.showAlchemyPane();
                return true;
            }
            else if (object.type === "Prisma Found") {
                var prismaCount = randIntMap(2, 7) * Math.round(this.hero.state.multiplier / 50);
                var multiplier = 1 + Math.floor(this.hero.floor / 16);
                this.map.specialWorldObjects.splice(this.map.specialWorldObjects.indexOf(object), 1);
                this.grantPrisma(prismaCount * multiplier);
            }
        };
        Game.prototype.grantPrisma = function (count) {
            this.prismaState.prisma += count;
            this.floatingText.push(new RogueRpg.FloatingText("+" + count + " prisma", new RogueRpg.Rgba(255, 255, 255), this.hero.drawX + 10, this.hero.drawY - 20, false, RogueRpg.FloatingText.DEFAULT_DURATION * 2));
        };
        Game.prototype.grantGold = function (count, atX, atY) {
            this.prismaState.gold += count;
            if (atX && atY) {
                this.floatingText.push(new RogueRpg.FloatingText("+" + RogueRpg.Helpers.formatNumber(count) + " gold", new RogueRpg.Rgba(230, 210, 70), atX, atY, false, RogueRpg.FloatingText.DEFAULT_DURATION));
            }
            else {
                this.goldAccumulation += count;
            }
        };
        Game.prototype.openChest = function (object) {
            var itemOffsetY = 120;
            var item = null;
            var sampleMonster = this.map.monsters[0] || this.getAMonsterForFloor(this.hero.floor);
            var itemLevel = Math.min(this.hero.level, sampleMonster.level);
            if (object.type === "Treasure Chest") {
                item = this.rollForDrop(itemLevel, /*50% for guaranteed rare*/ this.playerRandom(true) > 0.5, object.drawX, object.drawY + itemOffsetY);
                item = this.rollForDrop(itemLevel, false, object.drawX, object.drawY + itemOffsetY) || item;
                item = this.rollForDrop(itemLevel, false, object.drawX, object.drawY + itemOffsetY) || item;
            }
            else if (object.type === "Dark Treasure Chest") {
                if (Math.random() > 0.5) {
                    item = this.rollForDrop(itemLevel + 2, true, object.drawX, object.drawY + itemOffsetY, "Weapon", "Blade of the Dreadking");
                }
                else {
                    item = this.rollForDrop(itemLevel + 2, true, object.drawX, object.drawY + itemOffsetY, "Armor", "Dark Armor");
                }
            }
            var gemstonesFound = Math.random() > 0.8 && this.playerRandom(true) > 0.75;
            if (gemstonesFound) {
                this.dropGemstones(5 + Math.floor(this.floor / 4), false, object.drawX - 50 + Math.random() * 100, object.drawY - 50 + Math.random() * 100);
            }
            if (item || gemstonesFound) {
                RogueRpg.Mood.happy(30);
            }
            else {
                RogueRpg.Mood.sad(30);
            }
            object.type = "Used " + object.type;
            this.spriteAnimations.push(RogueRpg.Images.Sprites.Smoke.getInstance(new RogueRpg.Location(object.drawX + object.drawWidth / 2, object.drawY + object.drawHeight / 2)));
        };
        Game.randomPotion = function () {
            // The strongest potion that can drop will have a multiplier no higher than the floor
            // But after a multiplier of 4, increases every other floor
            // And after 
            var mult = RogueRpg.Helpers.hero.state.multiplier;
            mult = Math.min(mult, RogueRpg.Helpers.hero.floor);
            // Multipliers of 16x+ are insane! Scale down a bit. In this case, 'mult' is effectively 'floor'
            // so it goes 1 2 3 4 4 5 5 6 6 7 7 7 7 8 8 8 8 9 9 9 9 10 10 10 10 ...
            if (mult > 10) {
                mult = 7 + Math.floor((mult - 10) / 4);
            }
            else if (mult > 4) {
                mult = 4 + Math.floor((mult - 4) / 2);
            }
            // Finally, randomize it a bit, sometimes
            if (mult >= 2 && Math.random() > 0.5) {
                mult = 2 + Math.floor(Math.random() * (mult - 1));
            }
            var healingPotionChance = 0.45;
            if (mult > 2) {
                healingPotionChance -= 0.05;
            }
            if (mult > 5) {
                healingPotionChance -= 0.1;
            }
            if (Math.random() < healingPotionChance) {
                return new RogueRpg.Potion(RogueRpg.PotionEffect.Healing, 40);
            }
            if (Math.random() < 0.06 + Math.min(mult, 4) * 0.01) {
                return new RogueRpg.Potion(RogueRpg.PotionEffect.Clear);
            }
            if (Math.random() < 0.2) {
                return new RogueRpg.Potion(RogueRpg.PotionEffect.Maxlife, randInt(12, 20) * mult);
            }
            if (Math.random() < 0.15) {
                return new RogueRpg.Potion(RogueRpg.PotionEffect.Speed, randInt(1, 2));
            }
            if (mult >= 4 && Math.random() < 0.18) {
                return new RogueRpg.Potion(RogueRpg.PotionEffect.SkillPoint, 1);
            }
            if (mult >= 2 && Math.random() < 0.5) {
                if (Math.random() < 0.33) {
                    return new RogueRpg.Potion(RogueRpg.PotionEffect.Damage, randInt(1, 2) * mult);
                }
                else if (Math.random() < 0.33) {
                    return new RogueRpg.Potion(RogueRpg.PotionEffect.Reduction, randInt(1, 3) * mult);
                }
                else {
                    return new RogueRpg.Potion(RogueRpg.PotionEffect.Defense, randInt(1, 3) * mult);
                }
            }
            var type = RogueRpg.Helpers.SelectOne([
                RogueRpg.PotionEffect.Dex, RogueRpg.PotionEffect.Luck, RogueRpg.PotionEffect.Strength
            ]);
            if (mult > 5 && type === RogueRpg.PotionEffect.Luck) {
                mult = 5;
            }
            return new RogueRpg.Potion(type, 3 + randInt(1, 4) * mult);
        };
        Game.prototype.rollForDrop = function (level, isMonsterRare, drawX, drawY, itemCategory, itemType) {
            if (RogueRpg.Level[RogueRpg.L.DoubleExp]) {
                level = this.hero.level;
            }
            if (Math.random() > 0.9 - (0.1 * (RogueRpg.Level[RogueRpg.L.HeartDropChance] || 0))) {
                this.dropHeart(drawX, drawY);
            }
            var roll = Math.random(); //this.playerRandom(true); // Having more luck no longer means more items, only better items
            if (isMonsterRare || (this.hero.level == 1 && this.playerRandom(true) > 0.5)) {
                roll = 1; // Rare monsters always drop an item, or if you're level 1
            }
            if (roll > 0.75) {
                var type, quality;
                var rarity;
                if (!itemCategory) {
                    var typeRoll = Math.random();
                    if (typeRoll < .35) {
                        itemCategory = "Weapon";
                    }
                    else if (typeRoll < .59) {
                        itemCategory = "Armor";
                    }
                    else if (typeRoll < .80) {
                        itemCategory = "Shield";
                    }
                    else {
                        itemCategory = "Accessory"; // 20% rings
                    }
                }
                var qualityRoll = this.playerRandom(true);
                if (qualityRoll < .6) {
                    quality = level;
                }
                else if (qualityRoll < .9) {
                    quality = level + 1;
                }
                else {
                    quality = level + 2;
                }
                var rarityTable = [.25, .75, .95, .988, .998];
                var rarityRoll = this.playerRandom(true);
                // Rare monsters cannot drop less than a rare item
                while (isMonsterRare && rarityRoll < rarityTable[1]) {
                    rarityRoll = this.playerRandom(true);
                }
                /*if (this.hero.equippedHat == Hat.Hats["Fedora"]) {
                    for (var i = 0; i < rarityTable.length; i++) {
                        rarityTable[i] = rarityTable[i] - 0.003;
                    }
                }*/
                if (rarityRoll < rarityTable[0]) {
                    rarity = 1;
                    //if (this.hero.equippedHat == Hat.Hats["Wizard Hat"]) {return;}
                }
                else if (rarityRoll < rarityTable[1]) {
                    rarity = 2;
                }
                else if (rarityRoll < rarityTable[2] || RogueRpg.Level[RogueRpg.L.DoubleExp]) {
                    rarity = 3;
                }
                else if (rarityRoll < rarityTable[3]) {
                    rarity = 4;
                }
                else if (rarityRoll < rarityTable[4]) {
                    rarity = 5;
                    RogueRpg.Mood.happy(20);
                }
                else {
                    rarity = 6;
                    RogueRpg.Mood.happy(50);
                }
                if (this.floor > 15) {
                    if (rarity == 1 && Math.random() < 0.75) {
                        return;
                    }
                    if (rarity == 2 && Math.random() < 0.5) {
                        return;
                    }
                }
                var item = RogueRpg.Item.generateEquipment(itemCategory, quality, rarity, itemType);
                if (itemCategory === "Weapon") {
                    var prismaWeaponRanks = this.prismaState.levels["StrongerWeapons"] || 0;
                    if (prismaWeaponRanks > 0) {
                        var wep = item;
                        wep._baseDamage *= 1 + (prismaWeaponRanks * RogueRpg.PermanentUpgrade.Upgrades.StrongerWeapons.gainPerRank * 0.01); // Convert from Pct
                    }
                }
                if (itemCategory === "Armor") {
                    var prismaArmorRanks = this.prismaState.levels["StrongerArmors"] || 0;
                    if (prismaArmorRanks > 0) {
                        var wep = item;
                        wep._baseDamage *= 1 + (prismaArmorRanks * RogueRpg.PermanentUpgrade.Upgrades.StrongerArmors.gainPerRank * 0.01); // Convert from Pct
                    }
                }
                this.dropItem(item, drawX, drawY);
                return item;
            }
        };
        // TODO: Animate the drop process?
        Game.prototype.dropShards = function (count, location, spread) {
            for (var i = 0; i < count; i++) {
                var shard = new RogueRpg.MovingLocation(location.x + Math.random() * spread - spread / 2, location.y + Math.random() * spread - spread / 2);
                shard.v = 0;
                this.map.shardsOnGround.push(shard);
            }
        };
        Game.prototype.dropHeart = function (drawX, drawY) {
            var item = new RogueRpg.SpecialDroppedItem(0.25, drawX + 20, drawY + 40, "heart");
            this.map.healingOnGround.push(item);
        };
        // Responsible for placing an item on the ground and into the _itemsOnGround array to be picked up
        Game.prototype.dropItem = function (item, drawX, drawY) {
            var _this = this;
            var drop = new RogueRpg.DroppedItem();
            drop.item = item;
            drop.drawX = drawX - 40 + Math.random() * 80;
            drop.drawY = drawY - 30 + Math.random() * 80;
            drop.state = drop.getItemState();
            if (drop.item.rarity == 6) {
                var width = this.renderer.context.measureText(drop.item.fullName).width;
                this.activeAnimations.push(new RogueRpg.Animation("rarity6glow", drop.drawX + (width / 2), drop.drawY - 293, 3000));
            }
            this.map.itemsOnGround.push(drop);
            if (this.map.itemsOnGround.length > RogueRpg.Helpers.MaxItemCount) {
                this.map.itemsOnGround.splice(0, 1);
            }
            if (this.prismaState.levels[RogueRpg.PermanentUpgrade.Upgrades.AutoItemPickup.id]) {
                setTimeout(function () {
                    var index = _this.map.itemsOnGround.indexOf(drop);
                    if (index >= 0) {
                        _this.pickupItem(index);
                    }
                }, 600);
            }
        };
        Game.prototype.calculateItemEffect = function (item) {
            var originalDps = this.approximateDps();
            var originalDef = this.approximateDef();
            var newDps = originalDps;
            var newDef = originalDef;
            this.hero.clearStatCaches();
            if (item instanceof RogueRpg.Weapon) {
                var oldWeapon = this.hero.equippedWeapon;
                this.hero.equippedWeapon = item;
                newDps = this.approximateDps();
                newDef = this.approximateDef();
                this.hero.equippedWeapon = oldWeapon;
            }
            else if (item instanceof RogueRpg.Shield) {
                var oldShield = this.hero.equippedShield;
                this.hero.equippedShield = item;
                newDps = this.approximateDps();
                newDef = this.approximateDef();
                this.hero.equippedShield = oldShield;
            }
            else if (item instanceof RogueRpg.Armor) {
                var oldArmor = this.hero.equippedArmor;
                this.hero.equippedArmor = item;
                newDps = this.approximateDps();
                newDef = this.approximateDef();
                this.hero.equippedArmor = oldArmor;
            }
            else if (item instanceof RogueRpg.Accessory) {
                var oldAccessory = this.hero.equippedAccessory;
                this.hero.equippedAccessory = item;
                newDps = this.approximateDps();
                newDef = this.approximateDef();
                this.hero.equippedAccessory = oldAccessory;
            }
            this.hero.clearStatCaches();
            var effect = new RogueRpg.ItemEffect();
            originalDps = Math.max(1, originalDps);
            originalDef = Math.max(1, originalDef);
            effect.dpsRatio = Math.max(0.01, newDps / originalDps);
            effect.defRatio = Math.max(0.01, newDef / originalDef);
            return effect;
        };
        Game.prototype.approximateDps = function () {
            var damage = this.hero.weaponDamage;
            var delay = this.hero.weaponDelay;
            var dps = damage / (delay / 1000);
            if (RogueRpg.Level[RogueRpg.L.LuckyStrike]) {
                dps *= (this.expectedValueFromCurrentLuck() * 2); // Account for a skewed damage range
            }
            else {
                dps *= (this.expectedValueFromCurrentLuck() + 0.5); // Account for a skewed damage range
            }
            var critChance = this.hero.critChance;
            var critDamageBonus = 1.0 + RogueRpg.Level[RogueRpg.L.IncreasedCritDamage];
            dps *= 1 + critChance * critDamageBonus;
            return dps;
        };
        Game.prototype.approximateDef = function () {
            var eLife = this.hero.maxLife;
            if (!RogueRpg.Level[RogueRpg.L.NoLifesteal]) {
                eLife += 15 * (this.hero.getLifestealPercent() / 100) * this.approximateDps();
            }
            eLife = eLife / Math.max(0.25, (1 - this.hero.reduction / this.getEnemyDamageForLevel(this.hero.level))); // Avoid insanely large numbers
            eLife = eLife / (1 - this.hero.getDefenseReductionPercent(this.hero.level) / 100);
            return eLife;
        };
        // TODO: There is probably a better way to do this...
        Game.prototype.getEnemyDamageForLevel = function (level) {
            var dmg = 1;
            if (level <= 1)
                dmg = 2;
            else if (level <= 3)
                dmg = 8;
            else if (level <= 5)
                dmg = 20;
            else if (level <= 11)
                dmg = 35;
            else if (level <= 14)
                dmg = 100;
            else if (level <= 17)
                dmg = 150;
            else if (level <= 20)
                dmg = 250;
            else if (level <= 23)
                dmg = 350;
            else
                dmg = 10000000; // Otherwise, reduction is negligible so whatever
            return dmg * this.hero.state.multiplier;
        };
        Game.prototype.grantExperience = function (amount, ignoreMultipliers) {
            this.renderer.renderLevelBar = true;
            if (!ignoreMultipliers) {
                amount *= this.hero.getExperienceMultiplier();
            }
            amount = Math.round(amount);
            this.hero.currentExperience += amount;
            while (this.hero.currentExperience >= this.hero.neededExperience) {
                this.hero.currentExperience -= this.hero.neededExperience;
                this.hero.levelUp();
                // TODO: Animate before doing this
                if (RogueRpg.Level[RogueRpg.L.ExplodeOnLevel]) {
                    this.spriteAnimations.push(RogueRpg.Images.Sprites.Explosion.getInstance(new RogueRpg.Location(this.hero.drawX + this.hero.drawWidth / 1.5, this.hero.drawY - 10)));
                    this.damageArea(this.hero.x, this.hero.y, 3 * this.hero.weaponDamage * (1000 / this.hero.weaponDelay), 200, false);
                }
                else {
                    // this._spriteAnimations.push(Images.Sprites.LevelUp.getInstance(new Location(this.hero.drawX + this.hero.drawWidth / 1.5, this.hero.drawY - 10)));
                    this.renderer.animateLevelUpTimer = 4000;
                }
                // TODO: Show +life, +str, +dex, etc.
            }
        };
        // Updates units like pets, destructibles, etc
        Game.prototype.updateOtherUnits = function () {
            var _this = this;
            if (RogueRpg.Level[RogueRpg.L.Orb]) {
                var ORB_RADIUS = 37;
                if (typeof this.hero.orbAngle != 'number') {
                    this.hero.orbAngle = 0;
                }
                this.hero.orbAngle += Math.PI / 90;
                // Assume 1 rank for now
                // TODO: Note that there is some y-axis fudging going on to get overlap correct
                this.hero.orbPositions = [];
                for (var i = 0; i < RogueRpg.Level[RogueRpg.L.Orb]; i++) {
                    var angle = this.hero.orbAngle + i * (Math.PI * 2) / RogueRpg.Level[RogueRpg.L.Orb];
                    var loc = new RogueRpg.Location(this.hero.x + Math.cos(angle) * ORB_RADIUS, this.hero.y - 21 + Math.sin(angle) * ORB_RADIUS);
                    this.hero.orbPositions.push(loc);
                    this.damageArea(loc.x, loc.y - 20, Math.ceil(0.2 * this.hero.weaponDamage), 15, false, true, function (monster) {
                        monster.orbHitTimer = monster.orbHitTimer || 0;
                        monster.orbHitTimer -= Game.deltaTime;
                        if (monster.orbHitTimer <= 0) {
                            monster.orbHitTimer = _this.hero.weaponDelay / 3;
                            return true;
                        }
                        return false;
                    });
                }
            }
        };
        Game.prototype._updatePets = function () {
            this.hero.state.pets = this.hero.state.pets || [];
            for (var i = 0; i < this.hero.state.pets.length; i++) {
                var pet = this.hero.state.pets[i];
                if (RogueRpg.Helpers.dist(pet, this.hero) > 2000) {
                    pet.x = this.hero.x;
                    pet.y = this.hero.y;
                    pet.drawX = this.hero.drawX;
                    pet.drawY = this.hero.drawY;
                }
                pet.movingTimer = pet.movingTimer + Game.deltaTime;
                pet.relatiateTimer -= Game.deltaTime;
                if (pet.lifeTimer) {
                    pet.lifeTimer -= Game.deltaTime;
                    if (pet.lifeTimer < 0) {
                        this.hero.state.pets.splice(i);
                        i--;
                    }
                }
                this.moveDude(pet);
                var attackTarget = this.findNearestUnit(pet, this.map.monsters, pet.attackRange);
                if (attackTarget && pet.relatiateTimer < 0) {
                    pet.relatiateTimer = pet.delay;
                    pet.attackingTimer = pet.delay / 5;
                    this.damageMonster(attackTarget, pet.damage, false);
                }
            }
        };
        Game.prototype.updateMonsterActiveEffects = function (monster) {
            for (var i = 0; i < monster.activeEffects.length; i++) {
                var e = monster.activeEffects[i];
                e.timeLeft -= Game.deltaTime;
                if (e.timeLeft < 0) {
                    monster.activeEffects.splice(i, 1);
                    i--;
                }
            }
        };
        Game.prototype._updateMonsters = function () {
            var attacked = false;
            var agroMons = [];
            for (var i = 0; i < this.map.monsters.length; i++) {
                var monster = this.map.monsters[i];
                this.updateMonsterActiveEffects(monster);
                if (monster.tookDamageTimer > 0) {
                    monster.tookDamageTimer -= Game.deltaTime;
                }
                if (this.getEffectAmount(monster, "Stun")) {
                    continue;
                }
                var distance = RogueRpg.Helpers.distance(monster.x, this.hero.x, monster.y, this.hero.y);
                if (distance < RogueRpg.Helpers.DefaultAgroRadius && !monster.isAgroed) {
                    monster.path = this.doPathFinding(monster, this.hero);
                    monster.isAgroed = true;
                }
                if (monster.isAgroed) {
                    if (!monster.movingTimer) {
                        monster.movingTimer = 0;
                    }
                    monster.movingTimer = monster.movingTimer + Game.deltaTime;
                    if (distance < RogueRpg.Helpers.DefaultFollowRadius) {
                        if (distance > monster.attackRange - 10) {
                            this.moveDude(monster);
                            monster.movedThisLoop = true;
                        }
                        else {
                            monster.movedThisLoop = false;
                        }
                        agroMons.push(monster);
                        monster.facingRight = (monster.x < this.hero.x);
                    }
                }
                for (var j = 0; j < this.map.monsters.length; j++) {
                    var mon2 = this.map.monsters[j];
                    if (!mon2.isAgroed && mon2 != monster) {
                        var m2mDist = RogueRpg.Helpers.distance(monster.x, mon2.x, monster.y, mon2.y);
                        if (m2mDist < RogueRpg.Helpers.DefaultAssistRadius) {
                            mon2.isAgroed = true;
                            mon2.path = this.doPathFinding(mon2, this.hero);
                        }
                    }
                }
                if (distance < monster.attackRange) {
                    this.updateMonsterAttackCycle(monster);
                    attacked = true;
                }
            }
            // TODO: This needs to include pets too!
            this.handleCollisions(agroMons);
        };
        Game.prototype.handleCollisions = function (agroMonsters) {
            var units = agroMonsters; //(<Dude[]>agroMonsters).concat(this.hero.state.pets);
            for (var i = 0; i < units.length; i++) {
                var m1 = units[i];
                for (var j = i + 1; j < units.length; j++) {
                    var m2 = units[j];
                    var dist = RogueRpg.Helpers.dist(m1, m2);
                    var radius = m1.drawWidth / 4 + m2.drawWidth / 4;
                    if (dist < radius) {
                        var dx = m2.x - m1.x;
                        var dy = m2.y - m1.y;
                        var angle = Math.atan2(dy, dx);
                        //angle += 0.3;
                        // Power is inversely proportional to dist
                        // power caps out at monster speed
                        // power bottoms out at a portion of monster speed
                        var speed2 = m1.speed + m2.speed;
                        var power = Math.max(speed2, 0.4 * speed2 * (radius / (dist + 5)));
                        // Ocassionally fudge power to avoid stalemate situations
                        if (Math.random() > 0.95) {
                            power = power * 0.85;
                        }
                        this.moveUnitTo(m1, m1.x - Math.cos(angle) * power / 2, m1.y - Math.sin(angle) * power / 2, m1.speed);
                        this.moveUnitTo(m2, m2.x + Math.cos(angle) * power / 2, m2.y + Math.sin(angle) * power / 2, m2.speed);
                    }
                }
                var dist = RogueRpg.Helpers.dist(m1, this.hero);
                var radius = this.hero.drawWidth * 0.8;
                if (dist < radius) {
                    var dx = this.hero.x - m1.x;
                    var dy = this.hero.y - m1.y;
                    var angle = Math.atan2(dy, dx);
                    var power = Math.max(m1.speed, 0.8 * m1.speed * (radius / (dist + 5)));
                    this.moveUnitTo(m1, m1.x - Math.cos(angle) * power / 2, m1.y - Math.sin(angle) * power / 2, m1.speed);
                }
            }
        };
        Game.prototype.doPathFinding = function (from, to) {
            var map = this.map.tiles;
            var start = RogueRpg.Helpers.ToTileLocation(from.x, from.y);
            var goal = RogueRpg.Helpers.ToTileLocation(to.x, to.y);
            var frontier = new RogueRpg.Queue(); // An array where tiles are pushed
            frontier.enqueue(start);
            var cameFrom = Make2dArray(this.map.mapWidth, this.map.mapHeight);
            while (!frontier.empty()) {
                var current = frontier.dequeue();
                if (RogueRpg.Location.equals(current, goal)) {
                    break;
                }
                var neighbors = this.FindNeighbors(current);
                for (var i = 0; i < neighbors.length; i++) {
                    var next = neighbors[i];
                    if (map[next.y][next.x].walkable && !cameFrom[next.x][next.y]) {
                        frontier.enqueue(next);
                        cameFrom[next.x][next.y] = current;
                        if (RogueRpg.Location.equals(current, goal)) {
                            debugger;
                        }
                    }
                }
            }
            var current = goal;
            var path = [new RogueRpg.Location((current.x + 0.5) * this.map.tileSize, (current.y + 0.5) * this.map.tileSize)];
            while (current && !RogueRpg.Location.equals(current, start)) {
                current = cameFrom[current.x][current.y];
                if (current && !RogueRpg.Location.equals(current, start)) {
                    path.push(new RogueRpg.Location((current.x + 0.5) * this.map.tileSize, (current.y + 0.5) * this.map.tileSize));
                }
            }
            path.reverse();
            // Path now represents the shortest path. How do I translate this into actual movement?
            return path;
        };
        Game.prototype.FindNeighbors = function (loc) {
            var ns = [];
            if (loc.x > 0) {
                ns.push(new RogueRpg.Location(loc.x - 1, loc.y));
            }
            if (loc.x < this.map.mapWidth - 1) {
                ns.push(new RogueRpg.Location(loc.x + 1, loc.y));
            }
            if (loc.y > 0) {
                ns.push(new RogueRpg.Location(loc.x, loc.y - 1));
            }
            if (loc.y < this.map.mapHeight - 1) {
                ns.push(new RogueRpg.Location(loc.x, loc.y + 1));
            }
            return ns;
        };
        Game.prototype.updateMonsterAttackCycle = function (monster) {
            monster.relatiateTimer -= Game.deltaTime;
            monster.wasHitTimer -= Game.deltaTime;
            if (monster.relatiateTimer <= 0) {
                monster.relatiateTimer = monster.delay;
                var monsterIndex = this.map.monsters.indexOf(monster);
                var blockDamage = false;
                if (RogueRpg.Level[RogueRpg.L.CounterattackAOE]) {
                    var rand = RogueRpg.Level[RogueRpg.L.LuckOnCombatEffects] ? this.playerRandom() : Math.random();
                    if (rand > 1 - 0.07 * RogueRpg.Level[RogueRpg.L.CounterattackAOE]) {
                        this.damageArea(this.hero.x, this.hero.y, this.applyRange(this.hero.weaponDamage) * 0.5, 100, false);
                        this.floatingText.push(new RogueRpg.FloatingText("Counterattack!", new RogueRpg.Rgba(255, 0, 105), this.hero.drawX - 5, this.hero.drawY - 45, true));
                    }
                }
                if (RogueRpg.Level[RogueRpg.L.Riposte]) {
                    var rand = RogueRpg.Level[RogueRpg.L.LuckOnCombatEffects] ? this.playerRandom() : Math.random();
                    if (rand > 1 - 0.07 * RogueRpg.Level[RogueRpg.L.Riposte]) {
                        this.damageMonster(monster, this.applyRange(this.hero.weaponDamage));
                        blockDamage = true;
                        this.floatingText.push(new RogueRpg.FloatingText("Riposte!", new RogueRpg.Rgba(185, 0, 165), this.hero.drawX - 5, this.hero.drawY - 45, true));
                    }
                }
                // There is now an 8% chance to block any incoming damage by default when using a shield
                if (this.hero.equippedShield && !blockDamage) {
                    var rand = RogueRpg.Level[RogueRpg.L.LuckOnCombatEffects] ? this.playerRandom() : Math.random();
                    if (rand > 0.92) {
                        blockDamage = true;
                        this.floatingText.push(new RogueRpg.FloatingText("Blocked!", new RogueRpg.Rgba(185, 0, 165), this.hero.drawX - 5, this.hero.drawY - 45, true));
                    }
                }
                if (RogueRpg.Level[RogueRpg.L.Dodge] && !blockDamage) {
                    var rand = this.playerRandom(false, 400);
                    if (rand > 1 - 0.10 * RogueRpg.Level[RogueRpg.L.Dodge] && Math.random() > 0.01) {
                        blockDamage = true;
                        this.floatingText.push(new RogueRpg.FloatingText("Dodge!", new RogueRpg.Rgba(185, 0, 165), this.hero.drawX - 5, this.hero.drawY - 45, true));
                    }
                }
                this.animateAttack(this.hero, "attack");
                monster.attackingTimer = 50;
                if (monster.attackImages && monster.attackImages.length > 0) {
                    monster.attackingTimer = 250;
                }
                if (!blockDamage) {
                    this.attackHero(monster);
                    // Handle Counterattacks that only occur when you take damage
                    if (RogueRpg.Level[RogueRpg.L.CounterattackThornsWeapon]) {
                        this.damageMonster(monster, Math.ceil(this.hero.weaponDamage * 0.20 * RogueRpg.Level[RogueRpg.L.CounterattackThornsWeapon]), /*isAttack=*/ false);
                    }
                    if (RogueRpg.Level[RogueRpg.L.CounterattackResetDelay]) {
                        var rand = RogueRpg.Level[RogueRpg.L.LuckOnCombatEffects] ? this.playerRandom() : Math.random();
                        if (rand > 1 - 0.40 * RogueRpg.Level[RogueRpg.L.CounterattackResetDelay]) {
                            this.hero.remainingDelay = 0;
                        }
                    }
                }
                if (RogueRpg.Level[RogueRpg.L.ReduceDelayOnDamage]) {
                    this.hero.remainingDelay -= 120;
                }
                if (RogueRpg.Level[RogueRpg.L.AttackWhileMoving]) {
                    if (this.hero.remainingDelay <= 0) {
                        this.hero.remainingDelay = this.hero.weaponDelay;
                        var mult = this.movedThisCycle ? 2.0 : 1.0;
                        this.attackMonster(monsterIndex, mult);
                        this.hero.attackingTime = this.hero.weaponDelay / 2;
                    }
                }
            }
        };
        Game.prototype.handleIfMouseClickedOnButton = function (x, y) {
            if (this.hero.state.levelUpPoints > 0) {
                this.renderer.staticContext.font = "24px Helvetica";
                var textWidth = this.renderer.staticContext.measureText("楼层上升!").width;
                if (x > this.renderer.ww - textWidth - 50 && y >= 20 && y <= 82) {
                    this.showLevelUpPaneButtonPressed();
                    return true;
                }
            }
        };
        Game.prototype.handleIfMouseIsOverUnit = function (x, y) {
            var possibleTargets = [];
            for (var i = this.map.monsters.length - 1; i >= 0; i--) {
                var monster = this.map.monsters[i];
                if (monster && RogueRpg.Dude.containsX(x, monster, 5) && RogueRpg.Dude.containsY(y, monster, 5)) {
                    possibleTargets.push(monster);
                }
            }
            if (possibleTargets.length > 0) {
                possibleTargets.sort(function (a, b) { return a.life / a.maxLife - b.life / b.maxLife; });
                var monster = possibleTargets[0];
                var monsterIndex = this.map.monsters.indexOf(monster);
                if (RogueRpg.Helpers.distance(this.hero.x, monster.x, this.hero.y, monster.y) < this.hero.attackRange) {
                    this.handleBattlingMonster(monsterIndex);
                    return true;
                }
                else if (RogueRpg.HeroClass.Classes[this.hero.getHeroState().classIndex].skill == "Flying Kick" && this.hero.specialDelay <= 0) {
                    this.flyingKickCount = 0;
                    this.flyingKickTimer = 0;
                    this.triggerFlyingKick(monster);
                    return true;
                }
            }
            for (var i = 0; i < this.map.specialWorldObjects.length; i++) {
                var obj = this.map.specialWorldObjects[i];
                if (obj.isUnit) {
                    return true;
                }
            }
            return false;
        };
        Game.prototype.triggerFlyingKick = function (monster, startingTime, exclude, endAt) {
            var _this = this;
            if (startingTime === void 0) { startingTime = 0; }
            if (exclude === void 0) { exclude = []; }
            // In order to reduce old anims, speed up previous kicks 15% when a new ricochet begins
            for (var i = 0; i < this.activeAnimations.length; i++) {
                var anim = this.activeAnimations[i];
                if (anim.name.substr(0, 11) == "Flying Kick") {
                    anim.maxTime *= 0.9;
                    anim.timeElapsed *= 0.9;
                }
            }
            startingTime *= 0.9;
            this.hero.specialDelayStart = 500;
            this.hero.specialDelay = this.hero.specialDelayStart * (1 - this.hero.getCooldownReduction());
            var oldLocation = new RogueRpg.Location(this.hero.x, this.hero.y);
            this.hero.x = monster.x;
            this.hero.y = monster.y;
            this.activeAnimations.push(new RogueRpg.Animation("Smash", this.hero.x, this.hero.y, Game.deltaTime * 10));
            var damage = this.hero.weaponDamage;
            this.applyRange(damage);
            this.damageMonster(monster, 1000 * damage / this.hero.weaponDelay);
            var dist = RogueRpg.Helpers.dist(monster, oldLocation);
            var dx = monster.x - oldLocation.x;
            var dy = monster.y - oldLocation.y;
            var animationName = dx > 0 ? "Flying Kick" : "Flying Kick Reverse";
            for (var i = 0; i < dist; i += 35) {
                this.activeAnimations.push(new RogueRpg.Animation(animationName, oldLocation.x + (i / dist) * dx, oldLocation.y + (i / dist) * dy, 3 + i * 1.6 + startingTime));
            }
            if (RogueRpg.Level[RogueRpg.L.KickKnockback]) {
                exclude.push(monster);
                var range = 100;
                var targets = this.listMonstersInRange(this.hero.x, this.hero.y, range, function (mon) {
                    return mon != monster;
                });
                this.affectArea(this.hero.x, this.hero.y, 90, new RogueRpg.ActiveEffect("Slow", 1500, { "amount": 0.7 }));
                for (var i = 0; i < targets.length; i++) {
                    this.pushUnitFromUnit(targets[i], this.hero, 120);
                }
            }
            if (RogueRpg.Level[RogueRpg.L.KickRicochet]) {
                if (monster.life <= 0 || Math.random() < 0.5) {
                    var range = 200;
                    this.timeout(function () {
                        _this.hero.specialDelayStart = 400;
                        _this.hero.specialDelay = _this.hero.specialDelayStart;
                        var targets = _this.listMonstersInRange(_this.hero.x, _this.hero.y, range, function (mon) {
                            return mon.life > 0 && mon != monster && exclude.indexOf(monster) === -1;
                        });
                        if (targets.length > 0) {
                            if (!endAt) {
                                endAt = new RogueRpg.Location(_this.hero.x, _this.hero.y);
                            }
                            var target = targets[0];
                            _this.flyingKickCount++;
                            _this.flyingKickTimer = 0;
                            _this.triggerFlyingKick(target, 3 + i * 2 + startingTime, exclude, endAt);
                            return;
                        }
                        else {
                            // Display the flying kick count
                        }
                    }, 75);
                }
                else {
                    // Display the flying kick count
                }
            }
        };
        Game.prototype.pushUnitFromUnit = function (target, source, falloffLimit) {
            var _this = this;
            var angle = Math.atan2(target.y - source.y, target.x - source.x);
            // TODO: In order to get animation, maybe do this via an ActiveEffect
            var power = 1 - (RogueRpg.Helpers.dist(target, source) / falloffLimit);
            this.moveUnitX(target, 100 * power * Math.cos(angle));
            this.moveUnitY(target, 100 * power * Math.sin(angle));
            this.timeout(function () {
                _this.moveUnitX(target, 60 * power * Math.cos(angle));
                _this.moveUnitY(target, 60 * power * Math.sin(angle));
                _this.timeout(function () {
                    _this.moveUnitX(target, 40 * power * Math.cos(angle));
                    _this.moveUnitY(target, 40 * power * Math.sin(angle));
                    _this.timeout(function () {
                        _this.moveUnitX(target, 20 * power * Math.cos(angle));
                        _this.moveUnitY(target, 20 * power * Math.sin(angle));
                        _this.timeout(function () {
                            _this.moveUnitX(target, 10 * power * Math.cos(angle));
                            _this.moveUnitY(target, 10 * power * Math.sin(angle));
                        }, 50);
                    }, 50);
                }, 50);
            }, 50);
        };
        Game.prototype.timeout = function (callback, left) {
            this.callbacks.push({ callback: callback, timeLeft: left });
        };
        Game.prototype.listMonstersInRange = function (x, y, range, condition) {
            var monsters = [];
            for (var i = 0; i < this.map.monsters.length; i++) {
                var mon = this.map.monsters[i];
                if (RogueRpg.Helpers.dist({ x: x, y: y }, mon) < range) {
                    if (!condition || condition(mon)) {
                        monsters.push(mon);
                    }
                }
            }
            return monsters;
        };
        Game.prototype.affectArea = function (x, y, range, effect) {
            var units = this.listMonstersInRange(x, y, range);
            for (var i = 0; i < units.length; i++) {
                var u = units[i];
                RogueRpg.Dude.addActiveEffect(u, effect);
            }
        };
        Game.prototype.handleSpecialDoubleTap = function (x, y) {
            var _this = this;
            if (this.hero.specialDelay <= 0 && this._doubleTappedThisCycle) {
                var skill = RogueRpg.HeroClass.Classes[this.hero.getHeroState().classIndex].skill;
                if (skill == "Stab") {
                    this.hero.specialDelayStart = 8000;
                    var duration = 1000;
                    if (RogueRpg.Level[RogueRpg.L.StabDuration]) {
                        duration += 500;
                    }
                    this.hero.state.activeEffects.push(new RogueRpg.ActiveEffect("Stab", duration, { delayRemaining: 0 }));
                }
                else if (skill == "sweep") {
                    this.hero.specialDelayStart = 5000;
                    this.attackSpecial(x, y, "sweep");
                    if (RogueRpg.Level[RogueRpg.L.SweepSpins]) {
                        window.setTimeout(function () {
                            _this.attackSpecial(x, y, "sweep");
                        }, 250);
                    }
                }
                else if (skill == "snap") {
                    if (RogueRpg.Helpers.IsWalkable(new RogueRpg.Location(x, y), this.map)) {
                        this.attackSpecial(this.hero.x, this.hero.y, "snap");
                        this.hero.x = x;
                        this.hero.y = y;
                        this.hero.specialDelayStart = 6000;
                        if (RogueRpg.Level[RogueRpg.L.FlashCooldown]) {
                            this.hero.specialDelayStart = 4000;
                        }
                        this.attackSpecial(x, y, "snap");
                    }
                }
                else if (skill == "Smash") {
                    var distance = RogueRpg.Helpers.distance(x, this.hero.x, y, this.hero.y);
                    var ratio = distance / this.hero.attackRange;
                    if (ratio > 1) {
                        x = this.hero.x + (x - this.hero.x) / ratio;
                        y = this.hero.y + (y - this.hero.y) / ratio;
                    }
                    this.hero.specialDelayStart = 5000;
                    this.attackSpecial(x, y, "Smash");
                }
                else if (skill == "Flying Kick") {
                    return;
                }
                else if (skill == "arrows") {
                    this.hero.specialDelayStart = 5000;
                    this.attackSpecial(x, y, "arrows");
                }
                this.hero.specialDelay = this.hero.specialDelayStart * (1 - this.hero.getCooldownReduction());
                if (this.hero.remainingDelay <= 0) {
                    this.hero.remainingDelay = Math.min(150, this.hero.weaponDelay);
                }
            }
        };
        Game.prototype.handleBattlingMonster = function (monsterIndex) {
            var monster = this.map.monsters[monsterIndex];
            // this.hero.remainingDelay -= Game.deltaTime;
            if (this.hero.remainingDelay <= 0) {
                this.hero.remainingDelay = this.hero.weaponDelay;
                this.attackMonster(monsterIndex);
                this.hero.attackingTime = this.hero.weaponDelay / 2;
            }
        };
        Game.prototype.attackSpecial = function (x, y, type) {
            var _this = this;
            var damage = this.applyRange(this.hero.weaponDamage);
            if (type == "Smash") {
                this.spriteAnimations.push(RogueRpg.Images.Sprites.Smash.getInstance(new RogueRpg.Location(x, y)));
                this.damageArea(x, y, damage * (1250 / this.hero.weaponDelay), 60);
            }
            else if (type == "sweep") {
                this.spriteAnimations.push(RogueRpg.Images.Sprites.Sweep.getInstance(new RogueRpg.Location(this.hero.drawX + this.hero.drawWidth / 2, this.hero.drawY + this.hero.drawHeight / 2)));
                this.damageArea(this.hero.x, this.hero.y, damage, 110);
                if (RogueRpg.Level[RogueRpg.L.SweepStun]) {
                    this.affectArea(this.hero.x, this.hero.y, 110, new RogueRpg.ActiveEffect("Stun", 1000, {}));
                }
            }
            else if (type == "snap") {
                this.spriteAnimations.push(RogueRpg.Images.Sprites.Snap.getInstance(new RogueRpg.Location(this.hero.drawX + this.hero.drawWidth / 2, this.hero.drawY + this.hero.drawHeight / 2)));
                var newDmg = damage * 0.7 * Math.max(400 / this.hero.weaponDelay, 1);
                if (RogueRpg.Level[RogueRpg.L.FlashDamage]) {
                    newDmg *= 1.5;
                }
                this.damageArea(x, y, newDmg, 100);
            }
            else if (type == "arrows") {
                // TODO: Determine if I need to apply a damage range
                var damage = this.applyRange(this.hero.weaponDamage);
                var speed = 30;
                var repeatCount = RogueRpg.Level[RogueRpg.L.ArrowsDouble] ? 2 : 1;
                this.repeatTimes(function () {
                    if (RogueRpg.Level[RogueRpg.L.ArrowsAim]) {
                        _this.repeatTimes(function () {
                            var direction = Math.atan2(y - _this.hero.y, x - _this.hero.x);
                            direction += 0.3 - Math.random() * 0.6;
                            var xv = Math.cos(direction) * speed;
                            var yv = Math.sin(direction) * speed;
                            var projectile = new RogueRpg.Projectile(_this.hero, "Arrow", _this.hero.x, _this.hero.y, xv, yv, damage, 40);
                            _this.fireProjectile(projectile);
                        }, 11, 40);
                    }
                    else {
                        var arrowCount = 15;
                        for (var i = 0; i < arrowCount; i++) {
                            var direction = (i / arrowCount) * Math.PI * 2;
                            var xv = Math.cos(direction) * speed;
                            var yv = Math.sin(direction) * speed;
                            var projectile = new RogueRpg.Projectile(_this.hero, "Arrow", _this.hero.x, _this.hero.y, xv, yv, damage, 40);
                            _this.fireProjectile(projectile);
                        }
                    }
                }, repeatCount, 500);
            }
        };
        Game.prototype.repeatTimes = function (callback, times, delay) {
            var _this = this;
            if (times <= 0)
                return;
            callback();
            setTimeout(function () {
                _this.repeatTimes(callback, times - 1, delay);
            }, delay);
        };
        Game.prototype.attackMonster = function (monsterIndex, damageMultiplier, animate) {
            if (damageMultiplier === void 0) { damageMultiplier = 1; }
            if (animate === void 0) { animate = true; }
            var monster = this.map.monsters[monsterIndex];
            var damage = this.applyRange(this.hero.weaponDamage) * damageMultiplier;
            var isRangedAttack = RogueRpg.HeroClass.Classes[this.hero.getHeroState().classIndex].isRanged;
            if (isRangedAttack) {
                var speed = 30;
                var x = monster.x - this.hero.x;
                var y = monster.y - this.hero.y;
                var direction = Math.atan2(y, x);
                var xVelocity = Math.cos(direction) * speed;
                var yVelocity = Math.sin(direction) * speed;
                // Note: Subtract the velocity, because the proejctile will be updated this same game cycle
                var projectile = new RogueRpg.Projectile(this.hero, "Arrow", this.hero.x - xVelocity / 2, this.hero.y - yVelocity / 2, xVelocity, yVelocity, damage, 40);
                this.fireProjectile(projectile);
            }
            else {
                if (RogueRpg.Level[RogueRpg.L.Disarm]) {
                    if (Math.random() > 0.995) {
                        this.rollForDrop(monster.level, false, monster.drawX, monster.drawY, "Weapon");
                    }
                }
                if (animate) {
                    this.animateAttack(monster, "attack");
                }
                this.damageMonster(monster, damage);
            }
        };
        // Fire from unit, in direction
        Game.prototype.fireProjectile = function (projectile) {
            this.projectiles.push(projectile);
        };
        // TODO: Implement useIntersection
        Game.prototype.damageArea = function (x, y, damage, distance, isAttack, useIntersection, condition) {
            if (isAttack === void 0) { isAttack = true; }
            if (useIntersection === void 0) { useIntersection = true; }
            for (var i = this.map.monsters.length - 1; i >= 0; i--) {
                var mon = this.map.monsters[i];
                if (mon) {
                    if (!condition || condition(mon)) {
                        if (RogueRpg.Helpers.distance(x, mon.x, y, mon.y - mon.drawHeight / 2) < distance + (mon.drawWidth + mon.drawHeight) / 10) {
                            this.damageMonster(mon, damage, isAttack);
                        }
                    }
                }
            }
        };
        Game.prototype.damageMonster = function (monster, damage, isAttack) {
            if (isAttack === void 0) { isAttack = true; }
            if (!monster) {
                return;
            }
            monster.wasHitTimer = isAttack ? 150 : 50;
            var nonCritChance = 0.9 - RogueRpg.Level[RogueRpg.L.IncreasedCritChance] * 0.1;
            var isCrit = (RogueRpg.Level[RogueRpg.L.LuckOnCombatEffects] ? this.playerRandom(false) : Math.random()) > nonCritChance;
            if (isCrit) {
                var critMultiplier = 2 + RogueRpg.Level[RogueRpg.L.IncreasedCritDamage];
                damage = damage * critMultiplier;
            }
            if (RogueRpg.Level[RogueRpg.L.NoLifesteal] && isAttack) {
                this.healHeroFor(this.hero.maxLife * 0.008);
            }
            var actualDamage = this.rawDamageUnit(damage, monster, isCrit);
            if (!RogueRpg.Level[RogueRpg.L.NoLifesteal] && isAttack) {
                var lifeStolen = actualDamage * this.hero.getLifestealPercent() / 100;
                this.healHeroFor(lifeStolen);
            }
            else {
                // This is accounted for above
            }
            if (monster.life <= 0) {
                if (RogueRpg.Level[RogueRpg.L.Cleave]) {
                    this.hero.remainingDelay = 0;
                }
            }
        };
        // Returns the amount actually done
        Game.prototype.rawDamageUnit = function (damage, target, isCrit) {
            // damage multiplier is applied additively by Flurry skill
            if (typeof target.damageMultiplier !== 'number') {
                target.damageMultiplier = 1;
            }
            damage *= target.damageMultiplier;
            // toughness is computed damage reduction based on difficulty and floor
            if (typeof target.toughness !== 'number') {
                target.toughness = 1;
            }
            damage *= target.toughness;
            if (RogueRpg.Level[RogueRpg.L.AmplifyingAttacks]) {
                target.damageMultiplier += 0.02 * RogueRpg.Level[RogueRpg.L.AmplifyingAttacks];
            }
            damage = Math.round(damage);
            target.life -= damage;
            target.tookDamageTimer = 200;
            if (isCrit) {
                this.floatingText.push(new RogueRpg.FloatingText("-" + RogueRpg.Helpers.formatNumber(Math.floor(damage)) + "!", new RogueRpg.Rgba(255, 255, 0), target.drawX + target.drawWidth / 2 + (RogueRpg.Renderer.offsetMap[target.type] || 0), target.drawY - 45, true));
            }
            else {
                this.floatingText.push(new RogueRpg.FloatingText("-" + RogueRpg.Helpers.formatNumber(Math.floor(damage)), new RogueRpg.Rgba(255, 255, 255), target.drawX + target.drawWidth / 2 + (RogueRpg.Renderer.offsetMap[target.type] || 0), target.drawY - 45, true));
            }
            if (target.life <= 0) {
                this.handleMonsterDeath(target);
            }
            return damage;
        };
        Game.prototype.handleMonsterDeath = function (monster) {
            this.map.deadMonsters.push(new RogueRpg.DeadMonster(monster));
            if (!monster.skipLoot) {
                if (Math.random() < monster.dropRateModifier || 1) {
                    this.rollForDrop(monster.level, monster.isRare, monster.drawX, monster.drawY);
                }
            }
            var expMultiplier = 1;
            this.dropShards(Math.round(Math.sqrt(monster.level)), new RogueRpg.Location(monster.x, monster.y), 50);
            this.grantExperience(monster.experienceGrantedOnKill);
            if (Math.random() > 1 - (0.12 * RogueRpg.Level[RogueRpg.L.ChanceDoubleExp])) {
                this.grantExperience(monster.experienceGrantedOnKill);
            }
            this.handleMonsterDeathSpecialCases(monster);
            // Normal monsters have a 3% chance to drop gemstones, but rare are much more likely
            if (monster.isRare && Math.random() > 0.7 && this.playerRandom(true) > 0.6) {
                this.dropGemstones(5 + Math.floor(monster.level / 5), monster.isRare, monster.drawX, monster.drawY);
            }
            else if (Math.random() > 0.5 && this.playerRandom(true) > 0.97) {
                this.dropGemstones(3 + Math.floor(monster.level / 6), monster.isRare, monster.drawX, monster.drawY);
            }
            this.map.monsters.splice(this.map.monsters.indexOf(monster), 1);
            this.needsToPersist = true;
            //this.reviveMonster(monster);
        };
        Game.prototype.consumeHealingPotion = function () {
            var _this = this;
            if (!this.tutorialCompleted) {
                if (this.tutorialState == 12) {
                    this.tutorialState++;
                }
                else {
                    return;
                }
            }
            // TODO: Animate and do other things
            if (this.hero.state.healingPotionCount > 0) {
                if (this.hero.life < this.hero.maxLife) {
                    this.hero.state.healingPotionCount--;
                    this.healHeroFor(this.hero.maxLife * Game.HealingPotionHealPercent / 100);
                }
            }
            else {
                this.pauseLoop();
                this.showInGameStore(function () {
                    _this.beginLoop();
                });
            }
        };
        Game.prototype.reviveMonster = function (monster) {
            var pet = RogueRpg.Pet.PetFromMonster(monster);
            pet.lifeTimer = 7000;
            pet.agroRange = 2000;
            pet.speed += 15;
            pet.delay *= 0.66;
            this.hero.state.pets.push(pet);
        };
        Game.prototype.handleMonsterDeathSpecialCases = function (monster) {
            // TODO: The chance of an enemy randomly dropping something like this is actually really nice,
            // especially when tied to something like a rare monster. I need to re-introduce minor relics, which would
            // be like hats. In fact, maybe certain minor relics could be worn as a hat?
            if (monster.type == "Boarman") {
                this.rollForDrop(monster.level, true, monster.drawX, monster.drawY, "Weapon", "Decapitator");
                if (this.playerRandom(true) > 0.5) {
                    this.rollForDrop(monster.level, true, monster.drawX, monster.drawY);
                }
            }
        };
        Game.prototype.dropGemstones = function (level, isRare, drawX, drawY) {
            var gemstoneCount = level || randInt(2, 4);
            this.map.gemstonesOnGround.push(new RogueRpg.SpecialDroppedItem(gemstoneCount, drawX, drawY, "gemstones", 0.5));
        };
        Game.prototype.attackHero = function (monster) {
            var damage = this.applyRangeUnmodified(monster.damage);
            if (RogueRpg.Level[RogueRpg.L.WeakenOnHit] && monster.life < monster.maxLife) {
                damage *= 0.85 * RogueRpg.Level[RogueRpg.L.WeakenOnHit];
            }
            var reduction = this.hero.reduction;
            reduction += RogueRpg.Level[RogueRpg.L.ReductionWhileMoving] * 0.35 * reduction;
            damage = Math.max(damage - reduction, 0);
            if (damage == 0 && Math.random() > 0.5) {
                damage = 1;
            }
            var defenseModifier = 1 - this.hero.getDefenseReductionPercent(monster.level) / 100;
            damage = Math.round(damage * defenseModifier);
            this.damageHero(damage, monster);
        };
        // Takes 100% and makes it 50%-150%
        Game.prototype.applyRange = function (input, invertRange) {
            var fraction = 0.5;
            fraction -= RogueRpg.Level[RogueRpg.L.LuckyStrike] * 0.5;
            var val = fraction * input;
            var rand = invertRange ? 1 - this.playerRandom() : this.playerRandom();
            val += rand * input * (1 + RogueRpg.Level[RogueRpg.L.LuckyStrike] * 1);
            return val;
        };
        Game.prototype.applyRangeUnmodified = function (input) {
            var fraction = 0.5;
            var val = fraction * input;
            val += Math.random() * input;
            return val;
        };
        Game.prototype.animateAttack = function (unit, type, duration) {
            if (duration === void 0) { duration = 300; }
            if (unit) {
                if (type == "attack") {
                    var m = 10;
                    this.spriteAnimations.push(RogueRpg.Images.Sprites.Slash.getInstance(new RogueRpg.Location(unit.drawX + unit.drawWidth / 2 - m + Math.random() * 2 * m, unit.drawY + unit.drawHeight / 2 - m + Math.random() * 2 * m)));
                }
                else if (type == "Stab") {
                    /*var anim = new Animation(type, unit.drawX + unit.drawWidth / 2, unit.drawY + unit.drawHeight / 2, 400,
                        Helpers.angle(this.hero, unit));*/
                    var anim = new RogueRpg.Animation(type, this.hero.drawX + this.hero.drawWidth / 2, this.hero.drawY + this.hero.drawHeight / 2, 400, RogueRpg.Helpers.angle(this.hero, unit));
                    anim.velocity = 450;
                    anim.drag = 3000;
                    this.activeAnimations.push(anim);
                }
                else {
                    this.activeAnimations.push(new RogueRpg.Animation(type, unit.drawX + unit.drawWidth / 2, unit.drawY + unit.drawHeight / 2, duration));
                }
            }
        };
        // Handles damaging the hero and leading to death if necessary
        Game.prototype.damageHero = function (amount, source) {
            if (this.lockHeroMovement || this.clearingCurrentLevel) {
                return; // Take no damage during the tutorial
            }
            // amount += Level[L.Berserking] * 0.005 * (1 - this.hero.life / this.hero.maxLife);
            amount *= 1 + RogueRpg.Level[RogueRpg.L.Berserking] * 0.3;
            amount *= 1 - (0.1 * RogueRpg.Level[RogueRpg.L.ReduceDamage] || 0);
            if (RogueRpg.Level[RogueRpg.L.LuckyDefense]) {
                amount = this.applyRange(amount, /*invertRange=*/ true);
            }
            if (RogueRpg.Level[RogueRpg.L.DefWhileMoving] && this.movedThisCycle) {
                amount *= 0.666;
            }
            if (RogueRpg.Level[RogueRpg.L.Bloodied]) {
                amount *= 0.6;
            }
            var prismaDef = this.prismaState["LessDamage"] || 0;
            amount *= 1 - (prismaDef * RogueRpg.PermanentUpgrade.Upgrades.LessDamage.costPerRank * 0.01); // Convert to percent
            this.floatingText.push(new RogueRpg.FloatingText("-" + RogueRpg.Helpers.formatNumber(Math.round(amount)), new RogueRpg.Rgba(200, 0, 0), this.hero.drawX - 5, this.hero.drawY - 45, true));
            this.animateDamageTaken(amount, this.hero.life);
            this.hero.life -= amount;
            if (this.hero.life <= 0) {
                this.handleHeroDeath(source);
            }
        };
        // Animates the life bar chunking away at a given point
        Game.prototype.animateDamageTaken = function (damage, farEndOfRegion) {
            var chunk = new RogueRpg.LosingLifeChunk(damage / this.hero.maxLife, (farEndOfRegion - damage) / this.hero.maxLife);
            this.losingLifeChunks.push(chunk);
        };
        Game.prototype.handleHeroDeath = function (killer) {
            this.pauseLoop();
            //ga_storage._trackEvent('Death', killer.type);
            this.showDeathPrompt(killer);
        };
        Game.prototype.handleRelicExpAfterDeath = function () {
            by.id("inventoryAccess").style.visibility = "hidden";
            this.screenCaptureAfterDeath.height = this.renderer.wh;
            this.screenCaptureAfterDeath.width = this.renderer.ww;
            // var imgData = this.renderer.context.getImageData(0, 0, this.renderer.ww, this.renderer.wh);
            var ctx = this.screenCaptureAfterDeath.getContext("2d");
            this.screenCaptureAfterDeath.style.display = "block";
            ctx.drawImage(this.renderer.canvas, 0, 0, this.renderer.ww, this.renderer.wh);
            this._tappedThisCycle = false;
            this.shardTimer = 80;
            this.relicAnimationPhase = "FadingIn";
            this.bonusShards = Math.floor(this.shards * (RogueRpg.Level[RogueRpg.L.ShardFlatBonus] * 0.15 + (this.bonusShardsThisGame ? 0.3 : 0)));
            var shardBonusFromUpgradesRank = this.prismaState.levels[RogueRpg.PermanentUpgrade.Upgrades.ShardBonus.id] || 0;
            var totalShards = this.shards + this.bonusShards;
            this.bonusShards += Math.floor(totalShards * shardBonusFromUpgradesRank * RogueRpg.PermanentUpgrade.Upgrades.ShardBonus.gainPerRank * 0.01); // Convert from pct
            this.bonusShardsThisGame = false;
            this.relicStateTimer = 0;
            //this.originalRelicLevel = Relic.RelicLevels[this.lastRelicIndex];
            this.shardGainShowing = true;
            this.shownRelicPrompt = false;
            this.persistData();
            this.renderer.staticContext.clearRect(0, 0, this.renderer.ww, this.renderer.wh);
        };
        Game.prototype.shardGainLoop = function () {
            var _this = this;
            var wh = this.renderer.wh;
            var ww = this.renderer.ww;
            this.renderer.drawShardGainLoop();
            if (this.relicStateTimer >= 0.6) {
                this.relicAnimationPhase = "";
            }
            if (this.relicAnimationPhase == "FadingIn") {
                this.relicStateTimer += 0.008;
                return;
            }
            var updated = false;
            if (this.shards > 0) {
                updated = true;
                var amount = Math.min(this.shards, this.shards > 100 ? Math.floor(this.shards * 0.05) : (this.shards > 5 ? 5 : 1));
                var mult = 1;
                while (amount > 2) {
                    amount = Math.floor(amount / 2);
                    mult = mult * 2;
                }
                this.setShards(this.shards - (amount * mult));
                for (var i = 0; i < amount; i++) {
                    this.movingShards.push(new RogueRpg.MovingShard(new RogueRpg.Location(100, wh - 50), new RogueRpg.Location(ww / 2, 180), mult));
                }
            }
            else if (this.bonusShards > 0) {
                updated = true;
                var amount = Math.min(this.bonusShards, this.bonusShards > 100 ? Math.floor(this.bonusShards * 0.04) : (this.bonusShards < 5 ? 1 : 2));
                var mult = 1;
                while (amount > 3) {
                    amount = Math.floor(amount / 2);
                    mult = mult * 2;
                }
                this.bonusShards -= amount * mult;
                for (var i = 0; i < amount; i++) {
                    this.movingShards.push(new RogueRpg.MovingShard(new RogueRpg.Location(215, wh - 50), new RogueRpg.Location(ww / 2, 180), mult));
                }
            }
            for (var i = 0; i < this.movingShards.length; i++) {
                var ms = this.movingShards[i];
                RogueRpg.MovingShard.Update(ms);
                this.renderer.context.drawImage(RogueRpg.Images.get("shardEnergy"), ms.location.x, ms.location.y);
                if (RogueRpg.Helpers.distance(ms.location.x, ms.target.x, ms.location.y, ms.target.y) < 25) {
                    this.movingShards.splice(i, 1);
                    i--;
                    updated = true;
                    this.storedShards += ms.value;
                }
            }
            this.shardTimer -= 1;
            // showSecondaryConfirmPrompt
            if (this.shards == 0 && this.movingShards.length == 0 && this.shardTimer <= 0) {
                if (this._tappedThisCycle) {
                    // this.exitShardGain();
                }
                else if (!this.shownRelicPrompt) {
                    this.shownRelicPrompt = true;
                    this.persistData();
                    RogueRpg.Panels.showSecondaryConfirmPrompt("您现在可以使用收集的碎片升级您的纪念物!\n\n"
                        + "请立即执行此操作以变得更强大，然后重试！", "纪念物", function () {
                        _this.exitShardGain();
                        _this.showRelicListMenu(true, function () { });
                    }, "退出", function () {
                        _this.exitShardGain();
                    });
                }
            }
            this._tappedThisCycle = false;
        };
        Game.prototype.exitShardGain = function () {
            this.pauseLoop();
            this.screenCaptureAfterDeath.style.display = "none";
            this._tappedThisCycle = false;
            this.shardGainShowing = false;
            this.renderer.renderFloorSection = true;
            this.renderer.renderMinimap = true;
            this.renderer.renderLevelBar = true;
            this.renderer.renderShardsSection = true;
            by.id("inventoryAccess").style.visibility = "visible";
            this.beginNewGameAfterDeath();
            this.persistData();
        };
        Game.prototype.beginNewGameAfterDeath = function () {
            var _this = this;
            var potions = this.hero.state.potions;
            var healingPotionCount = this.hero.state.healingPotionCount;
            this.hero = new RogueRpg.Hero();
            this.hero.state.potions = potions;
            this.hero.state.healingPotionCount = healingPotionCount;
            RogueRpg.Helpers.hero = this.hero;
            this.startNewLevel();
            RogueRpg.Level = {
                Life: 0,
                Str: 0,
                Dex: 0,
                Dmg: 0,
                Def: 0,
                Speed: 0,
                Luck: 0
            };
            var levelUpNames = Object.getOwnPropertyNames(RogueRpg.L);
            for (var i = 0; i < levelUpNames.length; i++) {
                RogueRpg.Level[RogueRpg.L[levelUpNames[i]]] = 0; // set all to 0, later on values can be added by persistence
            }
            this.persistData();
            this.showClassSelectionPane(function () {
                if (!_this.tutorialCompleted) {
                    _this.setupTutorial();
                }
                _this.beginLoop();
            });
        };
        Game.prototype.inGemstoneSection = function (x, y) {
            return this.isInSection(x, y, RogueRpg.Renderer.GemstoneSectionStart);
        };
        Game.prototype.inShardSection = function (x, y) {
            return this.isInSection(x, y, RogueRpg.Renderer.ShardSectionStart);
        };
        Game.prototype.inInventoryAccess = function (x, y) {
            var left = this.renderer.ww - 60;
            var top = this.renderer.wh - 100;
            return x >= left && x <= left + 50 && y >= top && y <= top + 50;
        };
        Game.prototype.inPotionSection = function (x, y) {
            // Note: These should be kept in sync with the renderer
            var size = RogueRpg.Renderer.HealingPotionSize;
            var left = this.renderer.ww - RogueRpg.Renderer.HealingPotionRightOffset;
            var top = this.renderer.wh - RogueRpg.Renderer.HealingPotionBottomOffset;
            return x >= left && x <= left + size && y >= top && y <= top + size;
        };
        Game.prototype.isInSection = function (x, y, loc) {
            return x >= loc.left && x <= loc.left + loc.width
                && y > this.renderer.wh - loc.bottom && y <= this.renderer.wh - loc.bottom + loc.height;
        };
        /*private onMouseDown(e: MouseEvent) {
            this._mouseX = e.clientX;
            this._mouseY = e.clientY;
            this._mouseDown = true;
            this._tappedThisCycle = true;
            //  this.closeInventory();
        }

        private onMouseUp(e: MouseEvent) {
            //this._mouseX = e.clientX;
            //this._mouseY = e.clientY;
            this._mouseDown = false;
        }

        private onMouseMove(e: MouseEvent) {
            if (this._mouseDown) {
                this._mouseX = e.clientX;
                this._mouseY = e.clientY;
            }
        }*/
        Game.prototype.onTouchStart = function (e /*TouchEvent?*/) {
            var touch = e.touches[0];
            if (touch) {
                // TODO: Problems with this model, removed access here temporarily
                /*if (false && REAL_MONEY_ENABLED && this.inGemstoneSection(touch.clientX, touch.clientY)) {
                    this.gemstoneSectionDown = true;
                    e.preventDefault();
                    return true;
                } else */
                if (this.inShardSection(touch.clientX, touch.clientY)) {
                    this.shardSectionDown = true;
                    e.preventDefault();
                    return true;
                }
                else if (this.inInventoryAccess(touch.clientX, touch.clientY)) {
                    this.inventoryAccessDown = true;
                    e.preventDefault();
                    return true;
                }
                else if (this.inPotionSection(touch.clientX, touch.clientY)) {
                    this.healingPotionButtonDown = true;
                    e.preventDefault();
                    return true;
                }
                this._mouseX = touch.clientX;
                this._mouseY = touch.clientY;
                this._mouseDown = true;
                this._tappedThisCycle = true;
                //this.closeInventory();
            }
            e.preventDefault(); // TODO: game.js:4466 .
            return true;
        };
        Game.prototype.onTouchEnd = function (e /*TouchEvent?*/) {
            var _this = this;
            var touch = e.touches[0];
            if (touch) {
                /*this._mouseX = touch.clientX;
                this._mouseY = touch.clientY;*/
            }
            this._mouseDown = false;
            if (this.gemstoneSectionDown && this.hero.life >= 0) {
                this.showGemstoneMenu();
            }
            if (this.shardSectionDown && this.hero.life >= 0) {
                this.pauseLoop();
                this.showInGameStore(function () {
                    _this.beginLoop();
                });
            }
            if (this.inventoryAccessDown && this.hero.life >= 0) {
                this.showInventory();
            }
            if (this.healingPotionButtonDown && this.hero.life >= 0) {
                this.consumeHealingPotion();
            }
            this.shardSectionDown = false;
            this.gemstoneSectionDown = false;
            this.inventoryAccessDown = false;
            this.healingPotionButtonDown = false;
            e.preventDefault();
            return true;
        };
        Game.prototype.onTouchMove = function (e /*TouchEvent?*/) {
            var touch = e.touches[0];
            if (touch) {
                if (this._mouseDown) {
                    this._mouseX = touch.clientX;
                    this._mouseY = touch.clientY;
                    if (!this.inGemstoneSection(touch.clientX, touch.clientY)) {
                        this.gemstoneSectionDown = false;
                    }
                    else if (!this.inShardSection(touch.clientX, touch.clientY)) {
                        this.shardSectionDown = false;
                    }
                    else if (!this.inInventoryAccess(touch.clientX, touch.clientY)) {
                        this.inventoryAccessDown = false;
                    }
                    else if (!this.inPotionSection(touch.clientX, touch.clientY)) {
                        this.healingPotionButtonDown = false;
                    }
                }
            }
            e.preventDefault();
            return true;
        };
        Game.MonsterLevelForFloor = function (floor) {
            var monsterIndex = floor % Game.MonstersByFloor.length;
            var monsterForFloor = Game.MonstersByFloor[monsterIndex];
            var multiplierTier = Math.floor(floor / Game.MonstersByFloor.length);
            var mockMonster = new RogueRpg.Monster(Game.MonstersByFloor[monsterIndex], null, multiplierTier);
            return mockMonster.level;
        };
        Game.superHero = false; // One of the "cheats" for dev purposes
        Game.alwaysShowMiniMap = false; // One of the "cheat"s for dev purposes
        Game.renderPotionsOnMiniMap = false;
        Game.testCrashFix = true;
        Game.HealingPotionHealPercent = 40;
        Game.deltaTime = 0;
        Game.MonstersByFloor = [
            "Goblin",
            "Bat",
            "Mud Clown",
            "Goo",
            "Thorny Cactus",
            "Scorpiar",
            "Flying Skull",
            "Flytrap",
            "Wee Demon",
            "Mad Cube",
            "Bad Hero",
            "Ghost",
            "Snowman",
            "Eye",
            "Snake",
        ];
        // TODO: Currently unused - intended to support forward/back facing
        Game.bindings = {
            frontFacing: [
                { name: "hero" },
                { name: "armor", drawX: 0, drawY: 29 },
                { name: "helmet", drawX: 0, drawY: -9 },
                { name: "shield", drawX: 16, drawY: 17 },
                { name: "weapon", drawX: -25, drawY: 12 },
            ],
        };
        return Game;
    }());
    RogueRpg.Game = Game;
})(RogueRpg || (RogueRpg = {}));
function Make2dArray(rows, cols) {
    var ary = [];
    for (var i = 0; i < rows; i++) {
        var row = [];
        for (var j = 0; j < cols; j++) {
            row.push(null);
        }
        ary.push(row);
    }
    return ary;
}
function img(src, id, className) {
    var element = document.createElement("img");
    element.src = "images/" + src + ".png";
    if (id) {
        element.id = id;
    }
    if (className) {
        element.className = className;
    }
    return element;
}
function div(id, className, text) {
    var element = document.createElement("div");
    if (id) {
        element.id = id;
    }
    if (className) {
        element.className = className;
    }
    if (text) {
        element.innerText = text;
    }
    return element;
}
function span(text, className) {
    var element = document.createElement("span");
    if (text) {
        element.innerText = text;
    }
    if (className) {
        element.className = className;
    }
    return element;
}
function br() {
    return document.createElement("br");
}
// returns a random integer inclusive. Assumed that this is only used in item calcs
function randInt(low, high) {
    return Math.floor(low + RogueRpg.Helpers.game.playerRandom(true) * (1 + high - low));
}
function randIntMap(low, high) {
    return Math.floor(low + Math.random() * (1 + high - low));
}
// Casts a fraction to a percent
function percentIncr(num) {
    return Math.round((num - 1) * 100);
}
function green(text) {
    return "<span style='color: #00FF00'>" + text + "</span>";
}
function red(text) {
    return "<span style='color: #FF7744'>" + text + "</span>";
}
function blue(text) {
    return "<span style='color: #5599ff'>" + text + "</span>";
}
function light_blue(text) {
    return "<span style='color: #77BBff'>" + text + "</span>";
}
function yellow(text) {
    return "<span style='color: #FFEE00'>" + text + "</span>";
}
function bold(text) {
    return "<span style='font-weight: 800'>" + text + "</span>";
}
var by = {
    id: function (id) {
        return document.getElementById(id);
    },
    class: function (className) {
        return document.querySelector('.' + className);
    }
};
//# sourceMappingURL=game.js.map